#include "systick.h"
#include "AKD1000.h"
#include "common.h"
#include "systimer.h"

#define SYSTICK_LOAD_VAL (SystemCoreClock / 1000)

static volatile uint64_t _systicks;

static systick_callback_t systick_cb;

void SysTick_Handler(void) {
    _systicks++;

    if (systick_cb)
        systick_cb(_systicks);
}

uint64_t get_time_ms(void) {
    return (uint64_t)_systicks;
}

/*
 * Microseconds are calculated from the milliseconds and the rest is retrieved from difference
 * between the reload value and the current value of the CPU cycle counter
 */
uint64_t get_time_us(void) {
    uint64_t us_time;
    uint32_t flags = interrupt_lock();
    us_time = (uint64_t)((_systicks * 1000) +
                         (1000 * (uint64_t)(SYSTICK_LOAD_VAL - SysTick->VAL) / SYSTICK_LOAD_VAL));
    interrupt_unlock(flags);
    return us_time;
}

/*
 * nanoseconds are calculated from the nanoseconds and the rest is retrieved from difference between
 * the reload value and the current value of the CPU cycle counter.
 * The time time between a CPU clock transition is (1/SystemCoreClock).
 * @note With system clock frequency of 300 Mhz a granularity of 3.33 nanoseconds is achieved
 *
 */
uint64_t get_time_ns(void) {
    uint64_t ns_time;
    uint32_t flags = interrupt_lock();
    ns_time =
        (uint64_t)((_systicks * 1000000) +
                   ((1000000 * (uint64_t)(SYSTICK_LOAD_VAL - SysTick->VAL)) / SYSTICK_LOAD_VAL));
    interrupt_unlock(flags);
    return ns_time;
}

/* SysTicks block loads a counter with (SystemCoreClock / 1000) value and decrements the counter
 * on each CPU clock transition. Once count reaches 0 it will generate an interrupt and
 * increment _systick. The counter is automatically reloaded with (SystemCoreClock/1000)
 */
void systick_init(void) {
    SysTick_Config(SYSTICK_LOAD_VAL);
    /*Highest priority set for Systick interrupt*/
    NVIC_SetPriority(SysTick_IRQn, 0);
    NVIC_EnableIRQ(SysTick_IRQn);
}

/**
 * @brief  : function generates a delay in microseconds
 * @param  : period - timer period in microseconds
 * @retval : void
 */
void delay_us(uint32_t period) {
    if (period >= 1000) {
        uint64_t delay = get_time_us() + period;

        while (delay > get_time_us())
            ;
    } else {
        uint32_t ticks_per_us = SYSTICK_LOAD_VAL / 1000;
        int32_t t = SysTick->VAL - period * ticks_per_us;

        if (t < 0) {
            t += SYSTICK_LOAD_VAL;
            uint32_t reg;
            do {
                reg = SysTick->VAL;
            } while (reg < (uint32_t)t && reg > (SYSTICK_LOAD_VAL - t));
        } else {
            while (SysTick->VAL > (uint32_t)t)
                ;
        }
    }
}

/**
 * @brief  : systick_register_cb API is used to register a callback function.
 * @param  : cb the callback function to call on every systick interrupt
 * @retval : void
 */
int systick_register_cb(systick_callback_t cb) {
    BCFW_VALIDATE_NULL(cb);

    if (systick_cb) {
        return -1;
    }

    systick_cb = cb;

    return 0;
}

/**
 * @brief  : systick_deregister_cb API is used to de-register the callback function.
 * @param  : cb the callback function to be de-register
 * @return : error code incase of failure or 0 if success
 */
int systick_deregister_cb(systick_callback_t cb) {
    BCFW_VALIDATE_NULL(cb);

    if (cb != systick_cb) {
        return -1;
    }

    systick_cb = NULL;

    return 0;
}
