/******************************************************************************
 * @file     AKD1000.ld
 * @brief    GNU Linker Script for Cortex-M based device
 ******************************************************************************/
/*
 * Copyright (c) 2009-2019 Arm Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 *-------- <<< Use Configuration Wizard in Context Menu >>> -------------------
 */

/*---------------------- DPSRAM ----------------------------------
  <h> Dual port SRAM
    <o0> DPSRAM Base Address is <0x10000000>
    <o1> DPSRAM Size (in Bytes) 512KB
  </h>
  -----------------------------------------------------------------------------*/
__DPSRAM_BASE = 0x10000000;
__DPSRAM_SIZE = 0x00080000;

/*---------------------- ICODE SRAM ----------------------------------
  <h> I-Code SRAM
    <o0> ICODE SRAM Base Address is <0x01000000>
    <o1> ICODE SRAM Size (in Bytes) 128KB
  </h>
  -----------------------------------------------------------------------------*/
__ICODE_SRAM_BASE = 0x01000000;
__ICODE_SRAM_SIZE = 0x00020000;

/*---------------------- DCODE SRAM ----------------------------------
  <h> D-Code SRAM
    <o0> DCODE SRAM Base Address is <0x01100000>
    <o1> DCODE SRAM Size (in Bytes) 128KB
  </h>
  -----------------------------------------------------------------------------*/
__DCODE_SRAM_BASE = 0x01100000;
__DCODE_SRAM_SIZE = 0x00020000;

/*---------------------- LPDDR ----------------------------------
  <h> LPDDR
    <o0> LPDDR Base Address is <0x20000000>
    <o1> LPDDR Size (in Bytes) 3GB
  </h>
  -----------------------------------------------------------------------------*/
__LPDDR_BASE = 0x20000000;
__LPDDR_SIZE = 0xC0000000;

/*---------------------- CODE Configuration ----------------------------------
  <h> CODE Configuration
    <o0> CODE Base Address is DPSAM start <0x10000000>
    <o1> CODE Size (in Bytes) 128K
  </h>
  -----------------------------------------------------------------------------*/
__CODE_BASE = __DPSRAM_BASE;
__CODE_SIZE = 0x00040000;

/*--------------------- Embedded RAM Configuration ----------------------------
  <h> RAM Configuration
    <o0> RAM Base Address is DPSRAM start + 128K
    <o1> RAM Size (in Bytes) 256K
  </h>
 -----------------------------------------------------------------------------*/
__RAM_BASE = (__CODE_BASE + __CODE_SIZE);
__RAM_SIZE = 0x00020000;

/*------------------------- D-SRAM Configuration -------------------------------
  <h> D-SRAM Configuration
    <o0> D-SRAM Base Address is 0x1100000
    <o1> D-SRAM Size (in Bytes) 128K
  </h>
 -----------------------------------------------------------------------------*/
__DSRAM_BASE = 0x01100000;
__DSRAM_SIZE = 0x00020000;

/*------------------------- I-SRAM Configuration -------------------------------
  <h> I-SRAM Configuration
    <o0> I-SRAM Base Address is 0x1000000
    <o1> I-SRAM Size (in Bytes) 128K
  </h>
 -----------------------------------------------------------------------------*/
__ISRAM_BASE = 0x01000000;
__ISRAM_SIZE = 0x00020000;

/*
 *-------------------- <<< end of configuration section >>> -------------------
 */

MEMORY
{
  CODE (rx)  : ORIGIN = __CODE_BASE, LENGTH = __CODE_SIZE
  ISRAM (rx)  : ORIGIN = __ISRAM_BASE, LENGTH = __ISRAM_SIZE
  RAM   (rwx) : ORIGIN = __RAM_BASE, LENGTH = __RAM_SIZE
}

/* Linker script to place sections and symbol values. Should be used together
 * with other linker script that defines memory regions CODE and RAM.
 * It references following symbols, which must be defined in code:
 *   Reset_Handler : Entry of reset handler
 *
 * It defines following symbols, which code can use without definition:
 *   __exidx_start
 *   __exidx_end
 *   __copy_table_start__
 *   __copy_table_end__
 *   __zero_table_start__
 *   __zero_table_end__
 *   __etext
 *   __data_start__
 *   __preinit_array_start
 *   __preinit_array_end
 *   __init_array_start
 *   __init_array_end
 *   __fini_array_start
 *   __fini_array_end
 *   __data_end__
 *   __bss_start__
 *   __bss_end__
 *   __end__
 *   end
 *   __HeapLimit
 *   __StackLimit
 *   __StackTop
 *   __stack
 */
ENTRY(Reset_Handler)

SECTIONS
{
  .text :
  {
    KEEP(*(.vectors))
    *(.text*)

    KEEP(*(.init))
    KEEP(*(.fini))

    /* .ctors */
    *crtbegin.o(.ctors)
    *crtbegin?.o(.ctors)
    *(EXCLUDE_FILE(*crtend?.o *crtend.o) .ctors)
    *(SORT(.ctors.*))
    *(.ctors)

    /* .dtors */
    *crtbegin.o(.dtors)
    *crtbegin?.o(.dtors)
    *(EXCLUDE_FILE(*crtend?.o *crtend.o) .dtors)
    *(SORT(.dtors.*))
    *(.dtors)

    *(.rodata*)

    KEEP(*(.eh_frame*))
  } > CODE

  /*
   * SG veneers:
   * All SG veneers are placed in the special output section .gnu.sgstubs. Its start address
   * must be set, either with the command line option �--section-start� or in a linker script,
   * to indicate where to place these veneers in memory.
   */
/*
  .gnu.sgstubs :
  {
    . = ALIGN(32);
  } > CODE
*/
  .ARM.extab :
  {
    *(.ARM.extab* .gnu.linkonce.armextab.*)
  } > CODE
  
  .cmd.commands :
  {
    . = ALIGN(16);
    __cmd_start = .;
    KEEP(*(SORT(.cmd.*)))
    __cmd_end = .;
  } > CODE

  .test.cmd.commands :
  {
    . = ALIGN(16);
    __cmd_test_start = .;
    KEEP(*(SORT(.test.cmd.*)))
    __cmd_test_end = .;
  } > CODE

  .debug.cmd.commands :
  {
    . = ALIGN(16);
    __cmd_debug_start = .;
    KEEP(*(SORT(.debug.cmd.*)))
    __cmd_debug_end = .;
  } > CODE

  .config.cmd.commands :
  {
    . = ALIGN(16);
    __cmd_config_start = .;
    KEEP(*(SORT(.config.cmd.*)))
    __cmd_config_end = .;
  } > CODE

  .cli.channels :
  {
    . = ALIGN(16);
    __cli_channel_start = .;
    KEEP(*(SORT(.cli.channel.*)))
    __cli_channel_end = .;
  } > CODE

  .drv.drivers :
  {
    . = ALIGN(16);
    __drv_start = .;
    KEEP(*(SORT(.drv.*)))
    __drv_end = .;
  } > CODE

  .log_backends :
  {
    . = ALIGN(16);
    __log_backend_start = .;
    KEEP(*(SORT(.log_backend*)))
    __log_backend_end = .;
  } > CODE

  .log_modules :
  {
      . = ALIGN(16);
      __log_module_start = .;
      KEEP(*(SORT(.log_module*)))
      __log_module_end = .;
  } > CODE

  .properties :
  {
      . = ALIGN(16);
      __prop_start = .;
      KEEP(*(SORT(.prop*)))
      __prop_end = .;
  } > CODE

  __exidx_start = .;
  .ARM.exidx :
  {
    *(.ARM.exidx* .gnu.linkonce.armexidx.*)
  } > CODE
  __exidx_end = .;

  .copy.table :
  {
    . = ALIGN(4);
    __copy_table_start__ = .;
    LONG (__etext)
    LONG (__data_start__)
    LONG ((__data_end__ - __data_start__) / 4)
    /* Add each additional data section here */
    LONG (__eregtext)
    LONG (__isram_start__)
    LONG ((__isram_end__ - __isram_start__) / 4)
/*
    LONG (__etext2)
    LONG (__data2_start__)
    LONG (__data2_end__ - __data2_start__)
*/
    __copy_table_end__ = .;
  } > CODE

  .zero.table :
  {
    . = ALIGN(4);
    __zero_table_start__ = .;
    /* Add each additional bss section here */
    LONG (__bss_start__)
    LONG ((__bss_end__ - __bss_start__)/4)
/*
*/
    __zero_table_end__ = .;
  } > CODE

  /**
   * Location counter can end up 2byte aligned with narrow Thumb code but
   * __etext is assumed by startup code to be the LMA of a section in RAM
   * which must be 4byte aligned 
   */
__eregtext = ALIGN (4);

  .isram : AT(__eregtext)
  {
      __isram_start__ = .;
      *(.isram*)
      __isram_end__ = .;
  } > ISRAM

  . = __eregtext + SIZEOF(.isram);
  __etext = ALIGN (4);

  /**
   * RTT buffer at the start of RAM location
   */
  .rtt (NOLOAD):
  {
    . = ALIGN(8);
  } > RAM
  
  .data : AT (__etext)
  {
    __data_start__ = .;
    *(vtable)
    *(.data)
    *(.data.*)

    . = ALIGN(4);
    /* preinit data */
    PROVIDE_HIDDEN (__preinit_array_start = .);
    KEEP(*(.preinit_array))
    PROVIDE_HIDDEN (__preinit_array_end = .);

    . = ALIGN(4);
    /* init data */
    PROVIDE_HIDDEN (__init_array_start = .);
    KEEP(*(SORT(.init_array.*)))
    KEEP(*(.init_array))
    PROVIDE_HIDDEN (__init_array_end = .);


    . = ALIGN(4);
    /* finit data */
    PROVIDE_HIDDEN (__fini_array_start = .);
    KEEP(*(SORT(.fini_array.*)))
    KEEP(*(.fini_array))
    PROVIDE_HIDDEN (__fini_array_end = .);

    KEEP(*(.jcr*))
    . = ALIGN(4);
    /* All data end */
    __data_end__ = .;

  } > RAM

  /*
   * Secondary data section, optional
   *
   * Remember to add each additional data section
   * to the .copy.table above to asure proper
   * initialization during startup.
   */
/*
  __etext2 = ALIGN (4);

  .data2 : AT (__etext2)
  {
    . = ALIGN(4);
    __data2_start__ = .;
    *(.data2)
    *(.data2.*)
    . = ALIGN(4);
    __data2_end__ = .;

  } > RAM2
*/

  .bss :
  {
    . = ALIGN(4);
    __bss_start__ = .;
    *(.bss)
    *(.bss.*)
    *(COMMON)
    . = ALIGN(4);
    __bss_end__ = .;
  } > RAM AT > RAM

  /*
   * Secondary bss section, optional
   *
   * Remember to add each additional bss section
   * to the .zero.table above to asure proper
   * initialization during startup.
   */
/*
  .bss2 :
  {
    . = ALIGN(4);
    __bss2_start__ = .;
    *(.bss2)
    *(.bss2.*)
    . = ALIGN(4);
    __bss2_end__ = .;
  } > RAM2 AT > RAM2
*/

  .heap (COPY) :
  {
    . = ALIGN(8);
    __end__ = .;
    PROVIDE(end = .);
  } > RAM

  PROVIDE(__HeapLimit = __RAM_BASE + __RAM_SIZE);
  PROVIDE(__StackTop = __DSRAM_BASE + __DSRAM_SIZE);
  PROVIDE(__stack = __StackTop);

  PROVIDE(__dsram_start = __DSRAM_BASE);
  PROVIDE(__dsram_end = __DSRAM_BASE + __DSRAM_SIZE);
  PROVIDE(__isram_start = __ISRAM_BASE);
  PROVIDE(__isram_end = __ISRAM_BASE + __ISRAM_SIZE);

}
