/**
 * @file     AKD1000.h
 * @brief    CMSIS Core Peripheral Access Layer Header File for
 *           AKD1000 SoC
 */
/*
 * Copyright (c) 2021 Brainchip Inc.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef AKD1000_H
#define AKD1000_H

#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------  Interrupt Number Definition  ------------------------ */

typedef enum IRQn {
    /* -------------------  Processor Exceptions Numbers  ----------------------------- */
    NonMaskableInt_IRQn = -14,   /*  2 Non Maskable Interrupt */
    HardFault_IRQn = -13,        /*  3 HardFault Interrupt */
    MemoryManagement_IRQn = -12, /*  4 Memory Management Interrupt */
    BusFault_IRQn = -11,         /*  5 Bus Fault Interrupt */
    UsageFault_IRQn = -10,       /*  6 Usage Fault Interrupt */
    SVCall_IRQn = -5,            /* 11 SV Call Interrupt */
    DebugMonitor_IRQn = -4,      /* 12 Debug Monitor Interrupt */
    PendSV_IRQn = -2,            /* 14 Pend SV Interrupt */
    SysTick_IRQn = -1,           /* 15 System Tick Interrupt */

    /* ToDo: add here your device specific external interrupt numbers
             according the interrupt handlers defined in startup_Device.s
             eg.: Interrupt for Timer#1       TIM1_IRQHandler   ->   TIM1_IRQn */
    DiTimer1_IRQn = 0,            /*!< External Timer-0 Interrupt                       */
    DiTimer2_IRQn = 1,            /*!< External Timer-1 Interrupt                       */
    WatchDog_IRQn = 2,            /*!< WatchDog Interrupt                               */
    UART_IRQ0 = 3,                /*!< UART CHANNEL-0 Interrupt                         */
    UART_IRQ1 = 4,                /*!< UART CHANNEL-1 Interrupt                         */
    HDMAC_IRQ0 = 5,               /*!< DMA CHANNEL-0 Interrupt                          */
    HDMAC_IRQ1 = 6,               /*!< DMA CHANNEL-1 Interrupt                          */
    HDMAC_IRQ2 = 7,               /*!< DMA CHANNEL-2 Interrupt                          */
    HDMAC_IRQ3 = 8,               /*!< DMA CHANNEL-3 Interrupt                          */
    HDMAC_IRQ4 = 9,               /*!< DMA CHANNEL-4 Interrupt                          */
    HDMAC_IRQ5 = 10,              /*!< DMA CHANNEL-5 Interrupt                          */
    HDMAC_IRQ6 = 11,              /*!< DMA CHANNEL-6 Interrupt                          */
    HDMAC_IRQ7 = 12,              /*!< DMA CHANNEL-7 Interrupt                          */
    EXTINT0_AKIDACORE = 13,       /*!< External Interrupt-0                             */
    EXTINT1_RTC = 14,             /*!< External Interrupt-1                             */
    EXTINT2_I2C = 15,             /*!< External Interrupt-2                             */
    EXTINT3_I3C = 16,             /*!< External Interrupt-3                             */
    EXTINT4_I2S = 17,             /*!< External Interrupt-4                             */
    EXTINT5_SPI_RX = 18,          /*!< External Interrupt-5                             */
    EXTINT6_SPI_TX = 19,          /*!< External Interrupt-6                             */
    EXTINT7_SPI_ERR = 20,         /*!< External Interrupt-7                             */
    EXTINT8_TRNG = 21,            /*!< External Interrupt-8                             */
    EXTINT9_LPDDR = 22,           /*!< External Interrupt-9                             */
    EXTINT10_PCIEP_OR_USB = 23,   /*!< External Interrupt-10                            */
    EXTINT11_PCI_RC = 24,         /*!< External Interrupt-11                            */
    EXTINT12 = 25,                /*!< External Interrupt-12                            */
    EXTINT13_SYS_REG = 26,        /*!< External Interrupt-13                            */
    EXTINT14_LPDDR_FRQ_CHNG = 27, /*!< External Interrupt-14                            */
    EXTINT15_GPIO = 28,           /*!< External Interrupt-15                            */
                                  /* Interrupts 29 .. 224 are left out */
} IRQn_Type;

/* ================================================================================ */
/* ================      Processor and Core Peripheral Section     ================ */
/* ================================================================================ */

#define AHB_BASE 0xE8000000

// APB BASE
#define APB_BASE 0xE8010000

#define MRBC_OFFSET 0x0000
#define WDOG_OFFSET 0x2000
#define EXIU_OFFSET 0x3000
#define TIMER_OFFSET 0x5000
#define UART0_OFFSET 0x8000
#define UART1_OFFSET 0x9000
#define GPIO_OFFSET 0xC000

#define RTC_OFFSET 0x07FF8000 /*RTC is connected to APB, But the offset range is too high*/
#define TRNG_OFFSET 0x07FF9000
#define I2C_OFFSET 0x07FFA000
#define I3C_OFFSET 0x07FFB000
#define I2S_OFFSET 0x07FFC000
#define HDMAC_OFFSET 0x00001000

#define HDMAC_BASE (AHB_BASE + HDMAC_OFFSET)
#define MRBC_BASE (APB_BASE + MRBC_OFFSET)
#define WDOG_BASE (APB_BASE + WDOG_OFFSET)
#define EXIU_BASE (APB_BASE + EXIU_OFFSET)
#define TIMER_BASE (APB_BASE + TIMER_OFFSET)
#define UART0_BASE (APB_BASE + UART0_OFFSET)
#define UART1_BASE (APB_BASE + UART1_OFFSET)
#define GPIO_BASE (APB_BASE + GPIO_OFFSET)

/*RTC base address*/
#define RTC_BASE (APB_BASE + RTC_OFFSET)
#define TRNG_BASE (APB_BASE + TRNG_OFFSET)
#define I2C_BASE (APB_BASE + I2C_OFFSET)
#define I3C_BASE (APB_BASE + I3C_OFFSET)
#define I2S_BASE (APB_BASE + I2S_OFFSET)

#define SPI_BASE (0xF0000000UL + 0x03FF0000UL)
#define CRG11_BASE (0xf0001000)

/* -------  Start of section using anonymous unions and disabling warnings  ------- */
#if defined(__CC_ARM)
#pragma push
#pragma anon_unions
#elif defined(__ICCARM__)
#pragma language = extended
#elif defined(__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wc11-extensions"
#pragma clang diagnostic ignored "-Wreserved-id-macro"
#elif defined(__GNUC__)
/* anonymous unions are enabled by default */
#elif defined(__TMS470__)
/* anonymous unions are enabled by default */
#elif defined(__TASKING__)
#pragma warning 586
#elif defined(__CSMC__)
/* anonymous unions are enabled by default */
#else
#warning Not supported compiler type
#endif

/* --------  Configuration of Core Peripherals  ----------------------------------- */
#define __CM4_REV 0x0001U         /* Core revision r0p1 */
#define __MPU_PRESENT 1U          /* MPU present */
#define __VTOR_PRESENT 1U         /* VTOR present */
#define __NVIC_PRIO_BITS 3U       /* Number of Bits used for Priority Levels */
#define __Vendor_SysTickConfig 0U /* Set to 1 if different SysTick Config is used */
#define __FPU_PRESENT 1U          /* FPU present */

#include "core_cm4.h"       /* Processor and core peripherals */
#include "system_AKD1000.h" /* System Header */

/* --------  End of section using anonymous unions and disabling warnings  -------- */
#if defined(__CC_ARM)
#pragma pop
#elif defined(__ICCARM__)
/* leave anonymous unions enabled */
#elif (defined(__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050))
#pragma clang diagnostic pop
#elif defined(__GNUC__)
/* anonymous unions are enabled by default */
#elif defined(__TMS470__)
/* anonymous unions are enabled by default */
#elif defined(__TASKING__)
#pragma warning restore
#elif defined(__CSMC__)
/* anonymous unions are enabled by default */
#else
#warning Not supported compiler type
#endif

#ifdef __cplusplus
}
#endif

#endif /* AKD1000_H */
