/**
 * @file     exiu.h
 * @brief    The APIs declared in exiu.h are private APIs which are called by interrupt public APIs
 *           declared in interrupt.h
 */
#ifndef __EXIU_H_
#define __EXIU_H_

#include <stdint.h>

/**
 * Interrupt Detection Methods
 */
typedef enum {
    INTR_LOWLEVEL = 0, /*LOW LEVEL SETTING*/
    INTR_HIGHLEVEL,    /*HIGH LEVEL SETTING*/
    INTR_FALLINGEDGE,  /*FALLING EDGE SETTING*/
    INTR_RISINGEDGE    /*RISING EDGE SETTING*/
} intr_detection_method_t;

/**
 * exiu_interrupt_conf API is used to configure EXIU for a specific interrupt source.
 *
 * @param intr : specific interrupt source connected to EXIU.
 * @param detection_method: Detection Method for each interrupt (level/edge-sensitive etc.)
 *
 * @return 0 if success or error code in case of failure.
 */
int exiu_interrupt_conf(const uint8_t intr, intr_detection_method_t detection_method);

/**
 * exiu_interrupt_deconf API is used to de-configure EXIU for a specific interrupt source.
 *
 * @param intr : specific interrupt source connected to EXIU.
 *
 * @return 0 if success or error code in case of failure.
 */
int exiu_interrupt_deconf(const uint8_t intr);

/**
 * exiu_interrupt_clear API is used to clear the EXIU interrupt status for a specific interrupt
 * source.
 *
 * @param intr : specific interrupt source connected to EXIU.
 *
 * @return 0 if success or error code in case of failure.
 */
int exiu_interrupt_clear(const uint8_t intr);

/**
 * exiu_interrupt_enable API is used to enable the EXIU interrupt
 * from the given source.
 *
 * @param intr : specific interrupt source connected to EXIU.
 *
 * @return 0 if success or error code in case of failure.
 */
int exiu_interrupt_enable(const uint8_t intr);

/**
 * exiu_interrupt_disable API is used to disable the EXIU interrupt
 * from the given source.
 *
 * @param intr : specific interrupt source connected to EXIU.
 *
 * @return 0 if success or error code in case of failure.
 */
int exiu_interrupt_disable(const uint8_t intr);
#endif /*__EXIU_H_*/
