#ifndef __AKIDA_DMA_H__
#define __AKIDA_DMA_H__
/**
 * @file akida_dma.h
 * This is public header file of HDMAC firmware module.
 */

/** @addtogroup drivers
 * @{
 */

/**
 * @defgroup dma dma
 * DMA Management APIs
 */

/** @addtogroup dma
 *
 * This interface contains methods and definitions allowing to
 * configure and control AKIDA dma transfers
 *
 *  @{
 */
#include <stdbool.h>
#include <stdint.h>

/** An enum type to identify
 *  Peripheral that requests DMA
 */
typedef enum {
    HDMAC_PERIPHERAL_ID_UART_0_RX, /**< UART-0 RX*/
    HDMAC_PERIPHERAL_ID_UART_0_TX, /**< UART-0 TX*/
    HDMAC_PERIPHERAL_ID_UART_1_RX, /**< UART-1 RX*/
    HDMAC_PERIPHERAL_ID_UART_1_TX, /**< UART-1 TX*/
    HDMAC_PERIPHERAL_ID_I2S,       /**< I2S slave */
    HDMAC_PERIPHERAL_ID_SPI_RX,    /**< SPI slave for RX */
    HDMAC_PERIPHERAL_ID_SPI_TX,    /**< SPI slave for TX */
    HDMAC_PERIPHERAL_ID_SW,        /**< DMA request by software */
    HDMAC_PERIPHERAL_ID_MAX,       /**< Invalid peripheral */
} hdmac_peripheral_id_t;

/** An enum type to identify
 *  DMA Transfer modes in HDMAC
 */
typedef enum {
    HDMAC_TRANSFER_MODE_BLOCK,  /**< Block transfer mode selection   */
    HDMAC_TRANSFER_MODE_BURST,  /**< Burst transfer mode selection   */
    HDMAC_TRANSFER_MODE_DEMAND, /**< Demand transfer mode selection  */
    HDMAC_TRANSFER_MODE_NONE,   /**< Invalid transfer mode selection */
} hdmac_transfer_mode_t;

/** An enum type to identify
 *  DMA Beat Transfer Type in HDMAC
 */
typedef enum {
    HDMAC_BEAT_TRANSFER_SINGLE = 0x8, /**< Normal DMA transfer */
    HDMAC_BEAT_TRANSFER_INCR,         /**< Increment beat transfer of unkown length */
    HDMAC_BEAT_TRANSFER_WRAP4,        /**< Wrap beat transfer of 4 beats  */
    HDMAC_BEAT_TRANSFER_INCR4,        /**< Increment beat transfer of 4 beats       */
    HDMAC_BEAT_TRANSFER_WRAP8,        /**< Wrap beat transfer of 8 beats  */
    HDMAC_BEAT_TRANSFER_INCR8,        /**< Increment beat transfer of 8 beats       */
    HDMAC_BEAT_TRANSFER_WRAP16,       /**< Wrap beat transfer of 16 beats */
    HDMAC_BEAT_TRANSFER_INCR16,       /**< Increment beat transfer of 16 beats      */
    HDMAC_BEAT_TRANSFER_NONE,         /**< Invalid beat transfer */
} hdmac_beat_transfer_t;

/** An enum type to identify
 *  DMA stop status per channel
 */
typedef enum {
    HDMAC_STATUS_IDLE,          /**< Bitfield reset value,also to clear interrupts */
    HDMAC_STATUS_ADDR_OVERFLOW, /**< DMA transfer stopped due to Address overflow */
    HDMAC_STATUS_STOP_REQ,      /**< DMA transfer stopped due to stop request     */
    HDMAC_STATUS_SRC_ERROR,     /**< DMA transfer stopped due to src access error */
    HDMAC_STATUS_DEST_ERROR,    /**< DMA transfer stopped due to dest access error*/
    HDMAC_STATUS_NORMAL_END,    /**< DMA transfer stopped after completion        */
    HDMAC_STATUS_DMA_PAUSE = 7, /**< DMA transfer stopped due to pause request    */
    HDMAC_STATUS_NONE,          /**< Invalid option                               */
} hdmac_stop_status_t;

/** An enum type to identify
 *  DMA operation type
 */
typedef enum {
    HDMAC_OPERATION_UNPAUSE, /**< Un-pause a DMA transfer */
    HDMAC_OPERATION_DISABLE, /**< Stop DMA transfer on a channel  */
    HDMAC_OPERATION_PAUSE,   /**< Pause DMA transfer on a channel */
    HDMAC_OPERATION_NONE,    /**< Invalid option */
} hdmac_operation_t;

/** Function pointer of a
 *  callback that is subscribed by user
 *
 * @param[in] p_id Peripheral ID.
 * @param[in] err_code Error code returned by DMA.
 * @param[in] param place holder for information passed to user.
 */
typedef void (*dma_callback_t)(hdmac_peripheral_id_t p_id, uint8_t err_code, void* param);

/** A structure to abstract
 *  DMA channel configuration
 */
typedef struct {
    hdmac_transfer_mode_t transfer_mode; /**< Transfer mode selection  */
    uint8_t transfer_width;              /**< Transfer width, supported are 8,16,32bits          */
    hdmac_beat_transfer_t beat_transfer; /**< Beat transfer type       */
    uint8_t block_count; /**< range =<0x1-0x10>,Size of block= @a transferWidth X  @a blockCount*/
    uint32_t transfer_count; /**< range =<0x1-0x10000>,Number of blocks= @a transferCount X @a
                               blockCount*/
    uint32_t src_address;    /**< Source address           */
    uint32_t dest_address;   /**< Destination address      */
    bool is_src_addr_fixed;  /**< To have fixed source address      */
    bool is_dest_addr_fixed; /**< To have fixed destination address */
    bool single_req;         /**< Flag represents single or burst request, used in case of UART */
} hdmac_request_conf_t;

/** A structure to abstract
 *  DMA transfer status on a channel
 */
typedef struct {
    uint8_t block_count;             /**< Current block count     */
    uint16_t transfer_count;         /**< Current transfer count  */
    bool enable;                     /**< Current enable status of channel */
    bool sw_trigger;                 /**< Current status of software trigger */
    hdmac_stop_status_t stop_status; /**< Status of channel, read from HW */
} hdmac_status_t;

/**
 * API to Initialize DMA FW.
 * @return     SUCCESS   Successfully initialized DMA FW module.
 * @return     !=SUCCESS Error in initializing DMA FW module.
 */
int akida_hdmac_init(void);

/**
 * API to get default DMA request configuration.
 * @param[in,out] req_conf DMA configuration.
 * @return     SUCCESS   Successfully retreived default configuration
 * @return     !=SUCCESS Error in retreiving default configuration.
 */
int akida_hdmac_default_request_conf_get(hdmac_request_conf_t* req_conf);

/**
 * API to set DMA request configuration.
 * API is synchronous if cb = NULL else its asynchronous.
 *
 * @param[in] per_id Identifies peripheral that uses DMA.
 * @param[in] req_conf DMA request configuration.
 * @param[in] cb Callback provided by DMA user.
 * @param[in] cb_param parameters to callback.
 * @see hdmac_default_request_conf_get().
 * @return     SUCCESS   Successfully configured DMA request and trigger the same on a channel
 * mapped to @a per_id.
 * @return     !=SUCCESS Error in Configuring request configuration.
 */
int akida_hdmac_request_transfer(const hdmac_peripheral_id_t per_id, hdmac_request_conf_t* req_conf,
                                 dma_callback_t cb, void* cb_param);

/**
 * API to control various operations on a DMA transfer.
 * @param[in]  per_id  Identifies Peripheral that uses DMA.
 * @param[in]  dma_operation   DMA operation on a channel that is assigned to a peripheral.
 * @return     SUCCESS   Successfully performed given dma_operation on a channel mapped to @a
 * per_id.
 * @return     !=SUCCESS Error in setting DMA operation.
 */
int akida_hdmac_control_operation(hdmac_peripheral_id_t per_id, hdmac_operation_t dma_operation);

/**
 * API to get DMA transfer status on a channel.
 * @param[in]  per_id  Identifies Peripheral that uses DMA.
 * @param[out] dma_status  DMA transfer status that is assigned to peripheral.
 * @return     SUCCESS   Successfully retreived status on a peripheral
 * @return     !=SUCCESS Error in retreiving status.
 */
int akida_hdmac_get_transfer_status(hdmac_peripheral_id_t per_id, hdmac_status_t* dma_status);

/** @} dma
 * @} drivers
 */
#endif /* __AKIDA_DMA_H__ */
