#ifndef __CLOCK_H__
#define __CLOCK_H__

#include <stdint.h>

/** @addtogroup drivers
 * @{
 */

/**
 * @defgroup clock clock
 * Clock management APIs
 *
 * @addtogroup clock
 *
 * These apis allows to control the specific SoC clock tree.
 *
 * Each driver references the clock it uses from the clock tree in
 * order to be able to react to clock tree changes, or to optimize
 * power consumption by gating clock of unused peripherals
 *
 * @{
 */

/**
 * Get clock frequency
 *
 * @param clk_id the clock identifier to request frequency.
 *               if clk_id is -1, returns the core clock frequency.
 * @return the clock for the given clock id.
 */
uint32_t clock_get_freq(int clk_id);

/**
 * Set clock divider
 *
 * Set the clock divider for a given clock id. That will make the
 * clock frequency to be core clock / div
 *
 * @param clk_id the clock to change divider
 * @param div the new divider to set
 *
 * @return 0 if success
 */
int clock_set_div(int clk_id, int div);

/**
 * Acquire clock
 *
 * Notify clock manager that this clock is used. This method increments the
 * usage count for this clock, and will start the clock if it was not enabled.
 *
 * @param clk_id the clock to acquire
 *
 * @return 0 if success
 */
int clock_acquire(int clk_id);

/**
 * Release clock
 *
 * Notify clock manager that this clock is not used. That will decrement
 * this clock usage count and may stop the clock if counter reaches 0.
 *
 * @param clk_id the clock to release
 *
 * @return 0 if success
 */
int clock_release(int clk_id);

/** @} clock
 * @} drivers
 */
#endif /* __CLOCK_H__ */