#ifndef __DRIVER_H_
#define __DRIVER_H_

#include "clock.h"
struct config_item {
    char* name;
    int default_value;
    int current_value;
    int new_value;
};

/**
 * @defgroup drivers drivers
 *
 * bsp drivers
 */

/**
 * @addtogroup drivers
 * @{
 *
 * @defgroup sensors sensors
 * The class of drivers that interface with sensors
 *
 * @defgroup driver_model driver_model
 * The driver abstract APIs
 *
 * @addtogroup driver_model
 * @{
 */

/**
 * Get driver config item pointer
 *
 * @param drv the driver to access config items of
 * @param index the index of the config item to access
 *
 * @return the config item pointer requested
 */
#define DRIVER_CONFIG_GET_ITEM(drv, index) (((struct driver_common*)drv)->config_items[index])

/**
 * Set config item current value
 *
 * This macro replaces the existing config item current value with the new value
 *
 * @param drv the driver to access config items
 * @param index the config item of the driver to modify
 * @param value the new current value
 */
#define DRIVER_CONFIG_SET_CURRENT(drv, index, value) \
    (((struct driver_common*)drv)->config_items[index].current_value = value)

/**
 * Set config item new value
 *
 * This macro replaces the existing config item new value with the new value
 *
 * @param drv the driver to access config items
 * @param index the config item of the driver to modify
 * @param value the new value
 */
#define DRIVER_CONFIG_SET_NEW(drv, index, value) \
    (((struct driver_common*)drv)->config_items[index].new_value = value)

/**
 * Set config item current value with config item new value
 *
 * This macro replaces the existing config item current value with the config item new value
 *
 * @param drv the driver to access config items
 * @param index the config item of the driver to modify
 */
#define DRIVER_CONFIG_APPLY(drv, index)                                \
    (((struct driver_common*)drv)->config_items[index].current_value = \
         (((struct driver_common*)drv)->config_items[index].new_value))

/**
 * Common structure for driver definition.
 */
struct driver_common {
    /**
     * Initialize a driver
     *
     * This method is called at boot for all drivers
     * @param drv the driver to initialize
     * @param priv the data specific to the driver
     *
     * @return int error code or 0 if success
     */
    int (*init)(struct driver_common* drv, void* priv);

    /**
     * Deinitialize a driver
     *
     * This method can be called in shutdown in order to
     * put hardware in a safe power-down condition.
     *
     * @param drv the driver to deinitialize
     * @param priv the data specific to the driver
     *
     * @return int error code or 0 if success
     */
    int (*deinit)(struct driver_common* drv, void* priv);

    /**
     * Read configuration
     *
     * This method fills current value for all configuration
     * items defined for this driver
     *
     * @param drv the driver to deinitialize
     *
     * @return int error code or 0 if success
     */
    int (*read_config)(struct driver_common* drv);

    /**
     * Apply configuration
     *
     * This method applies the new values set in the configuration
     * elements and make them new values.
     *
     * @param drv the driver to deinitialize
     *
     * @return int error code or 0 if success
     */
    int (*apply_config)(struct driver_common* drv);

    struct config_item* config_items;
    int config_items_count;

    /**
     * Pointer to the driver's registry item
     */
    struct driver_item* reg;
};

/** A device driver have to be initialized at startup */
#define DRIVER_TYPE_DEVICE 1
/** An interface type driver will be initialized when
 * attached to a bus */
#define DRIVER_TYPE_INTERFACE 2

/**
 * Structure holding a driver item in the registry.
 */
struct driver_item {
    /** Name of the driver or interface */
    char* name;
    /** Class name of the driver or interface
     * This field allows to lookup driver by roles
     * for instance, all i2c bus will be of class "i2c",
     * all power monitoring will be of class "pmon"
     */
    char* class;
    /** The actual driver instance pointer */
    struct driver_common* driver;
    /** The type of driver (interface or driver instance) */
    int type;
    /** The clock used by this driver */
    int clk_id;
    /** The private data passed to init in case of driver instance */
    void* priv;
};

/**
 * Retrieves the clock id for the given driver
 *
 * @param "_driver_" the driver to retrieve clock id from
 *
 * @returns the clock id
 */
#define DRIVER_GET_CLK_ID(_driver_) (((_driver_)->reg)->clk_id)

/**
 * Declare a driver with a clock id
 *
 * @param "_name_" the name of the driver
 * @param "_class_" the class name of the driver
 * @param "_driver_" the driver structure ( @ref driver_common ) of the driver
 * @param "_priv_" the priv pointer passed to the driver_init method
 * @param "_clock_" the clock id for this driver
 */
#define DECLARE_DRIVER_CLK(_name_, _class_, _driver_, _priv_, _clock_) \
    _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, DRIVER_TYPE_DEVICE, _clock_)

/**
 * Declare a driver
 *
 * @param "_name_" the name of the driver
 * @param "_class_" the class name of the driver
 * @param "_driver_" the driver structure (ref driver_common_ ) of the driver
 * @param "_priv_" the priv pointer passed to the driver_init method
 */
#define DECLARE_DRIVER(_name_, _class_, _driver_, _priv_) \
    _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, DRIVER_TYPE_DEVICE, -1)

/**
 * Declare an interface driver
 *
 * An interface driver is a driver that must be attached to a bus for it to function.
 * It allows to have more than one instance of a driver in the system with the same interface.
 *
 * @param "_name_" the name of the driver
 * @param "_class_" the class name of the driver
 * @param "_driver_" the driver structure (ref driver_common_ ) of the driver
 * @param "_priv_" the priv pointer passed to the driver_init method
 */
#define DECLARE_INTERFACE(_name_, _class_, _driver_, _priv_) \
    _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, DRIVER_TYPE_INTERFACE, -1)

/**
 * Declare a driver in the registry.
 */
#if defined(__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050)
#define _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, _type_, _clock_)  \
    struct driver_item __attribute__((section("DRV"))) __attribute__((used)) \
        __driver_##_name_ = {.name = #_name_,                                \
                             .class = #_class_,                              \
                             .driver = _driver_,                             \
                             .type = _type_,                                 \
                             .clk_id = _clock_,                              \
                             .priv = _priv_};

#elif defined(__GNUC__)
#define _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, _type_, _clock_)                     \
    struct driver_item __attribute__((section(".drv." #_class_ #_name_))) __attribute__((used)) \
        __driver_##_name_ = {.name = #_name_,                                                   \
                             .class = #_class_,                                                 \
                             .driver = _driver_,                                                \
                             .type = _type_,                                                    \
                             .clk_id = _clock_,                                                 \
                             .priv = _priv_};

#endif
/**
 * This function should be called by the runtime before any driver is used.
 * It will iterate over all drivers and call their init function.
 *
 * @return 0 if init succesfull, error code if an error occured
 */
int driver_init(void);

/**
 * Search for a driver given its name.
 *
 * @param name the name of the driver to look for
 *
 * @return struct driver_common * the pointer to the driver handler NULL if driver not found.
 */
struct driver_common* driver_lookup(char* name);

/**
 * Search for an interface given its name.
 *
 * @param name the name of the interface to look for
 *
 * @return struct driver_common * the pointer to the interface operations.
 */
struct driver_common* interface_lookup(char* name);

/**
 * This function should be called by the runtime before any driver is used.
 * It will iterate over all drivers and call their init function.
 *
 * @return 0 if init successful, error code if an error occurred
 */
int driver_set_config(struct driver_common* drv, char* name, int value);

/**
 * This function should be called by the runtime before any driver is used.
 * It will iterate over all drivers and call their init function.
 *
 * @return 0 if init successful, error code if an error occurred
 */
int driver_apply_config(struct driver_common* drv);

/**
 * This function dynamically add a driver to the list of available drivers.
 *
 * @param drv the driver to add
 * @return 0 if successfull, error code in case of error
 */
int driver_add_dynamic(struct driver_common* drv);

/**
 * This function iterates over all drivers
 *
 * For each driver the callback function will be called with the driver
 * and the private pointer as parameters.
 * The callback can return non-zero in order to stop iterating.
 *
 * @param cb the callback called for each available driver
 * @param param the parameter passed to the callback in addition to the current driver
 */
void driver_foreach(int (*cb)(struct driver_item* drv, void* param), void* param);

/**
 * @} driver_model
 * @} drivers
 */
#endif
