#ifndef __GPIO_H_
#define __GPIO_H_

#include <stdint.h>
#include "driver.h"

#define GPIO_DIR_IN 0
#define GPIO_DIR_OUT 1

/**
 * @file gpio.h
 * This is public header file of gpio.
 **/

/** @addtogroup drivers
 * @{
 */

/** @addtogroup gpio gpio
 *  Documentation details of GPIO module.
 *  @{
 **/

/**
 * GPIO interrupt trigger modes
 */
typedef enum {
    GPIO_INTR_LOWLVL = 0,   /**< Low level GPIO interrupt trigger mode */
    GPIO_INTR_HIGHLVL,      /**< High level GPIO interrupt trigger mode */
    GPIO_INTR_RISING_EDGE,  /**< Rising edge GPIO interrupt trigger mode */
    GPIO_INTR_FALLING_EDGE, /**< Falling edge GPIO interrupt trigger mode */
    GPIO_INTR_BOTH_EDGE     /**< Both edge GPIO interrupt trigger mode */
} gpio_intr_mode_t;

/**
 * GPIO resistor modes
 */
typedef enum {
    GPIO_PULL_NO_PULL = 0, /**< PULL_NO_PULL resistor mode */
    GPIO_PULL_PU,          /**< PULL_UP resistor mode */
    GPIO_PULL_PD,          /**< PULL_DOWN resistor mode */
} gpio_mode_t;

/**
 * Structure that describes the current soc gpio capabilities.
 */

typedef struct {
    uint8_t num_ports;  /**<number of GPIO ports on the current soc */
    uint8_t port_width; /**<number of GPIO's per port */
} gpio_cap_t;

/**
 * Structure that describes the gpio flags and its bit fields.
 */

typedef union {
    struct {
        /**GPIO mode, refer gpio_mode_t */
        gpio_mode_t gpio_mode : 3;
        /**GPIO- Input  interrupt trigger mode refer gpio_intr_mode_t */

        /*                _______ <-----High Level
         RisingEdge->    /       \  <---Falling Edge
         LowLevel-->____/         \______
         Interrupt mode bothEdge -> interrupt on rising & falling edges.
        */
        gpio_intr_mode_t intr_mode : 3;
    } flags;
    uint32_t flagval;
} gpio_config_flag_t;

struct gpio_driver;

/**
 * gpio callback function to notify to the registrar, invokes upon gpio interrupt trigger
 *
 * @param drv the gpio driver in use
 * @param val GPIO bit value read from the pin
 * @param cb_param information to be passed to application
 */

typedef void (*gpio_callback)(struct gpio_driver* drv, uint8_t val, void* cb_param);

struct gpio_ops {
    /**
     * Get the current soc gpio capabilities, number of ports and gpio's per port
     *
     * @param drv the driver to use
     * @param caps capability structure updated with current soc gpio capabilities.
     * @returns 0 in case of success else return error.
     */
    int (*get_capabilities)(struct gpio_driver* drv, gpio_cap_t* caps);

    /**
     * Configure the requested port_id and corresponding bit GPIO.
     *
     * @param  drv the driver to use
     * @param  port_id GPIO Port ID
     * @param  bit Specific GPIO on the given port
     * @param  dir GPIO direction 0-input, 1-output
     * @param  flags GPIO configuration flags
     * @param  cb register a callback pointer to notify when the GPIO input status is
     *                  changed. Callback pointer is ignored for the output directed GPIO.
     * @param cb_param the parameter to pass to the callback
     * @returns 0 in case of success else return error.
     */
    int (*set_config)(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t dir,
                      gpio_config_flag_t flags, gpio_callback cb, void* cb_param);

    /**
     * Read the requested port and corresponding bit GPIO value.
     *
     * @param  drv the driver to use
     * @param  port_id GPIO Port ID
     * @param  bit corresponding GPIO bit on the given port_id
     * @param  rval pointer to store the read value from the port_id port and bit GPIO
     * @returns 0 in case of success else return error
     */
    int (*read)(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t* rval);

    /**
     *  Read the requested GPIO port all the bit values.
     *
     * @param  drv the driver to use
     * @param  port_id GPIO Port ID
     * @param  rval pointer to store the read value from the port_id port
     * @returns 0 in case of success else return error
     */
    int (*read_port)(struct gpio_driver* drv, uint8_t port_id, uint32_t* rval);

    /**
     * Write to the requested port and corresponding bit GPIO with the given value
     *
     * @param  drv the driver to use
     * @param  port_id GPIO Port ID
     * @param  bit corresponding GPIO bit on the given port_id
     * @param  value  value to be written on the GPIO.
     * @returns 0 in case of success else return error
     */
    int (*write)(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t value);

    /**
     * Write the given value to the specified GPIO port.
     *
     * @param  drv the driver to use
     * @param  port_id GPIO Port ID
     * @param  value  value to be written on the given GPIO port_id .
     * @returns 0 in case of success else return error
     */
    int (*write_port)(struct gpio_driver* drv, uint8_t port_id, uint32_t value);
};

/**
 * GPIO interface driver handle.
 */
struct gpio_driver {
    /** Common driver interface */
    struct driver_common driver;
    /** GPIO driver interface implementation*/
    struct gpio_ops* ops;
    /** Private pointer */
    void* priv;
};

#define DECLARE_GPIO(_name_, _ops_, _init_, _priv_)                                  \
    struct gpio_driver __gpio_data_##_name_ = {.driver.init = _init_, .ops = _ops_}; \
    DECLARE_DRIVER(_name_, gpio, &__gpio_data_##_name_.driver, _priv_)

/**
 * Get the current soc gpio capabilities, number of ports and gpio's per port
 *
 * @param drv the driver to use
 * @param caps capability structure updated with current soc gpio capabilities.
 * @returns 0 in case of success else return error
 */
int gpio_get_capabilities(struct gpio_driver* drv, gpio_cap_t* caps);

/**
 * Configure the requested port_id and corresponding bit GPIO.
 *
 * @param  drv the driver to use
 * @param  port_id GPIO Port ID
 * @param  bit Specific GPIO on the given port
 * @param  dir GPIO direction 0-input, 1-output
 * @param  flags GPIO configuration flags pushpull, opendrain etc.,
 * @param  cb  register a callback pointer to notify when the GPIO input status is changes.
 *               this pointer is void for the output directed GPIO.
 * @param cb_data the parameter to pass to the callback
 * @returns 0 in case of success else return error
 */
int gpio_set_config(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t dir,
                    gpio_config_flag_t flags, gpio_callback cb, void* cb_data);

/**
 * Read the requested port and corresponding bit GPIO value.
 *
 * @param  drv the driver to use
 * @param  port_id GPIO Port ID
 * @param  bit corresponding GPIO bit on the given port_id
 * @param  rval pointer to store the read value from the port_id port and bit GPIO
 * @returns 0 in case of success else return error
 */
int gpio_read(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t* rval);

/**
 * Read the requested GPIO port all the bit values.
 *
 * @param  drv the driver to use
 * @param  port_id GPIO Port ID
 * @param  rval pointer to store the read value from the port_id port
 * @returns 0 in case of success else return error.
 */
int gpio_read_port(struct gpio_driver* drv, uint8_t port_id, uint32_t* rval);

/**
 * Write to the requested port and corresponding bit GPIO with the given value
 *
 * @param  drv the driver to use
 * @param  port_id GPIO Port ID
 * @param  bit corresponding GPIO bit on the given port_id
 * @param  value  value to be written on the GPIO.
 * @returns 0 in case of success else return error
 */
int gpio_write(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t value);

/**
 * Write the given value to the specified GPIO port.
 *
 * @param  drv the driver to use
 * @param  port_id GPIO Port ID
 * @param  value  value to be written on the given GPIO port_id
 * @returns 0 in case of success else return error
 */
int gpio_write_port(struct gpio_driver* drv, uint8_t port_id, uint32_t value);
/** @} gpio
 * @} drivers
 */
#endif  //__GPIO_H_
