#ifndef __I2C_BUS_H_
#define __I2C_BUS_H_

/**
 * @file i2c_bus.h
 * This is public header file of i2c bus.
 */

/** @addtogroup drivers
 * @{
 */

/** @addtogroup i2c_bus i2c_bus
 *  Documentation details of i2c bus Abstraction layer.
 *  @{
 */

#include <stdint.h>

#include "driver.h"
#include "elog.h"

struct i2c_bus_driver;

typedef void (*i2c_callback_t)(struct i2c_bus_driver* drv, void* param);

/* forward declaration of i2c_device */
struct i2c_device;

struct i2c_bus_ops {
    /**
     * initiate a read from an i2c device
     *
     * @param drv the i2c driver in use
     * @param buf the buffer to save read data
     * @param len the number of bytes to read
     * @param cb the callback function to call on completion, NULL to use sync mode
     * @param cb_param the parameter to pass to the callback
     *
     * @return 0 if async request started, number of read bytes if synchronous
     */
    int (*read)(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb, void* cb_param);

    /**
     * initiate a write to an i2c device
     *
     * @param drv the i2c driver in use
     * @param buf the buffer to write data from
     * @param len the number of bytes to write
     * @param cb the callback function to call on completion, NULL to use sync mode
     * @param cb_param the parameter to pass to the callback
     *
     * @return 0 if async request started, number of written bytes if synchronous
     */
    int (*write)(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb, void* cb_param);

    /**
     * read register from an i2c device
     *
     * This function is always synchronous
     *
     * @param drv the i2c driver in use
     * @param reg the register address to read from
     * @param buf the buffer to save read data
     * @param len the number of bytes to read
     *
     * @return 0 if async request started, number of read bytes if synchronous
     */
    int (*read_reg)(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len);

    /**
     * write a register of an i2c device
     *
     * This function is always synchronous
     *
     * @param drv the i2c driver in use
     * @param reg the register address to write to
     * @param buf the buffer containing the data to write
     * @param len the number of bytes to write
     *
     * @return 0 if async request started, number of written bytes if synchronous
     */
    int (*write_reg)(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len);

    /**
     * attach a device to a bus
     *
     * This function is called when a device is attached to a bus
     *
     * @param drv the i2c driver in use
     * @param dev the device attached to the bus
     *
     * @return 0 if attach succedded, or error code in case of error
     */
    int (*attach)(struct i2c_bus_driver* drv, struct i2c_device* dev);

    /**
     * detach a device from the bus if it's attached to
     *
     * This function is called when a device needs to be detached from the bus
     *
     * @param drv the i2c driver in use
     * @param dev the i2c device to detach
     *
     * @return 0 if detach succedded, or error code in case of error
     */
    int (*detach)(struct i2c_bus_driver* drv, struct i2c_device* dev);
};

/**
 * I2C bus driver handle.
 */
struct i2c_bus_driver {
    /** Common driver interface */
    struct driver_common driver;
    /** I2C driver interface implementation*/
    struct i2c_bus_ops* ops;
    /** List of attached I2C devices **/
    struct i2c_device* devices;
    /** Private pointer */
    void* priv;
};

/** declare an I2C bus driver instance */
#define DECLARE_I2C_BUS(_name_, _ops_, _init_, _priv_, _clk_)  \
    struct i2c_bus_driver __i2c_data_##_name_ = {              \
        .driver.init = _init_, .ops = _ops_, .devices = NULL}; \
    DECLARE_DRIVER_CLK(_name_, i2c, &__i2c_data_##_name_.driver, _priv_, _clk_)

/**
 * Declare an Config items
 * to store copy of config items from I2C device interface
 */
#define DECLARE_CONFIG_ITEMS(_name_, _no_of_items_) \
    static struct config_item _name_##_config_items[_no_of_items_];

/**
 * Declare an I2C device driver instance
 * the I2C device driver must be attached to a bus instance using
 * i2c_attach()
 */

#define DECLARE_I2C_DEVICE(_name_, _address_, _priv_, _no_of_items_)                 \
    DECLARE_CONFIG_ITEMS(_name_, _no_of_items_)                                      \
    static struct i2c_device _name_ = {.name = #_name_,                              \
                                       .address = _address_,                         \
                                       .priv = _priv_,                               \
                                       .driver.config_items = _name_##_config_items, \
                                       .driver.config_items_count = _no_of_items_};
/**
 * Declare an I2C device driver instance for non config items
 * the I2C device driver must be attached to a bus instance using
 * i2c_attach()
 */
#define DECLARE_I2C_DEVICE_NOCONF(_name_, _address_, _priv_)        \
    static struct i2c_device _name_ = {.name = #_name_,             \
                                       .address = _address_,        \
                                       .priv = _priv_,              \
                                       .driver.config_items = NULL, \
                                       .driver.config_items_count = 0};

struct i2c_device {
    /* common driver structure. */
    struct driver_common driver;
    /* driver interface ops structure */
    void* drv_ops;
    /* driver interface private data */
    void* drv_priv;

    /** Name of the device (for driver lookup) */
    char* name;
    /** Next device in the list */
    struct i2c_device* next;
    /** Bus to which the device is attached */
    struct i2c_bus_driver* bus;
    /** i2c address of the device on the bus */
    int address;
    /** instance private data */
    void* priv;
};

/**
 * Read data from an i2c slave.
 *
 * This function can be used in a synchronous way if callback parameter is NULL.
 *
 * @param drv the driver to use
 * @param buf the buffer where to put the read data
 * @param len the number of bytes to read
 * @param cb the async callback
 * @param cb_param the parameter passed to the callback function
 *
 * @return number of bytes read or negative error code.
 */
int i2c_read(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb, void* cb_param);

/**
 * Write data to an i2c slave.
 *
 * This function can be used in a synchronous way if callback parameter is NULL.
 *
 * @param drv the driver to use
 * @param buf the buffer data to transfer to slave
 * @param len the number of bytes to write
 * @param cb the async callback
 * @param cb_param the parameter passed to the callback function
 *
 * @return number of bytes writen or negative error code.
 */
int i2c_write(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb, void* cb_param);

/**
 * Write data to an i2c slave register.
 *
 * This function can only be called in a synchronous way.
 *
 * @param drv the driver to use
 * @param reg the register address to write
 * @param buf the buffer data to transfer to slave
 * @param len the number of bytes to write
 *
 * @return number of bytes writen or negative error code.
 */
int i2c_write_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len);

/**
 * Write data to an i2c slave.
 *
 * This function can be used in a synchronous way if callback parameter is NULL.
 *
 * @param drv the driver to use
 * @param reg the register address to read
 * @param buf the buffer data to transfer from slave
 * @param len the number of bytes to write
 *
 * @return number of bytes writen or negative error code.
 */
int i2c_read_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len);

/**
 * Attach a device to an i2c bus
 *
 * This function connects a device to a bus
 *
 * @param drv the i2c bus to attach device to
 * @param dev the device to attach to the bus
 * @param intf the driver interface that the device implements
 *
 * @return 0 if succeeds or negative error code.
 */
int i2c_attach(struct i2c_bus_driver* drv, struct i2c_device* dev, struct driver_common* intf);

/**
 * Detach a device from the i2c bus it's attached to
 *
 * This function disconnect a device from a bus
 *
 * @param dev the device to detach from the bus
 *
 * @return 0 if succeeds or negative error code.
 */
int i2c_detach(struct i2c_device* dev);

#define I2C_LOGI(...) elog_msg(&__log_I2C, ELOG_LEVEL_INFO, __VA_ARGS__)
#if defined CONFIG_DEBUG_LOG_I2C
#define I2C_LOGD(msg, ...) \
    elog_msg(&__log_I2C, ELOG_LEVEL_DEBUG, "%s:%u " #msg, __FUNCTION__, __LINE__, ##__VA_ARGS__)
#else
#define I2C_LOGD(...)
#endif
#define I2C_LOGE(...) elog_msg(&__log_I2C, ELOG_LEVEL_ERROR, __VA_ARGS__)
extern struct log_module __log_I2C;
/** @} i2c_bus
 * @} drivers
 */
#endif /* __I2C_BUS_H_ */
