/**
 * @file     interrupt.h
 * @brief    The API functions declared in this header file are called by upper levels (which
 *           include interrupt.h) to configure/use the interrupt services.
 *
 * @note     This interrupt public APIs are callable from LLCS(ex. DMA) and LLDs only.
 */

#ifndef __INTR_H_
#define __INTR_H_

#include <stdbool.h>
#include <stdint.h>

/** @addtogroup drivers
 * @{
 */

/** @addtogroup interrupt interrupt
 *  @note     This interrupt public APIs are callable from LLCS(ex. DMA) and LLDs only.
 *  @{
 **/

/**
 * Callback function type intr_handler_t
 * Callback invokes as an interrupt handler once interrupt is raised
 *
 */
typedef void (*intr_handler_t)(void);

/**
 * Sets the priority of a device specific interrupt.
 *
 * @param intr: Device specific interrupt number.
 * @param priority: priority for the interrupt.(lower value implies higher priority)
 *
 * @return 0 if success or return error code.
 */
int intr_set_priority(int intr, uint8_t priority);

/**
 * Reads the priority of a device specific interrupt.
 *
 * @param intr: Device specific interrupt number.
 *
 * @return priority of interrupt if success or return error code.
 *
 */
int intr_get_priority(int intr);

/**
 * Enables a device specific interrupt in the NVIC interrupt controller.
 *
 * @param intr - Device specific interrupt number.
 *
 * @note intr must not be negative, because we won't allow enable/disable of Exceptions; these are
 *       by default enabled
 *
 * @return 0 if success or return error code.
 */
int intr_enable(int intr);

/**
 * Returns the enable status of a device specific interrupt from the NVIC interrupt controller
 *
 * @param intr:  Device specific interrupt number.
 *
 * @note intr must not be negative, because we won't allow enable/disable of Exceptions; these are
 *       by default enabled
 *
 * @return 0 if interrupt is disabled or 1 if interrupt is enabled.
 */
bool intr_is_enabled(int intr);

/**
 * Disables a device specific interrupt in the NVIC interrupt controller and de-configure EXIU for a
 * specific interrupt connected with EXIU controller. Validates the interrupt, if pending and clear
 * the interrupt before disabling it
 *
 * @param intr:  Device specific interrupt number.
 *
 * @note intr must not be negative, because we won't allow enable/disable of Exceptions; these are
 *       by default enabled
 *
 * @return 0 if success or return error code.
 */
int intr_disable(int intr);

/**
 * Clears the pending bit of a device specific interrupt in the NVIC pending register.
 *
 * @param intr : Device specific interrupt number.
 *
 * @note intr must not be negative, because we won't allow enable/disable of Exceptions; these are
 *       by default enabled
 *
 * @return 0 if success or return error code.
 */
int intr_clear_pending(int intr);

/**
 * Sets the pending bit of a device specific interrupt in the NVIC pending register.
 *
 * @param intr: Device specific interrupt number.
 *
 * @note intr must not be negative, because we won't allow enable/disable of Exceptions; these are
 *       by default enabled
 *
 * @return 0 if success or return error code.
 */
int intr_set_pending(int intr);

/**
 * Sets an interrupt handler address in SRAM based interrupt vector table.
 *
 * @note VTOR must been relocated to SRAM before.
 *
 * @param intr Interrupt number
 * @param handle interrupt handler function pointer
 *
 * @note The interrupt number can be positive to specify a device specific interrupt, or negative to
 *       specify a processor exception
 *
 * @return 0 if success or return error code.
 */
int intr_set_vector(int intr, intr_handler_t handle);

/** Reads an interrupt handler address from interrupt vector table.
 *
 * @param intr Interrupt number
 *
 * @note The interrupt number can be positive to specify a device specific interrupt, or negative to
 *       specify a processor exception
 *
 * @return interrupt handler if success or NULL in case of failure
 */
intr_handler_t intr_get_vector(int intr);
/** @} interrupt
 * @} drivers
 */
#endif /*__INTR_H_*/
