#ifndef __PARTITION_H__
#define __PARTITION_H__

#include <stdint.h>

#include "driver.h"

/** @addtogroup infra
 * @{
 */

/** @addtogroup partitions partitions
 *  Documentation details of disk and partition abstraction
 *  @{
 */

#define PARTITION_NAME_SIZE 16
#define DISK_NAME_SIZE 16

struct disk;

/**
 * partition abstraction structure
 */
struct partition {
    /** Size of the partition ( in bytes ) */
    uint32_t size;

    /** Block size of the media (in bytes)*/
    uint32_t block_size;

    /** Size of the erase block ( in bytes ) */
    uint32_t erase_size;

    /** Size of the write block ( in bytes ) */
    uint32_t write_size;

    /** partition start from the begining of the disc */
    uint32_t offset;

    /** disk attached to the partition */
    struct disk* disk;

    /** name of the partition */
    char name[PARTITION_NAME_SIZE];
};

/** flag indicating that sectors must be erased before write */
#define DISK_FLAGS_NEED_ERASE 0x1

/**
 * Disk abstraction structure.
 *
 */
struct disk {
    struct driver_common drv;
    /** pointer to the function allowing to read data from a partition */
    int (*read)(struct partition* p, uint32_t start, uint8_t* buffer, uint32_t size);
    /** pointer to the function allowing to write data to a partition */
    int (*write)(struct partition* p, uint32_t start, uint8_t* buffer, uint32_t size);
    /** pointer to the function allowing to erase a block in the partition */
    int (*erase)(struct partition* p, uint32_t start, uint32_t count);
    /** size of the disc */
    int size;
    /** number of partitions in the disk */
    int partition_count;
    /** flags of the disk */
    int flags;
    /** disk private data */
    void* priv;
    /** partition table */
    struct partition* partitions;
};

#define DECLARE_DISK(_name_, _struct_, _priv_) DECLARE_DRIVER(_name_, disk, _struct_, _priv_)

/**
 * Read data from a partition
 *
 * @param p the partition to read from
 * @param start the start address in the partition to read from
 * @param buffer the buffer to place the read data
 * @param size the number of bytes to be read
 *
 * @return the number of bytes read or negative if error
 */
int partition_read(struct partition* p, uint32_t start, uint8_t* buffer, uint32_t size);

/**
 * Write data to a partition
 *
 * @param p the partition to write to
 * @param start the start address in the partition to write to
 * @param buffer the buffer with the data to write
 * @param size the number of bytes to write
 *
 * @return the number of bytes written or negative if error
 */
int partition_write(struct partition* p, uint32_t start, uint8_t* buffer, uint32_t size);

/**
 * erase blocks from partition
 *
 * @param p the partition to erase from
 * @param start the start address in the blocks to erase
 * @param size the number of bytes to erase (will round up to block size)
 *
 * @return the number of bytes read or negative if error
 */
int partition_erase(struct partition* p, uint32_t start, uint32_t size);

/**
 * iterate over all partitions in the system.
 *
 * For each partition the callback function will be called with the partition
 * and the private pointer as parameters.
 * The callback can return non-zero in order to stop iterating.
 *
 * @param callback the callback function that will be called for each
 * partition in the system
 * @param priv the priv parameter passed to the callback
 */
void partition_foreach(int (*callback)(struct partition* p, void* priv), void* priv);

/**
 * return the first partition matching a given name.
 *
 * This function iterates over all available partitions, and will return
 * the first one with the given name. If a non NULL disk name is specified,
 * it will only lookup int the disk with specified name.
 *
 * @param disk_name the disk name to be scanned for partition
 * @param partition_name the name of the partition to lookup
 *
 * @return partition if found, NULL if not found.
 */
struct partition* partition_lookup(char* disk_name, char* partition_name);

/**
 * @} partitions
 * @} infra
 */
#endif /* __PARTITION_H__ */