#ifndef __SPI_BUS_H_
#define __SPI_BUS_H_

/** @addtogroup drivers
 * @{
 */

/** @addtogroup spi_bus spi_bus
 *  Documentation details of SPI_BUS FW.
 *  @{
 */

#include <stdbool.h>
#include <stdint.h>
#include "driver.h"

struct spi_bus_driver;

/**
 * SPI Transfer mode
 **/
typedef enum {
    SPI_XFER_EXTENDED_SPI_MODE, /**< Operate with 1-line for Rx, 1-line for Tx */
    SPI_XFER_DUAL_SPI_MODE,     /**< Operate with 2-Lines for both Rx,Tx */
    SPI_XFER_QUAD_SPI_MODE,     /**< Operate with 4-Line for both Rx,Tx  */
    SPI_XFER_INVALD_SPI_MODE    /**< Invalid Transfer protocol           */
} spi_xfer_mode_t;

/**
 * Per-Slave SPI Controller Configuration data structure
 **/
typedef struct {
    uint8_t clk_div;      /**< Clock divider */
    bool shift_lsb_first; /**< Bit shift direction in a byte, TRUE->LSB, FALSE->MSB */
    bool slave_sel_pol;   /**< Slave select polarity TRUE -> HIGH, FLASE -> LOW */
} spi_slave_conf_t;

/**
 * Per-SPI transfer Configuration data structure
 **/
typedef struct {
    spi_xfer_mode_t mode; /**< Mode for this transfer */
    int size;             /**< Number of bytes to xfer */
    uint8_t* txd;         /**< Transmit data pointer, NULL for dummy writes */
    uint8_t* rxd;         /**< Receive data pointer, NULL for ignored read */
} spi_transfer_t;

/**
 * SPI transaction Configuration data structure
 **/
typedef struct {
    int slave_id;              /**< The slave concerned by this transaction */
    int xfer_count;            /**< The number of transfers in this transaction */
    spi_transfer_t* xfer_list; /**< The descriptors of the transfers in the transaction */
} spi_transaction_t;

/**
 * Generic callback function.
 * Invoked in case of SPI transaction completed or error occured.
 *
 * @param drv The spi driver in use
 * @param err_code Contains negative value in case of an error.
 * @param size Contains number of bytes read/written.
 * @param param Contains information passed to user.
 */
typedef void (*spi_bus_callback_t)(struct spi_bus_driver* drv, int err_code, uint32_t size,
                                   void* param);

struct spi_bus_ops {
    int (*slave_config_set)(struct spi_bus_driver* drv, const uint8_t slv_id,
                            const spi_slave_conf_t* conf);

    int (*read)(struct spi_bus_driver* drv, const uint8_t slv_id,
                const spi_xfer_mode_t transfer_mode, void* buf, uint32_t size,
                spi_bus_callback_t cb, void* cb_param);

    int (*write)(struct spi_bus_driver* drv, const uint8_t slv_id,
                 const spi_xfer_mode_t transfer_mode, void* buf, uint32_t size,
                 spi_bus_callback_t cb, void* cb_param);

    int (*transceive)(struct spi_bus_driver* drv, const uint8_t slv_id,
                      const spi_xfer_mode_t transfer_mode, void* txb, void* rxb, uint32_t tx_size,
                      uint32_t rx_size, spi_bus_callback_t cb, void* cb_param);

    int (*transact)(struct spi_bus_driver* drv, spi_transaction_t* transact, spi_bus_callback_t cb,
                    void* cb_param);
};

/**
 * spi bus driver handle.
 */
struct spi_bus_driver {
    /** Common driver interface */
    struct driver_common driver;
    /** spi driver interface implementation*/
    struct spi_bus_ops* ops;
    /** Private pointer */
    void* priv;
};

#define DECLARE_SPI_BUS(_name_, _ops_, _init_, _priv_)                                 \
    struct spi_bus_driver __spi_data_##_name_ = {.driver.init = _init_, .ops = _ops_}; \
    DECLARE_DRIVER(_name_, spi, &__spi_data_##_name_.driver, _priv_)

/**
 * Configure SPI Slave
 *
 * @param drv The SPI driver in use
 * @param slv_id Slave Id of the SPI device
 * @param conf SPI Slave configuration
 *
 * @return 0 If configuration is set successfully, negative value incase of error
 */
int spi_slave_config_set(struct spi_bus_driver* drv, const uint8_t slv_id,
                         const spi_slave_conf_t* conf);

/**
 * SPI read operation
 * -Useful in case of SPI transfers that doesnt involve SPI commands
 *
 * @param drv The SPI driver in use
 * @param slv_id Slave Id of the SPI device
 * @param transfer_mode Transfer mode configuration
 * @param buf Updated with Read data
 * @param size Transfer size in bytes
 * @param cb the Callback function to call on completion, NULL to use sync mode
 * @param cb_param The parameter to pass to the callback
 *
 * @return 0 If async request started successfully, negative value in case of error.
 *         Number of bytes read in case of successfull synchronous request, else negative value.
 *
 */
int spi_bus_read(struct spi_bus_driver* drv, const uint8_t slv_id,
                 const spi_xfer_mode_t transfer_mode, void* buf, uint32_t size,
                 spi_bus_callback_t cb, void* cb_param);

/**
 * Perform SPI write operation
 * -Useful in case of SPI transfers that doesnt involve SPI commands
 *
 * @param drv The SPI driver in use
 * @param slv_id Slave Id of the SPI device
 * @param transfer_mode Transfer mode configuration
 * @param buf User provided buffer contains data to be written
 * @param size Transfer size in bytes
 * @param cb the Callback function to call on completion, NULL to use sync mode
 * @param cb_param The parameter to pass to the callback
 *
 * @return 0 If async request started successfully, negative value incase of error.Number of bytes
 *         Number of bytes written in case of successfull synchronous request, else negative value.
 */
int spi_bus_write(struct spi_bus_driver* drv, const uint8_t slv_id,
                  const spi_xfer_mode_t transfer_mode, void* buf, uint32_t size,
                  spi_bus_callback_t cb, void* cb_param);

/**
 * Perform Simaltaneous SPI read and write operations
 * -Useful in case of SPI transfers that doesnt involve SPI commands
 *
 * @param drv The SPI driver in use
 * @param slv_id Slave Id of the SPI device
 * @param transfer_mode Transfer mode configuration
 * @param txb User provided buffer contains data to be written
 * @param rxb User provided buffer contains data read.
 * @param tx_size Transmit size in bytes
 * @param rx_size Receive size in bytes
 * @param cb the Callback function to call on completion, NULL to use sync mode
 * @param cb_param The parameter to pass to the callback
 *
 * @return 0 If async request started successfully, negative value incase of error.Number of bytes
 *         Number of bytes written in case of successfull synchronous request, else negative value.
 */
int spi_bus_transceive(struct spi_bus_driver* drv, const uint8_t slv_id,
                       const spi_xfer_mode_t transfer_mode, void* txb, void* rxb, uint32_t tx_size,
                       uint32_t rx_size, spi_bus_callback_t cb, void* cb_param);

/**
 * Perform SPI transaction (group of transfers)
 * -Useful in case of SPI transfers that involve SPI commands
 *
 * @param drv The SPI driver in use
 * @param transact SPI transaction configuration
 * @param cb the Callback function to call on completion, NULL to use sync mode
 * @param cb_param The parameter to pass to the callback
 *
 * @return 0 If async request started successfully, negative value incase of error.Number of bytes
 *         Number of bytes written in case of successfull synchronous request, else negative value.
 */
int spi_bus_transact(struct spi_bus_driver* drv, spi_transaction_t* transact, spi_bus_callback_t cb,
                     void* cb_param);

/**
 * Helper function to initialize read transaction of SPI BUS
 *
 * @param t_list Pointer to transfer item
 * @param buff output buffer for read transaction
 * @param size size of transfer
 */
void spi_bus_transaction_init_read(spi_transfer_t* t_list, uint8_t* buff, uint32_t size);

/**
 * Helper function to initialize write transaction of SPI BUS
 *
 * @param t_list Pointer to transfer item
 * @param buff input buffer for write transaction
 * @param size size of transfer
 */
void spi_bus_transaction_init_write(spi_transfer_t* t_list, const uint8_t* buff, uint32_t size);

#define SPI_BUS_LOGI(...) elog_msg(&__log_SPI_BUS, ELOG_LEVEL_INFO, __VA_ARGS__)
#if defined CONFIG_DEBUG_LOG_SPI_BUS
#define SPI_BUS_LOGD(msg, ...) \
    elog_msg(&__log_SPI_BUS, ELOG_LEVEL_DEBUG, "%s:%u " #msg, __FUNCTION__, __LINE__, ##__VA_ARGS__)
#else
#define SPI_BUS_LOGD(...)
#endif
#define SPI_BUS_LOGE(...) elog_msg(&__log_SPI_BUS, ELOG_LEVEL_ERROR, __VA_ARGS__)
extern struct log_module __log_SPI_BUS;

/** @} spi_bus
 * @} drivers
 */
#endif /* __SPI_BUS_H_ */
