#ifndef __SPI_FLASH_H_
#define __SPI_FLASH_H_

/** @addtogroup drivers
 * @{
 */

/** @addtogroup spi_flash spi_flash
 *  Documentation details of SPI_FLASH FW.
 *  @{
 */

#include <stdint.h>
#include "driver.h"
#include "spi_bus.h"

struct spi_flash_driver;

/**
 * Flash sector types
 **/
typedef enum {
#if CONFIG_ENABLE_BULK_ERASE
    SPI_FLASH_SECTOR_TYPE_WHOLE_FLASH, /**< Defines the whole flash */
#endif
    SPI_FLASH_SECTOR_TYPE_64KB = 0x10, /**< Flash divided into 64KB sectors */
    SPI_FLASH_SECTOR_TYPE_32KB = 0x0f, /**< Flash divided into 32KB sectors */
    SPI_FLASH_SECTOR_TYPE_4KB = 0x0c,  /**< Flash divided into 04KB sectors */
    SPI_FLASH_SECTOR_TYPE_NONE,        /**< Invalid sector type */
} spi_flash_sector_type_t;

/**
 * SPI Flash Information
 **/
typedef struct {
    uint32_t flash_size;       /**< Flash size in bytes */
    uint16_t vendor_id;        /**< Flash vendor Id ,specified by JEDEC */
    uint16_t device_id;        /**< Flash Device Id ,contains memory type and memory capacity */
    bool is_quad_rd_supp;      /**< Flag represents if Quad read is supported */
    bool is_dual_rd_supp;      /**< Flag represents if Dual read is supported */
    uint32_t min_era_sec_size; /**< Minimum erasable sector size */
} spi_flash_information_t;

/**
 * Known spi flash vendor ids
 *
 */
#define SPI_FLASH_VENDOR_CYPRESS 0x1
#define SPI_FLASH_VENDOR_MICRON 0x20

/**
 * Generic callback function.
 *
 * @param drv the spi flash driver in use
 * @param err_code contains negative value in case of an error.
 * @param size contains number of bytes read/written.
 * @param param Information to be passed to application
 */
typedef void (*spi_flash_callback_t)(struct spi_flash_driver* drv, int err_code, uint32_t size,
                                     void* param);

struct spi_flash_ops {
    int (*get_info)(struct spi_flash_driver* drv, spi_flash_information_t* flash_info);

    int (*read)(struct spi_flash_driver* drv, const uint32_t flash_addr, void* rd_buff,
                const uint32_t size, spi_flash_callback_t cb, void* cb_param);

    int (*write)(struct spi_flash_driver* drv, const uint32_t flash_addr, const void* wr_buff,
                 const uint32_t size, spi_flash_callback_t cb, void* cb_param);

    int (*erase)(struct spi_flash_driver* drv, const uint32_t sector_id, const uint32_t n_sectors,
                 const spi_flash_sector_type_t type, spi_flash_callback_t cb, void* cb_param);
    int (*erase_bulk)(struct spi_flash_driver* drv, spi_flash_callback_t cb, void* cb_param);
};

/**
 * spi flash driver handle.
 */
struct spi_flash_driver {
    /** Common driver interface */
    struct driver_common driver;
    /** spi driver interface implementation*/
    struct spi_flash_ops* ops;
    /** Private pointer */
    void* priv;
};

#define DECLARE_SPI_FLASH(_name_, _ops_, _init_, _read_conf_, _apply_conf_, _confstr_, _conf_cnt_, \
                          _priv_)                                                                  \
    struct spi_flash_driver __spi_data_##_name_ = {                                                \
        .driver.init = _init_,                                                                     \
        .ops = _ops_,                                                                              \
        .driver.read_config = _read_conf_,                                                         \
        .driver.apply_config = _apply_conf_,                                                       \
        .driver.config_items = _confstr_,                                                          \
        .driver.config_items_count = _conf_cnt_,                                                   \
    };                                                                                             \
    DECLARE_DRIVER(_name_, spi, &__spi_data_##_name_.driver, _priv_)

/**
 * Get identified Flash information
 * It is a synchronous call
 *
 * @param drv The SPI Flash driver in use
 * @param flash_info Updated with flash information like flash size,..
 *
 * @return 0  If flash is initialized successfully and flash_info is updated with valid values,
 * negative value in case of flash initialized.
 */
int spi_flash_get_info(struct spi_flash_driver* drv, spi_flash_information_t* flash_info);

/**
 *  Perform read operation on Flash
 *
 * @param drv the SPI Flash driver in use
 * @param flash_addr Flash device address to perform read.
 * @param rd_buff User buffer updated with the read data.
 * @param size Number of bytes to read.
 * @param cb the callback function to call on completion or error, NULL to use sync mode
 * @param cb_param the parameter to pass to the callback
 *
 * @return Negative value in case of error, number of bytes read if it is synchronous request, 0 in
 * case of a successful asynchronous request
 */
int spi_flash_read(struct spi_flash_driver* drv, const uint32_t flash_addr, void* rd_buff,
                   const uint32_t size, spi_flash_callback_t cb, void* cb_param);

/**
 *  Perform write operation on Flash
 *
 * @param drv the SPI Flash driver in use
 * @param flash_addr Flash device address to perform write.
 * @param wr_buff User buffer which contains data to be written.
 * @param size Number of bytes to write.
 * @param cb the callback function to call on completion or error, NULL to use sync mode
 * @param cb_param the parameter to pass to the callback
 *
 * @return Negative value in case of error, number of bytes written if it is synchronous request, 0
 * in case of a successful asynchronous request
 */
int spi_flash_write(struct spi_flash_driver* drv, const uint32_t flash_addr, const void* wr_buff,
                    const uint32_t size, spi_flash_callback_t cb, void* cb_param);

/**
 *  Perform erase operation on Flash sectors
 *
 * @param drv the SPI Flash driver in use
 * @param sector_id Starting sector to erase.
 * @param n_sectors Number of sectors to erase.
 * @param type sector type (64KB, 32KB or 4KB).
 * @param cb the callback function to call on completion or error, NULL to use sync mode
 * @param cb_param the parameter to pass to the callback
 *
 * @return Negative value in case of error, 0 if it is successful synchronous or asynchronous
 * request *
 */
int spi_flash_erase(struct spi_flash_driver* drv, const uint32_t sector_id,
                    const uint32_t n_sectors, spi_flash_sector_type_t type, spi_flash_callback_t cb,
                    void* cb_param);
/**
 *  Perform bulk erase operation on Flash sectors
 *
 * @param drv the SPI Flash driver in use
 * @param cb the callback function to call on completion or error, NULL to use sync mode
 * @param cb_param the parameter to pass to the callback
 *
 * @return Negative value in case of error, 0 if it is successful synchronous or asynchronous
 * request *
 */
int spi_flash_erase_bulk(struct spi_flash_driver* drv, spi_flash_callback_t cb, void* cb_param);

#define SPI_FLASH_LOGI(...) elog_msg(&__log_SPI_FLASH, ELOG_LEVEL_INFO, __VA_ARGS__)
#if defined CONFIG_DEBUG_LOG_SPI_FLASH
#define SPI_FLASH_LOGD(msg, ...)                                                        \
    elog_msg(&__log_SPI_FLASH, ELOG_LEVEL_DEBUG, "%s:%u " #msg, __FUNCTION__, __LINE__, \
             ##__VA_ARGS__)
#else
#define SPI_FLASH_LOGD(...)
#endif
#define SPI_FLASH_LOGE(...) elog_msg(&__log_SPI_FLASH, ELOG_LEVEL_ERROR, __VA_ARGS__)
extern struct log_module __log_SPI_FLASH;

/** @} spi_flash
 * @} drivers
 */
#endif /* __SPI_FLASH_H_ */
