#ifndef __UART_H_
#define __UART_H_

#include <stdint.h>
#include "driver.h"
#include "elog.h"

/* UART supported Stop bits constants */
#define UART_STOP_BIT_ONE 1
#define UART_STOP_BIT_TWO 2

/* UART supported WordLength constants */
#define UART_WORDLEN_5 5
#define UART_WORDLEN_6 6
#define UART_WORDLEN_7 7
#define UART_WORDLEN_8 8

/* UART supported Parity constants */
#define UART_PARITY_ODD 0
#define UART_PARITY_EVEN 1
#define UART_PARITY_NO 2

/* UART supported HardWare Flow controls */
#define UART_HWFLOW_CTRL_NONE 0
#define UART_HWFLOW_CTRL_CTS 1
#define UART_HWFLOW_CTRL_RTS 2
#define UART_HWFLOW_CTRL_ALL 3

/**
 *   @file uart.h
 *   This is public header file of uart interfaces..
 **/

/** @addtogroup drivers
 * @{
 */

/** @addtogroup uart uart
 *  Documentation details of uart.
 *  @{
 **/

typedef struct {
    uint32_t baud_rate;         /**< UART port bit rate configuration value */
    uint32_t word_len : 4;      /**< UART port word length  configuration value */
    uint32_t stop_bits : 2;     /**< UART port stop bit configuration value */
    uint32_t parity : 2;        /**< UART port parity bit configuration value */
    uint32_t flow_ctrl : 2;     /**< UART port flow control configuration value */
    uint32_t sticky_parity : 1; /**< Sticky parity bit generation and detection*/
} uart_config_t;

struct uart_driver;

/**
 * UART callback function, invokes when upon the transfer request is completed.
 *
 * @param drv the uart driver in use
 * @param cb_param Information to be passed to application
 * @param status Transfer status to be notified to the caller ex: 1-timeout
 * @param len number of character read from the UART or sent over the UART.
 */

typedef void (*uart_xfer_complete_callback)(struct uart_driver* drv, void* cb_param, int status,
                                            uint32_t len);

struct uart_stream_ops {
    /**
     * Opens and configures the  UART HW Port.
     *
     * @param drv the driver to use
     * @param uart_cfg pointer to an UART configuration structure.
     * @param refer uart_config_t for detailed configuration parameters
     *
     * @returns 0 incase of sucessfull open, else error code.
     */

    int (*open)(struct uart_driver* drv, const uart_config_t* uart_cfg);

    /**
     * initiate a read from an UART instance
     *
     * @param drv the UART driver in use
     * @param buf character buffer pointer where the UART read content is copied to.
     * @param len number of character expected to read from the UART.
     * @param timeout timeout in milliseconds, read operation is performed
     *                         until timeout is reached, or the specified numbers of
     *                         characters are read, whichever is earlier; and
     *                         terminated on timeout.. Incase of async mode, callback
     *                         is invoked with timeout status and bytes received.
     * @param cb the callback function to call on completion incase of Interrupt
     * and DMA transfer mode, NULL to use polling (or) sync mode
     * @param cb_params the parameter to pass to the callback
     *
     * @return 0 if async read request started.
     *               Number of read bytes if synchronous read is success, else returns error.
     *
     */
    int (*read)(struct uart_driver* drv, uint8_t* buf, const uint32_t len, const uint32_t timeout,
                uart_xfer_complete_callback cb, void* cb_param);

    /**
     * initiate a write to an UART instance
     *
     * @param drv the UART driver in use
     * @param buf character buffer pointer from  the UART  content is copied to.
     * @param len number of character expected to sent over UART.
     * @param timeout timeout in milliseconds, write operation is performed
     *                          until timeout is reached, or the specified numbers of
     *                          characters are transmitted, whichever is earlier; and
     *                          terminated on timeout.. Incase of async mode, callback
     *                          is invoked with timeout status and bytes transmitted.
     * @param cb the callback function to call on completion incase of Interrupt
     * and DMA transfer mode, NULL to use polling (or) sync mode
     * @param cb_params the parameter to pass to the callback
     *
     * @return 0 if async write request started.
     *               Number of bytes if synchronous write is success, else returns error.
     */

    int (*write)(struct uart_driver* drv, const uint8_t* buf, const uint32_t len,
                 const uint32_t timeout, uart_xfer_complete_callback cb, void* cb_param);

    /**
     * Closes the requested instance UART HW Port.
     *
     * @param drv the driver to use.
     *
     * @returns 0 upon success, error code on any failures.
     */
    int (*close)(struct uart_driver* drv);
};

/**
 *The following are the character based operations defined specific usage for an
 * Firmware based applications
 */

struct uart_char_ops {
    /**
     * initiate a read a character from an UART instance
     *
     * @param drv the UART driver in use
     *
     * @return the received char casted to int in case of success or else error code.
     */
    int (*get_char)(struct uart_driver* drv);

    /**
     * initiate a write to an UART instance
     *
     * @param drv the UART driver in use
     * @param ch character to sent over UART port.
     *
     * @returns 0 upon success, error code on any failures.
     */
    int (*put_char)(struct uart_driver* drv, const uint8_t ch);

    /**
     * Configures the  UART HW Port.
     *
     * @param drv the driver to use
     * @param uart_cfg pointer to an UART configuration structure.
     * @param refer uart_config_t for detailed configuration parameters
     *
     * @returns 0 incase of sucessfullly configuration, else error code.
     */

    int (*set_config)(struct uart_driver* drv, const uart_config_t* uart_cfg);
};

/**
 * UART interface driver handle.
 */
struct uart_driver {
    /** Common driver interface */
    struct driver_common driver;
    /** UART driver interface implementation*/
    struct uart_stream_ops* ops;
    struct uart_char_ops* char_ops;
    /** Private pointer */
    void* priv;
};

#define DECLARE_UART_INTF(_name_, _ops_, _char_ops_, _init_, _priv_, _clk_) \
    struct uart_driver __uart_data_##_name_ = {                             \
        .driver.init = _init_, .ops = _ops_, .char_ops = _char_ops_};       \
    DECLARE_DRIVER_CLK(_name_, uart, &__uart_data_##_name_.driver, _priv_, _clk_)

/**
 * Opens and configures the  UART HW Port.
 *
 * @param drv the driver to use
 * @param uart_cfg pointer to an UART configuration structure.
 *
 * @returns 0 incase of sucessfull open, else error code.
 */

int uart_open(struct uart_driver* drv, const uart_config_t* uart_cfg);

/**
 * Read from an UART port
 *
 * @param drv the UART driver in use
 * @param buf character buffer pointer where the UART read content is copied to.
 * @param len number of character expected to read from the UART.
 * @param timeout timeout in milliseconds, read operation is performed
 *                         until timeout is reached, or the specified numbers of
 *                         characters are read, whichever is earlier; and
 *                         terminated on timeout.. Incase of async mode, callback
 *                         is invoked with timeout status and bytes received.
 * @param cb the callback function to call on completion incase of Interrupt
 * and DMA transfer mode, NULL to use polling (or) sync mode
 * @param cb_param the parameter to pass to the callback
 *
 * @return 0 if async read request started.
 *               Number of read bytes if synchronous read is success, else returns error.
 *
 */
int uart_read(struct uart_driver* drv, uint8_t* buf, const uint32_t len, const uint32_t timeout,
              uart_xfer_complete_callback cb, void* cb_param);

/**
 * Write to an UART port
 *
 * @param drv the UART driver in use
 * @param buf character buffer pointer from  the UART  content is copied to.
 * @param len number of character expected to sent over UART.
 * @param timeout timeout in milliseconds, write operation is performed
 *                          until timeout is reached, or the specified numbers of
 *                          characters are transmitted, whichever is earlier; and
 *                          terminated on timeout.. Incase of async mode, callback
 *                          is invoked with timeout status and bytes transmitted.
 * @param cb the callback function to call on completion incase of Interrupt
 * and DMA transfer mode, NULL to use polling (or) sync mode
 * @param cb_param the parameter to pass to the callback
 *
 * @return 0 if async write request started.
 *               Number of bytes if synchronous write is success, else returns error.
 */
int uart_write(struct uart_driver* drv, const uint8_t* buf, const uint32_t len,
               const uint32_t timeout, uart_xfer_complete_callback cb, void* cb_param);

/**
 * Closes the UART HW Port.
 *
 * @param drv the driver to use.
 *
 * @returns 0 upon success, error code on any failures.
 */
int uart_close(struct uart_driver* drv);

/**
 * Read a character from an UART instance. The character which is read is an unsigned char
 * which is converted to an integer value.
 *
 * @param drv the UART driver in use
 *
 * @return the received char casted to int in case of success or else error code.
 */
int uart_get_char(struct uart_driver* drv);

/**
 * Write to an UART instance
 *
 * @param drv the UART driver in use
 * @param ch character to sent over UART port.
 *
 * @returns 0 upon success, error code on any failures.
 */

int uart_put_char(struct uart_driver* drv, const uint8_t ch);

/**
 * Configures the  UART HW Port.
 *
 * @param drv the driver to use
 * @param uart_cfg pointer to an UART configuration structure.
 *
 * @returns 0 incase of sucessfullly configuration, else error code.
 */

int uart_set_config(struct uart_driver* drv, const uart_config_t* uart_cfg);

#define UART_LOGI(...) elog_msg(&__log_UART, ELOG_LEVEL_INFO, __VA_ARGS__)
#if defined CONFIG_DEBUG_LOG_UART
#define UART_LOGD(msg, ...) \
    elog_msg(&__log_UART, ELOG_LEVEL_DEBUG, "%s:%u " #msg, __FUNCTION__, __LINE__, ##__VA_ARGS__)
#else
#define UART_LOGD(...)
#endif
#define UART_LOGE(...) elog_msg(&__log_UART, ELOG_LEVEL_ERROR, __VA_ARGS__)
extern struct log_module __log_UART;
/** @} uart
 * @} drivers
 */
#endif /*  __UART_H_  */
