#ifndef __WDOG_H_
#define __WDOG_H_

#include <stdint.h>

#include "driver.h"
/**
 *   @file wdog.h
 *   This is public header file of wdog interfaces.
 **/

/** @addtogroup drivers
 * @{
 */

/** @addtogroup watchdog watchdog
 *  Documentation details of watch dog timer
 * @{
 **/

struct wdog_driver;

/**
 * Expire handler of type wdog_expire_handler_t
 * Expire handler invokes once the wdog timeout expires
 *
 * @param drv the wdog driver in use
 * @param param the parameter to pass to the expire handler
 */
typedef void (*wdog_expire_handler_t)(struct wdog_driver* drv, void* param);

struct wdog_ops {
    int (*set_timeout)(struct wdog_driver* drv, uint32_t timeout);

    int (*set_expire_handler)(struct wdog_driver* drv, wdog_expire_handler_t handle, void* param);

    int (*start)(struct wdog_driver* drv);

    void (*kick)(struct wdog_driver* drv);

    int (*stop)(struct wdog_driver* drv);
};

/**
 * Watchdog driver handle.
 */
struct wdog_driver {
    /** Common driver interface */
    struct driver_common driver;
    /** Watchdog driver interface implementation*/
    struct wdog_ops* ops;
    /** Private pointer */
    void* priv;
};

#define DECLARE_WDOG(_name_, _ops_, _init_, _read_conf_, _apply_conf_, _confstr_, _conf_cnt_, \
                     _priv_, _clk_)                                                           \
    struct wdog_driver __wdog_data_##_name_ = {.driver.init = _init_,                         \
                                               .driver.read_config = _read_conf_,             \
                                               .driver.apply_config = _apply_conf_,           \
                                               .driver.config_items = _confstr_,              \
                                               .driver.config_items_count = _conf_cnt_,       \
                                               .ops = _ops_};                                 \
    DECLARE_DRIVER_CLK(_name_, wdog, &__wdog_data_##_name_.driver, _priv_, _clk_)

/**
 * Configure Watchdog load value based on the timeout
 *
 * @param drv the wdog driver in use
 * @param timeout the timeout period in milliseconds.
 *
 * @return 0 if success or return error code in case of failure
 */
int wdog_set_timeout(struct wdog_driver* drv, uint32_t timeout);

/**
 * Start the wdog device
 *
 * @param drv the wdog driver in use
 *
 * @return 0 if success or return error code in case of failure
 */
int wdog_start(struct wdog_driver* drv);

/**
 * Kick the WDOG
 *
 * @param drv the wdog driver in use
 */
void wdog_kick(struct wdog_driver* drv);

/**
 * Register a WDOG expire handler
 * @param drv the wdog driver in use
 * @param handle the expire handler to call once timeout expires
 * @param param the parameter to pass to the expire handler
 *
 * @return 0 if success or return error code in case of failure
 */
int wdog_set_expire_handler(struct wdog_driver* drv, wdog_expire_handler_t handle, void* param);

/**
 * Stop the wdog device
 *
 * @param drv the wdog driver in use
 *
 * @return 0 if success or return error code in case of failure
 */
int wdog_stop(struct wdog_driver* drv);

/** @} wdog
 * @} drivers
 */
#endif /* __WDOG_H_ */
