#include "akida_dma.h"
#include <stdlib.h>
#include <string.h>
#include "akida_dma_reg.h"
#include "cli_cmd.h"
#include "common.h"
#include "elog.h"
#include "interrupt.h"
#include "nsocv1/dma_common.h"
#include "soc.h"

#define HDMAC_CHANNEL_MAX (8)
#define MAX_TIMEOUT (10000)  // Based on better judgment this value may be changed
#define START (0)
#define STOP (1)
#define HALT (2)

#define HDMAC_TRANSFER_REQUEST_EXTERNAL (0)
#define HDMAC_TRANSFER_REQUEST_PERIPHERAL (1)
#define HDMAC_TRANSFER_REQUEST_SOFTWARE (2)

#define GET_HANDLER(x) HDMAC_IRQ_Handler##x

/** A structure to abstract
 **  DMA registers
 **/
typedef volatile struct {
    __IO uint32_t DMACR;  /**< Controller configuration register */
    uint32_t Reserved[3]; /**< Reserved */
    struct HDMAC_Ch {
        __IO uint32_t DMACA;  /**< Configuration A register */
        __IO uint32_t DMACB;  /**< Configuration B register */
        __IO uint32_t DMACSA; /**< Source address           */
        __IO uint32_t DMACDA; /**< Destination address      */
    } ch[HDMAC_CHANNEL_MAX];  /**< Perchannel registers     */
} HDMAC_t;

HDMAC_t* HDMAC = ((HDMAC_t*)HDMAC_BASE);

/* Local data structures ,START*/
/** Peripheral Id to channel Id mapping */
static int per_to_channel_map[HDMAC_CHANNEL_MAX] = {0, 1, 2, 3, 4, 5, 6, 7};

static int peri_to_reqpin_map[HDMAC_CHANNEL_MAX] = {
    BC_HDMAC_DMACA_IS_IDREQ_3, BC_HDMAC_DMACA_IS_IDREQ_1, BC_HDMAC_DMACA_IS_IDREQ_7,
    BC_HDMAC_DMACA_IS_IDREQ_5, BC_HDMAC_DMACA_IS_DREQ_H,  BC_HDMAC_DMACA_IS_DREQ_H,
    BC_HDMAC_DMACA_IS_DREQ_H,  BC_HDMAC_DMACA_IS_ST,
};
typedef struct {
    dma_callback_t cb;
    void* cb_param;
} dma_priv_t;
static dma_priv_t priv_data[HDMAC_CHANNEL_MAX];

/* Local data structures ,END*/

/* Local function declerations ,START*/

__STATIC_INLINE void hdmac_controller_conf_set(bool is_fixed_priority, uint8_t start);
__STATIC_INLINE void hdmac_channel_perform_dma_operation_priv(uint8_t ch_id,
                                                              uint8_t channel_operation);
__STATIC_INLINE void hdmac_process_intr(uint8_t ch_id);

const char* __hdmac_mode_str[] = {"BLOCK", "BURST", "DEMAND", "NONE"};

const char* __hdmac_beat_str[] = {"SINGLE", "INCR",   "WRAP4",  "INCR4", "WRAP8",
                                  "INCR8",  "WRAP16", "INCR16", "NONE"};

const char* __hdmac_bool_str[] = {"FALSE", "TRUE"};

const char* __hdmac_stop_status_str[] = {"Unkown",
                                         "Address over flow error",
                                         "DMA Transfer stopped",
                                         "Source access error ",
                                         "Destination access error",
                                         "DMA transfer done",
                                         "",
                                         "DMA Transfer Paused",
                                         "NONE"};

void HDMAC_IRQ_Handler_0(void);
void HDMAC_IRQ_Handler_1(void);
void HDMAC_IRQ_Handler_2(void);
void HDMAC_IRQ_Handler_3(void);
void HDMAC_IRQ_Handler_4(void);
void HDMAC_IRQ_Handler_5(void);
void HDMAC_IRQ_Handler_6(void);
void HDMAC_IRQ_Handler_7(void);

/* Local function declerations ,END*/

/* Public function definitions ,START*/

int akida_hdmac_init(void) {
    uint8_t loop_var = 0;

    intr_set_vector(INTR_HDMAC0, HDMAC_IRQ_Handler_0);
    intr_set_vector(INTR_HDMAC1, HDMAC_IRQ_Handler_1);
    intr_set_vector(INTR_HDMAC2, HDMAC_IRQ_Handler_2);
    intr_set_vector(INTR_HDMAC3, HDMAC_IRQ_Handler_3);
    intr_set_vector(INTR_HDMAC4, HDMAC_IRQ_Handler_4);
    intr_set_vector(INTR_HDMAC5, HDMAC_IRQ_Handler_5);
    intr_set_vector(INTR_HDMAC6, HDMAC_IRQ_Handler_6);
    intr_set_vector(INTR_HDMAC7, HDMAC_IRQ_Handler_7);
    for (loop_var = 0; loop_var < HDMAC_CHANNEL_MAX; loop_var++) {
        if (intr_enable(INTR_HDMAC0 + loop_var)) {
            return EFAILURE;
        }
    }
    hdmac_controller_conf_set(true, START);
    /**
     * @note Return values would be replaced by macros
     **/
    return SUCCESS;
}

int akida_hdmac_default_request_conf_get(hdmac_request_conf_t* req_conf) {
    if (NULL == req_conf) {
        return ENULL_POINTER;
    }
    req_conf->transfer_mode = HDMAC_TRANSFER_MODE_BLOCK;
    req_conf->transfer_width = 8;
    req_conf->beat_transfer = HDMAC_BEAT_TRANSFER_SINGLE;
    req_conf->block_count = 0x10;
    req_conf->single_req = false;
    req_conf->is_src_addr_fixed = false;
    req_conf->is_dest_addr_fixed = false;
    return SUCCESS;
}

int akida_hdmac_request_transfer(const hdmac_peripheral_id_t per_id, hdmac_request_conf_t* req_conf,
                                 dma_callback_t cb, void* cb_param) {
    uint8_t ch_id = per_to_channel_map[per_id], req_type = 0;
    uint32_t reg_val = 0;

    if (NULL == req_conf) {
        return ENULL_POINTER;
    }
    // Only supported modes are BLOCK and BURST mode and any other mode requested should be
    // invalidated
    if ((req_conf->transfer_mode == HDMAC_TRANSFER_MODE_DEMAND) &&
        (per_id == HDMAC_PERIPHERAL_ID_SW)) {
        return EINVALID_PARM;
    }
    REG_WRITE_MASK(HDMAC->ch[ch_id].DMACB, BC_HDMAC_DMACB_SS_SET(HDMAC_STATUS_IDLE),
                   BC_HDMAC_DMACB_SS_MASK);
    switch (per_id) {
        case HDMAC_PERIPHERAL_ID_UART_0_RX:
        case HDMAC_PERIPHERAL_ID_UART_1_RX:
        case HDMAC_PERIPHERAL_ID_UART_0_TX:
        case HDMAC_PERIPHERAL_ID_UART_1_TX:
            req_type = HDMAC_TRANSFER_REQUEST_PERIPHERAL;
            break;
        case HDMAC_PERIPHERAL_ID_SPI_RX:
        case HDMAC_PERIPHERAL_ID_SPI_TX:
        case HDMAC_PERIPHERAL_ID_I2S:
            req_type = HDMAC_TRANSFER_REQUEST_EXTERNAL;
            break;
        case HDMAC_PERIPHERAL_ID_SW:
            req_type = HDMAC_TRANSFER_REQUEST_SOFTWARE;
            break;
        default:
            req_type = HDMAC_TRANSFER_REQUEST_SOFTWARE;
            break;
    }
    REG_WRITE_MASK(HDMAC->ch[ch_id].DMACA,
                   (BC_HDMAC_DMACA_BT_SET(req_conf->beat_transfer)) |
                       (BC_HDMAC_DMACA_BC_SET((req_conf->block_count - 1))) |
                       (BC_HDMAC_DMACA_TC_SET((req_conf->transfer_count - 1))),
                   BC_HDMAC_DMACA_BT_MASK | BC_HDMAC_DMACA_BC_MASK | BC_HDMAC_DMACA_TC_MASK);

    /** Configure Transfer mode and transfer width*/
    reg_val = req_conf->transfer_width / 16;
    REG_WRITE_MASK(
        HDMAC->ch[ch_id].DMACB,
        (BC_HDMAC_DMACB_MS_SET(req_conf->transfer_mode) | BC_HDMAC_DMACB_TW_SET(reg_val)),
        BC_HDMAC_DMACB_MS_MASK | BC_HDMAC_DMACB_TW_MASK | BC_HDMAC_DMACB_TW_MASK);

    /** Configure flags fixed or increment addressess */
    REG_WRITE_MASK(HDMAC->ch[ch_id].DMACB,
                   (req_conf->is_src_addr_fixed ? BC_HDMAC_DMACB_FS_SET(1) : 0) |
                       (req_conf->is_dest_addr_fixed ? BC_HDMAC_DMACB_FD_SET(1) : 0),
                   BC_HDMAC_DMACB_FS_MASK | BC_HDMAC_DMACB_FD_MASK);

    /** Configure source address */
    REG_WRITE(HDMAC->ch[ch_id].DMACSA, req_conf->src_address);

    /** Configure destination address */
    REG_WRITE(HDMAC->ch[ch_id].DMACDA, req_conf->dest_address);

    /** Configure interrupt mask */
    if (cb) {
        REG_WRITE_MASK(HDMAC->ch[ch_id].DMACB, BC_HDMAC_DMACB_EI_SET(1) | BC_HDMAC_DMACB_CI_SET(1),
                       BC_HDMAC_DMACB_EI_MASK | BC_HDMAC_DMACB_CI_MASK);
    } else {
        REG_WRITE_MASK(HDMAC->ch[ch_id].DMACB, 0, BC_HDMAC_DMACB_EI_MASK | BC_HDMAC_DMACB_CI_MASK);
    }

    /** Configure input select*/
    reg_val = 0;
    if (false == req_conf->single_req || per_id >= HDMAC_PERIPHERAL_ID_I2S) {
        reg_val |= BC_HDMAC_DMACA_IS_SET(peri_to_reqpin_map[per_id]);
    } else if (per_id <= HDMAC_PERIPHERAL_ID_UART_1_TX) {
        switch (per_id) {
            case HDMAC_PERIPHERAL_ID_UART_0_RX:
                reg_val |= BC_HDMAC_DMACA_IS_SET(BC_HDMAC_DMACA_IS_IDREQ_2);
                break;
            case HDMAC_PERIPHERAL_ID_UART_0_TX:
                reg_val |= BC_HDMAC_DMACA_IS_SET(BC_HDMAC_DMACA_IS_IDREQ_0);
                break;
            case HDMAC_PERIPHERAL_ID_UART_1_RX:
                reg_val |= BC_HDMAC_DMACA_IS_SET(BC_HDMAC_DMACA_IS_IDREQ_6);
                break;
            case HDMAC_PERIPHERAL_ID_UART_1_TX:
                reg_val |= BC_HDMAC_DMACA_IS_SET(BC_HDMAC_DMACA_IS_IDREQ_4);
                break;
            default:
                break;
        }
    }
    REG_WRITE_MASK(HDMAC->ch[ch_id].DMACA, reg_val, BC_HDMAC_DMACA_IS_MASK);
    priv_data[ch_id].cb = cb;
    priv_data[ch_id].cb_param = cb_param;

    if (HDMAC_TRANSFER_REQUEST_SOFTWARE == req_type) {
        REG_WRITE_MASK(HDMAC->ch[ch_id].DMACA, BC_HDMAC_DMACA_ST_SET(1), BC_HDMAC_DMACA_ST_MASK);
    } else {
        REG_WRITE_MASK(HDMAC->ch[ch_id].DMACA, BC_HDMAC_DMACA_ST_SET(0), BC_HDMAC_DMACA_ST_MASK);
    }
    hdmac_channel_perform_dma_operation_priv(ch_id, START);
    if (NULL == cb) {
        while (1) {
            /*TODO: Implement proper timeout */
            if (!(BC_HDMAC_DMACA_EB_GET(REG_READ(HDMAC->ch[ch_id].DMACA)))) {
                /** exit if enable bit is set to 0 by HDMAC */
                break;
            } else if (HDMAC_STATUS_DMA_PAUSE ==
                       BC_HDMAC_DMACB_SS_GET(REG_READ(HDMAC->ch[ch_id].DMACB))) {
                /** exit if DMA PAUSE request is issued by user during transfer*/
                break;
            } else if (HDMAC_STATUS_STOP_REQ ==
                       BC_HDMAC_DMACB_SS_GET(REG_READ(HDMAC->ch[ch_id].DMACB))) {
                /** exit if DMA STOP request is issued by user during transfer*/
                break;
            }
        }
    }
    return SUCCESS;
}

int akida_hdmac_control_operation(hdmac_peripheral_id_t per_id, hdmac_operation_t dma_operation) {
    uint8_t ch_id;
    ch_id = per_to_channel_map[per_id];

    REG_WRITE_MASK(HDMAC->ch[ch_id].DMACB, BC_HDMAC_DMACB_SS_SET(HDMAC_STATUS_IDLE),
                   BC_HDMAC_DMACB_SS_MASK);
    switch (dma_operation) {
        case HDMAC_OPERATION_UNPAUSE:
            if (REG_READ(HDMAC->ch[ch_id].DMACA) & BC_HDMAC_DMACA_PB_MASK) {
                hdmac_channel_perform_dma_operation_priv(ch_id, START);
            } else {
                /** There was no pause attempted on this channel */
                return EFAILURE;
            }
            break;
        case HDMAC_OPERATION_DISABLE:
            if (REG_READ(HDMAC->ch[ch_id].DMACA) & BC_HDMAC_DMACA_EB_MASK) {
                hdmac_channel_perform_dma_operation_priv(ch_id, STOP);
            } else {
                /** There was no ongoing transfer on this channel */
                return EFAILURE;
            }
            break;
        case HDMAC_OPERATION_PAUSE:
            if (REG_READ(HDMAC->ch[ch_id].DMACA) & BC_HDMAC_DMACA_EB_MASK) {
                hdmac_channel_perform_dma_operation_priv(ch_id, HALT);
            } else {
                /** There was no ongoing transfer on this channel */
                return EFAILURE;
            }
            break;
        default:
            break;
    }

    return SUCCESS;
}

int akida_hdmac_get_transfer_status(hdmac_peripheral_id_t per_id, hdmac_status_t* dma_status) {
    uint8_t ch_id;
    uint32_t reg_val = 0;

    if (NULL == dma_status) {
        return ENULL_POINTER;
    }
    ch_id = per_to_channel_map[per_id];
    /** Reading previous stop status of a channel*/
    reg_val = REG_READ(HDMAC->ch[ch_id].DMACB);
    dma_status->stop_status = BC_HDMAC_DMACB_SS_GET(reg_val);
    /** Reading current block count and transfer count*/
    reg_val = REG_READ(HDMAC->ch[ch_id].DMACA);
    dma_status->enable = BC_HDMAC_DMACA_EB_GET(reg_val) ? true : false;
    dma_status->sw_trigger = BC_HDMAC_DMACA_ST_GET(reg_val) ? true : false;
    dma_status->block_count = BC_HDMAC_DMACA_BC_GET(reg_val);
    dma_status->transfer_count = BC_HDMAC_DMACA_TC_GET(reg_val);
    return SUCCESS;
}
/* Public function definitions ,END*/

/* Local function definitions ,START*/

__STATIC_INLINE void hdmac_controller_conf_set(bool is_fixed_priority, uint8_t start) {
    REG_WRITE_MASK(HDMAC->DMACR,
                   (BC_HDMAC_DMACR_PR_SET(is_fixed_priority ? 0 : 1) |
                    BC_HDMAC_DMACR_DE_SET((start == START ? 1 : 0))),
                   (BC_HDMAC_DMACR_PR_MASK | BC_HDMAC_DMACR_DE_MASK));
}

__STATIC_INLINE void hdmac_channel_perform_dma_operation_priv(uint8_t ch_id,
                                                              uint8_t channel_operation) {
    switch (channel_operation) {
        case START:
            REG_WRITE_MASK(HDMAC->ch[ch_id].DMACA, BC_HDMAC_DMACA_EB_SET(1),
                           (BC_HDMAC_DMACA_EB_MASK | BC_HDMAC_DMACA_PB_MASK));
            break;
        case STOP:
            REG_WRITE_MASK(HDMAC->ch[ch_id].DMACA, BC_HDMAC_DMACA_EB_SET(0),
                           (BC_HDMAC_DMACA_EB_MASK));
            break;
        case HALT:
            REG_WRITE_MASK(HDMAC->ch[ch_id].DMACA, BC_HDMAC_DMACA_PB_SET(1),
                           (BC_HDMAC_DMACA_PB_MASK));
            break;
        default:
            break;
    }
}

__STATIC_INLINE void hdmac_process_intr(uint8_t ch_id) {
    hdmac_stop_status_t stop_status = HDMAC_STATUS_NONE;
    dma_callback_t usr_cb = priv_data[ch_id].cb;
    void* param = priv_data[ch_id].cb_param;

    REG_WRITE_MASK(HDMAC->ch[ch_id].DMACB, 0, BC_HDMAC_DMACB_EI_MASK | BC_HDMAC_DMACB_CI_MASK);
    stop_status = BC_HDMAC_DMACB_SS_GET(REG_READ(HDMAC->ch[ch_id].DMACB));

    /* Setting Stopstatus bitfeild to 0 clears Interrupts */
    REG_WRITE_MASK(HDMAC->ch[ch_id].DMACB, BC_HDMAC_DMACB_SS_SET(HDMAC_STATUS_IDLE),
                   BC_HDMAC_DMACB_SS_MASK);

    if (usr_cb) {
        usr_cb(ch_id, stop_status, param);
    }
}

/* Local function definitions ,END*/

/* Irq handler definitions ,START*/
/* DMA perchannel interrupt handlers */
void HDMAC_IRQ_Handler_0(void) {
    hdmac_process_intr(0);
}
void HDMAC_IRQ_Handler_1(void) {
    hdmac_process_intr(1);
}
void HDMAC_IRQ_Handler_2(void) {
    hdmac_process_intr(2);
}
void HDMAC_IRQ_Handler_3(void) {
    hdmac_process_intr(3);
}
void HDMAC_IRQ_Handler_4(void) {
    hdmac_process_intr(4);
}
void HDMAC_IRQ_Handler_5(void) {
    hdmac_process_intr(5);
}
void HDMAC_IRQ_Handler_6(void) {
    hdmac_process_intr(6);
}
void HDMAC_IRQ_Handler_7(void) {
    hdmac_process_intr(7);
}

/* Irq handler definitions ,END*/

int dma_debug_cli(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    hdmac_request_conf_t req;
    hdmac_status_t req_status;
    hdmac_peripheral_id_t p_id;
    char* rem;
    uint8_t rv = 0;

    if (argc > 1) {
        if (!strncmp(argv[1], "getconfig", 9) && (2 != argc)) {
            rv = EFAILURE;
        } else if (!strncmp(argv[1], "status", 6) && (3 != argc)) {
            rv = EFAILURE;
        }
        if (rv) {
            cli_printf(ctx, "%d Invalid no of arguments\n", argc);
        }
    } else {
        rv = EFAILURE;
        cli_printf(ctx, "%d Invalid no of arguments\n", argc);
    }

    if (!rv && !strncmp(argv[1], "getconfig", 9)) {
        cli_printf(ctx, "Reading the default configuration\n");
        rv = akida_hdmac_default_request_conf_get(&req);

        cli_printf(ctx,
                   "Transfer mode = %s\n"
                   "Transfer width = %u bits\n"
                   "Beat transfer = %s\n",
                   __hdmac_mode_str[req.transfer_mode], req.transfer_width,
                   __hdmac_beat_str[(req.beat_transfer - 8)]);
        cli_printf(ctx,
                   "Block count = 0x%02x\nSingle req = %s\nSrc addr fixed = %s\nDest addr "
                   "fixed = %s\n",
                   req.block_count, __hdmac_bool_str[req.single_req],
                   __hdmac_bool_str[req.is_src_addr_fixed],
                   __hdmac_bool_str[req.is_dest_addr_fixed]);
        cli_printf(ctx, "Syntax: debug dma getconfig\n");
        if (rv) {
            cli_printf(ctx, "DBG-CLI-4 Result:Fail\n");
        } else {
            cli_printf(ctx, "DBG-CLI-4 Result:Pass\n");
        }

    } else if (!rv && !strncmp(argv[1], "status", 6)) {
        p_id = strtoul(argv[2], &rem, 10);
        if (p_id > 7) {
            rv = EFAILURE;
            cli_printf(ctx, "%s Invalid peripheral ID\n", argv[2]);
        }
        if (!rv) {
            rv = akida_hdmac_get_transfer_status(p_id, &req_status);
            cli_printf(ctx,
                       "Per=%u enable=%u  BC=%u  TC=%u  sw_trigger=%u  Status=%s Status_val=%u\n",
                       p_id, req_status.enable, req_status.block_count, req_status.transfer_count,
                       req_status.sw_trigger,
                       __hdmac_stop_status_str[(hdmac_stop_status_t)req_status.stop_status],
                       (hdmac_stop_status_t)req_status.stop_status);
        }

        cli_printf(ctx, "Syntax: dma status <per_id>\n");
        if (rv) {
            cli_printf(ctx, "DBG-CLI-5 Result:Fail\n");
        } else {
            cli_printf(ctx, "DBG-CLI-5 Result:Pass\n");
        }
    } else {
        rv = 1;
        cli_printf(ctx, "%s command not supported\n", argv[1]);
    }

    if (rv) {
        cli_printf(ctx, "CLI error\n");
        cli_printf(ctx, "COMMAND HELP::\n");
        cli_printf(ctx, "SYNTAX FOR DMA GET CONFIG :: debug dma getconfig\n");
        cli_printf(ctx, "DESCRIPTION FOR DMA GET CONFIG ::\n");
        cli_printf(ctx, "getconfig   : Gets default configuration\n");
        cli_printf(ctx, "SYNTAX FOR DMA TRANSFER STATUS	:: debug dma status <per_id>\n");
        cli_printf(ctx, "DESCRIPTION FOR DMA TRANSFER STATUS::\n");
        cli_printf(ctx, "p_id   : Represents peripheral id \n");
    }
    return rv;
}

/** @page d_commands DEBUG
 * @section debug_dma DEBUG_DMA
 * @ref DBG-CLI-4 \n
 * @ref DBG-CLI-5 \n
 */

/** @page d_commands DEBUG
 * 	@subsection DBG-CLI-4 debug dma getconfig
 *   CLI to print default configured values \n
 * 	__syntax			: debug dma getconfig__ \n
 * 	__getconfig__		: Prints default configuration \n
 */

/** @page d_commands DEBUG
 * 	@subsection DBG-CLI-5 debug dma status
 *   CLI to print status of the peripheral ID \n
 * 	__syntax		: debug dma status &lt;p_id&gt;__ \n
 * 	__p_id__	    : Represents peripheral id \n
 */

DECLARE_DEBUG_CMD(dma, dma_debug_cli);
