#include <stdint.h>
#include "bsp_error.h"
#include "common.h"
#include "exiu.h"
#include "gpio.h"
#include "gpio_reg.h"
#include "interrupt.h"

typedef struct {
    struct gpio_driver* drv;
    gpio_callback cb;
    void* cb_param;
    uint8_t port;
    uint8_t bit;
    gpio_intr_mode_t intr_mode;
} gpio_instance_t;

gpio_instance_t gpio = {.drv = NULL, .cb = NULL, .cb_param = NULL, .port = 0, .bit = 0};

#ifndef CONFIG_NO_INTERRUPTS
__STATIC_INLINE void gpio_interrupt_edge_config(uint8_t port_id, uint8_t bit) {
    uint8_t bitval1, bitval2;
    bitval2 = bitval1 = (GPIO->pdr[port_id] >> bit) & 0x01;

    /*                _______ <-----High Level
     RisingEdge->    /       \  <---Falling Edge
     LowLevel-->____/         \______
     Interrupt mode bothEdge -> interrupt on rising & falling
     edges.
    */
    do {
        bitval1 = bitval2;
        if (bitval1) {
            exiu_interrupt_conf(INTR_GPIO, INTR_FALLINGEDGE);
        } else {
            exiu_interrupt_conf(INTR_GPIO, INTR_RISINGEDGE);
        }
        /* check bit value before and after interrupt detection, state can change*/
        bitval2 = (GPIO->pdr[gpio.port] >> gpio.bit) & 0x01;
    } while (bitval1 != bitval2);
}

__STATIC_INLINE void gpio_interrupt_config(uint8_t port_id, uint8_t bit, gpio_config_flag_t config,
                                           gpio_callback cb, void* cb_param) {
    intr_detection_method_t detection_method;
    gpio_intr_mode_t intr_mode = config.flags.intr_mode;
    gpio.cb = cb;
    gpio.cb_param = cb_param;
    gpio.port = port_id;
    gpio.bit = bit;
    gpio.intr_mode = intr_mode;
    // disable gpio interrupt at NVIC, EXIU
    intr_disable(INTR_GPIO);
    // clearing gpio interrupt mask on port = port_id, pin = bit
    SCRST->GPIO2EXTINT |= (0x1 << (port_id * SOC_GPIO_PER_PORT + bit));

    /*gpio interrupt trigger configuration*/
    /* Incase of the both edge trigger mode configure as rising edge here.
    Falling edge configuration is taken care as part of the interrupt handling*/
    if (GPIO_INTR_BOTH_EDGE == intr_mode) {
        gpio_interrupt_edge_config(port_id, bit);
    } else {
        if (GPIO_INTR_LOWLVL == config.flags.intr_mode) {
            detection_method = INTR_LOWLEVEL;
        } else if (GPIO_INTR_HIGHLVL == config.flags.intr_mode) {
            detection_method = INTR_HIGHLEVEL;
        } else if (GPIO_INTR_RISING_EDGE == config.flags.intr_mode) {
            detection_method = INTR_RISINGEDGE;
        } else {
            detection_method = INTR_FALLINGEDGE;
        }
        exiu_interrupt_conf(INTR_GPIO, detection_method);
    }
    intr_clear_pending(INTR_GPIO);
    // enable interrupt
    intr_enable(INTR_GPIO);
}
#endif

__STATIC_INLINE void gpio_clear_config(uint8_t port_id, uint8_t bit) {
    // clearing gpio state
    gpio.cb = NULL;
    gpio.cb_param = NULL;
    gpio.port = 0;
    gpio.bit = 0;
    gpio.intr_mode = 0;

#ifndef CONFIG_NO_INTERRUPTS
    // clear pending gpio interrupts
    intr_clear_pending(INTR_GPIO);
    // disable gpio interrupt at NVIC, EXIU
    intr_disable(INTR_GPIO);
#endif
    // clearing gpio interrupt mask on port = port_id, pin = bit
    SCRST->GPIO2EXTINT &= ~(0x1 << (port_id * SOC_GPIO_PER_PORT + bit));
}

#ifndef CONFIG_NO_INTERRUPTS
void gpio_intr_handler() {
    uint8_t bitval1, bitval2;

    if (NULL != gpio.cb) {
        bitval2 = bitval1 = (GPIO->pdr[gpio.port] >> gpio.bit) & 0x01;

        if (GPIO_INTR_BOTH_EDGE == gpio.intr_mode) {
            gpio_interrupt_edge_config(gpio.port, gpio.bit);
        }
        /*Notify gpio value to the registrar using the callback function */
        gpio.cb(gpio.drv, bitval2, gpio.cb_param);
    }
    intr_clear_pending(INTR_GPIO);
}
#endif

static int akida_gpio_get_capabilities(struct gpio_driver* drv, gpio_cap_t* caps) {
    (void)*drv;

    caps->num_ports = SOC_GPIO_PORTS;
    caps->port_width = SOC_GPIO_PER_PORT;

    return SUCCESS;
}

static int akida_gpio_set_config(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t dir,
                                 gpio_config_flag_t config, gpio_callback cb, void* cb_param) {
    uint32_t port_mask = (0x1 << (port_id * SOC_GPIO_PER_PORT + bit));

    gpio.drv = drv;

    if (((gpio.port == port_id) && (gpio.bit == bit)) && (NULL != gpio.cb)) {
        gpio_clear_config(port_id, bit);
    }

    if (GPIO_DIR_IN == dir) {
        if ((NULL != gpio.cb) && (NULL != cb)) {
            /*Multiple gpio callback registration is not supported due to HW limitation*/
            return E_NORESOURCEAVAILABLE;
        }

        GPIO->ddr[port_id] &= ~(0x1 << bit);
#ifndef CONFIG_NO_INTERRUPTS
        if (NULL == gpio.cb && cb) {
            gpio_interrupt_config(port_id, bit, config, cb, cb_param);
        }
#else
        (void)cb_param;
#endif
    } else if (GPIO_DIR_OUT == dir) {
        GPIO->ddr[port_id] |= (0x1 << bit);
        SCRST->GPIO2EXTINT &= ~port_mask;
    }

    if (GPIO_PULL_NO_PULL == config.flags.gpio_mode) {
        /* 0-No pull */
        SCRST->GPIOPDEN |= port_mask;
    } else {
        SCRST->GPIOPDEN &= ~port_mask;
        if (GPIO_PULL_PU == config.flags.gpio_mode) {
            /* 1-pullup */
            SCRST->GPIOPUEN |= port_mask;
        } else if (GPIO_PULL_PD == config.flags.gpio_mode) {
            /* 2-pulldown */
            SCRST->GPIOPUEN &= ~port_mask;
        }
    }

    return SUCCESS;
}

static int akida_gpio_read(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t* rval) {
    (void)*drv;

    *rval = (GPIO->pdr[port_id] >> bit) & 0x01;

    return SUCCESS;
}

static int akida_gpio_read_port(struct gpio_driver* drv, uint8_t port_id, uint32_t* rval) {
    (void)*drv;

    *rval = GPIO->pdr[port_id];

    return SUCCESS;
}

static int akida_gpio_write(struct gpio_driver* drv, uint8_t port_id, uint8_t bit, uint8_t value) {
    (void)*drv;

    if (value)
        GPIO->pdr[port_id] |= (1 << bit);
    else
        GPIO->pdr[port_id] &= ~(1 << bit);

    return SUCCESS;
}

static int akida_gpio_write_port(struct gpio_driver* drv, uint8_t port_id, uint32_t value) {
    (void)*drv;

    GPIO->pdr[port_id] = value;

    return SUCCESS;
}

static int akida_gpio_lld_init(struct driver_common* drv, void* priv) {
    (void)*drv;
    (void)priv;

    for (int i = 0; i < SOC_GPIO_PORTS; i++) {
        GPIO->ddr[i] = 0xff;
        GPIO->pfr[i] = 0x00;
    }

#ifndef CONFIG_NO_INTERRUPTS
    /*gpio interrupt vector configuration*/
    intr_set_vector(INTR_GPIO, gpio_intr_handler);
    // clearing gpio interrupt mask on port = port_id, pin = bit
    SCRST->GPIO2EXTINT = 0;
    intr_clear_pending(INTR_GPIO);
    intr_enable(INTR_GPIO);
#endif
    return SUCCESS;
}

struct gpio_ops gpio_lld_ops = {
    .get_capabilities = akida_gpio_get_capabilities,
    .set_config = akida_gpio_set_config,
    .read = akida_gpio_read,
    .read_port = akida_gpio_read_port,
    .write = akida_gpio_write,
    .write_port = akida_gpio_write_port,
};

DECLARE_GPIO(gpio_lld, &gpio_lld_ops, akida_gpio_lld_init, NULL);
