#include "common.h"
#include "i2c_bus.h"
#include "i2c_reg.h"
#include "systick.h"

#define I2C_FSR_CONF_DIV 20000000
#define I2C_FRQ_STANDARD_DIV 100000
#define I2C_FRQ_FAST_DIV 400000

#define I2C_BUS_CLK(clk) ((clk) / I2C_FSR_CONF_DIV + 1)

/* I2C FREQUENCY IN STANDARD MODE */
#define I2C_MASTER_CLK_STANDARD(clk) ((((((clk) / I2C_FRQ_STANDARD_DIV) - 2)) / 2))

/* I2C FREQUENCY IN  FAST MODE */
#define I2C_MASTER_CLK_FAST(clk) ((((((clk) / I2C_FRQ_FAST_DIV) - 2) * 2) / 3))

/* Configure the CS bits in CCR register in standard mode */
#define I2C_CCR_CS_STANDARD(clk) ((I2C_MASTER_CLK_STANDARD(clk) - 1) & I2C_CCR_CS_MASK)

/* Configure the CS bits in CSR register in standard mode*/
#define I2C_CSR_CS_STANDARD(clk) (((I2C_MASTER_CLK_STANDARD(clk) - 1) >> 5) & I2C_CSR_CS_MASK)

/* Configure the CS bits in CCR register in fast mode*/
#define I2C_CCR_CS_FAST(clk) ((I2C_MASTER_CLK_FAST(clk) - 1) & I2C_CCR_CS_MASK)

/* Configure the CS bits in CSR register in fast mode */
#define I2C_CSR_CS_FAST(clk) (((I2C_MASTER_CLK_FAST(clk) - 1) >> 5) & I2C_CSR_CS_MASK)

/*
 * WFI: Wait For Interrupt
 * FM : Fast Mode
 * SM : Standard Mode
 * Timeout in us
 */
#define I2C_WFI_FM_TIMEOUT 2
#define I2C_WFI_SM_TIMEOUT 10

#define I2C_WFI_TIMEOUT(X) (X == I2C_CLK_FAST ? I2C_WFI_FM_TIMEOUT : I2C_WFI_SM_TIMEOUT)

/*Wait For Interrupt Timeout*/
#define I2C_WAIT_INTERRUPT_TIMEOUT 1000

/*Max number of data byte*/
#define DATA_BUF_MAX 100

typedef enum { I2C_CLK_STANDARD, I2C_CLK_FAST } I2C_CLK_MODE_t;

typedef enum { I2C_TRX_TRANSFER, I2C_TRX_RECEIVE } I2C_TRX_MODE_t;

#define I2C_CLK_MODE I2C_CLK_STANDARD

uint32_t g_clk;

static inline void i2c_conf_freq(I2C_CLK_MODE_t mode, uint32_t clock) {
    unsigned int ccr, csr;

    if (I2C_CLK_FAST == mode) {
        ccr = I2C_CCR_CS_FAST(clock);
        csr = I2C_CSR_CS_FAST(clock);

        ccr |= I2C_CCR_FM_MASK;
    } else {
        csr = I2C_CSR_CS_STANDARD(clock);
        ccr = I2C_CCR_CS_STANDARD(clock);
    }

    /* Set Clock and enable, Set standard mode */
    REG_WRITE(I2C->CSR, csr);
    REG_WRITE(I2C->CCR, ccr | I2C_CCR_EN_MASK);
}

static inline void i2c_bus_stop(void);

static void i2c_clear_error(void) {
    /*clear IRQ (INT=0, BER=0), Interrupt Disable*/
    REG_WRITE(I2C->BCR, 0);

    i2c_conf_freq(I2C_CLK_MODE, g_clk);

    REG_WRITE(I2C->BCR, 0);
}

/***********Wait For Interrupt**********/
static inline int i2c_wait_for_intr() {
    unsigned int bcr;
    int timeout = I2C_WAIT_INTERRUPT_TIMEOUT;

    do {
        delay_us(I2C_WFI_TIMEOUT(I2C_CLK_MODE));

        bcr = (I2C->BCR);

        if (bcr & I2C_BCR_INT_MASK) {
            return SUCCESS;
        }
    } while (timeout--);

    if (REG_READ_MASK(I2C->BCR, I2C_BCR_BER_MASK)) {
        i2c_clear_error();
    } else {
        i2c_bus_stop();
    }

    return EFAILURE;
}

static int i2c_start(uint8_t addr, I2C_TRX_MODE_t mode, bool restart) {
    int status = 0;

    if (I2C_TRX_RECEIVE == mode) {
        REG_WRITE(I2C->DAR, (addr << 1) | 1);
    } else {
        REG_WRITE(I2C->DAR, (addr << 1));
    }

    if (false == restart) {
        if (REG_READ_MASK(I2C->BSR, I2C_BSR_BB_MASK) &&
            !REG_READ_MASK(I2C->BCR, I2C_BCR_MSS_MASK)) {
            I2C_LOGI("%d Bus Busy bsr ", __LINE__);
            return EBUSY;
        }

        if (REG_READ_MASK(I2C->BSR, I2C_BSR_BB_MASK)) { /* Bus is busy */
            I2C_LOGI("Continuous Start");
            REG_WRITE_MASK(I2C->BCR, I2C_BCR_SCC_MASK, I2C_BCR_SCC_MASK);

        } else {
            if (REG_READ_MASK(I2C->BCR, I2C_BCR_MSS_MASK)) {
                I2C_LOGI("not in master mode");
                return EBAD_STATE;
            }

            /* Start Condition + Enable Interrupts */
            REG_WRITE(I2C->BCR, (I2C_BCR_BEIE_MASK | I2C_BCR_GCAA_MASK | I2C_BCR_INTE_MASK |
                                 I2C_BCR_MSS_MASK)); /*BCR = 0x46*/
        }
    } else {
        /* Start Condition + Enable Interrupts */
        REG_WRITE_MASK(I2C->BCR, I2C_BCR_SCC_MASK, I2C_BCR_SCC_MASK); /*BCR = 0x46*/
    }

    if ((status = i2c_wait_for_intr())) {
        return status;
    }

    if (!REG_READ_MASK(I2C->BSR, I2C_BSR_BB_MASK)) {
        status = EFAILURE;
    } else if (REG_READ_MASK(I2C->BSR, I2C_BSR_LRB_MASK)) {
        status = EFAILURE;
    }

    return status;
}

static inline void i2c_bus_stop(void) {
    REG_WRITE(I2C->BCR, 0);
    while (REG_READ_MASK(I2C->BSR, I2C_BSR_BB_MASK))
        ;
}

static int i2c_bus_write(uint8_t addr, uint8_t* buf, int len, bool restart) {
    int indx;
    int status = 0;

    if ((status = i2c_start(addr, I2C_TRX_TRANSFER, false))) {
        /*if the i2c_start fail, issue a stop signal to gracefully exit the protocol*/
        i2c_bus_stop();
        return status;
    }

    for (indx = 0; indx < len; indx++) {
        REG_WRITE(I2C->DAR, buf[indx]);

        REG_CLEAR_BIT_FIELD(I2C->BCR, I2C_BCR_INT_MASK);

        if ((status = i2c_wait_for_intr())) {
            status = EFAILURE;
            break;
        }

        if (REG_READ_MASK(I2C->BSR, I2C_BSR_LRB_MASK)) {
            status = EFAILURE;
            break;
        }
    }

    if (false == restart)
        i2c_bus_stop();

    return status == 0 ? len : status;
}

static int i2c_bus_read(uint8_t addr, uint8_t* buf, int len, bool restart) {
    unsigned int bcr;
    int indx = 0;
    int status = 0;

    if ((status = i2c_start(addr, I2C_TRX_RECEIVE, restart))) {
        i2c_bus_stop();
        return status;
    }

    for (indx = 0; indx < len; indx++) {
        bcr = (I2C->BCR);

        if (indx == len - 1) {
            bcr &= ~(I2C_BCR_INT_MASK | I2C_BCR_ACK_MASK);
        } else {
            bcr &= ~(I2C_BCR_INT_MASK);
            bcr |= I2C_BCR_ACK_MASK;
        }

        REG_WRITE(I2C->BCR, bcr);

        if ((status = i2c_wait_for_intr())) {
            i2c_bus_stop();
            return status;
        }

        buf[indx] = REG_READ(I2C->DAR);
    }

    i2c_bus_stop();
    return indx;
}

static int akida_i2c_read_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len) {
    if (NULL == buf) {
        return ENULL_POINTER;
    }

    buf[0] = reg;

    if (i2c_bus_write(drv->address, buf, 1, true) != 1) {
        return EFAILURE;
    }

    return i2c_bus_read(drv->address, buf, len, true);
}

/**********i2c Multiple data send***********/
static int akida_i2c_write_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len) {
    uint8_t data_buf[DATA_BUF_MAX + 1];

    if (NULL == buf) {
        return ENULL_POINTER;
    }

    data_buf[0] = reg;

    memcpy(data_buf + 1, buf, len);

    int ret = i2c_bus_write(drv->address, data_buf, len + 1, false);

    /* i2c_write_reg returns the number of bytes written minus the reg address */
    return ret <= 0 ? ret : ret - 1;
}

static int akida_i2c_write(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb,
                           void* cb_param) {
    (void)cb;
    (void)cb_param;

    if (NULL == buf) {
        return ENULL_POINTER;
    }

    return i2c_bus_write(drv->address, buf, len, false);
}

static int akida_i2c_read(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb,
                          void* cb_param) {
    (void)cb;
    (void)cb_param;

    if (NULL == buf) {
        return ENULL_POINTER;
    }

    return i2c_bus_read(drv->address, buf, len, false);
}

static int akida_i2c_attach(struct i2c_bus_driver* drv, struct i2c_device* dev) {
    (void)drv;
    (void)dev;
    return 0;
}

static int akida_i2c_detach(struct i2c_bus_driver* drv, struct i2c_device* dev) {
    (void)drv;
    (void)dev;
    return 0;
}

static int akida_i2c_init(struct driver_common* drv, void* priv) {
    (void)drv;
    (void)priv;

    /* Configure own Address*/
    REG_WRITE(I2C->ADR, 0);

    /*Configure I2C module CLK frequency*/
    uint32_t clock = clock_get_freq(DRIVER_GET_CLK_ID(drv));

    g_clk = clock;

    I2C_LOGI("I2C clock freq %d", clock);

    REG_WRITE(I2C->FSR, I2C_BUS_CLK(clock));

    i2c_conf_freq(I2C_CLK_MODE, clock);

    REG_WRITE(I2C->BCR2, 0);

    /* clear and disable Interrupt(INT=0, BER=0) */
    REG_WRITE(I2C->BCR, 0);

    return SUCCESS;
}

struct i2c_bus_ops akida_i2c_ops = {.read_reg = akida_i2c_read_reg,
                                    .write_reg = akida_i2c_write_reg,
                                    .read = akida_i2c_read,
                                    .write = akida_i2c_write,
                                    .attach = akida_i2c_attach,
                                    .detach = akida_i2c_detach};

DECLARE_I2C_BUS(akida_i2c, &akida_i2c_ops, akida_i2c_init, NULL, CRG11_CLK_ID_I2C);
