#include "soc.h"

#include "i2c_bus.h"
#include "i3c_reg.h"
#include "systick.h"

/*Wait For Interrupt*/
#define I3C_WAIT_INTERRUPT_TIMEOUT 10

#define I3C_WFI_DELAY 100 /*Timeout in us*/

/*I2C Bus Frequency*/
#define I2C_OPERATING_FRQ 400000
/*I3C Bus Frequency*/
#define I3C_OPERATING_FRQ 12500000

#define I3C_DEVS_CTRL_DEV_CLR(x) (1 << x)

#define I3C_MASTER_ID 0x41
#define I3C_SELF_INDX 0
#define I3C_MAX_DEVICES 7
#define I3C_SLAVE_INDX 1

#define I3C_MST_ICR_ALL 0x3DFFF

typedef enum {
    I3C_BURST_INC,
    I3C_SINGLE_INC,
    I3C_SINGLE_STATIC,
    I3C_BURST,
} i3c_xmit_mode_t;

typedef struct {
    uint8_t cmd_id;
    /*Below commented parameters are used for I3C bus communication*/
#if 0
    bool is_ddr;
    bool is_ccc;
    bool is_bch;
#endif
    bool rsbc;
    i3c_xmit_mode_t mode;
    bool is_read_cmd;
    bool is_cmd_code;
    uint16_t code;
} i3c_cmd_conf_t;

typedef enum {
    SL_OP_MODE_FM,
    SL_OP_MODE_FM_PLUS,
} i3c_sl_op_mode_t;

typedef union {
    uint8_t dcr;
    struct {
        uint8_t code;
        uint8_t index;
        i3c_sl_op_mode_t mode;
    } lvr;
} i3c_dcr_lvr_t;

typedef struct {
    uint8_t bcr;
    bool is_i3c;
    uint64_t pid;
    i3c_dcr_lvr_t dcr_lvr;
} i3c_dev_conf_t;

static struct i3c_fifo_depth {
    int tx;
    int rx;
    int tx_limit;
} fifo_depth;

static void write_tx_fifo(uint8_t* buf, int nobyte) {
    while (nobyte) {
        int data = 0;

        if ((nobyte / 4)) {
            data = ((buf[3] << 24) | (buf[2] << 16) | (buf[1] << 8) | buf[0]);
            buf += 4;
            nobyte -= 4;
        } else if ((nobyte % 4) == 3) {
            data = ((buf[2] << 16) | (buf[1] << 8) | buf[0]);
            buf += 3;
            nobyte -= 3;
        } else if ((nobyte % 4) == 2) {
            data = ((buf[1] << 8) | buf[0]);
            buf += 2;
            nobyte -= 2;
        } else {
            data = buf[0];
            buf += 1;
            nobyte -= 1;
        }

        REG_WRITE(I3C->tx_fifo, data);
    }
}

static int read_rx_fifo(uint8_t* buf, int len) {
    uint32_t data = 0;
    int d_indx = 0;
    for (int indx = 0; indx < len; indx += 4) {
        data = I3C->rx_fifo;

        if (REG_READ_MASK(I3C->mst_intr.isr, I3C_MST_ISR_RX_UNF_MASK)) {
            REG_WRITE_MASK(I3C->mst_intr.icr, I3C_MST_ISR_RX_UNF_MASK, I3C_MST_ISR_RX_UNF_MASK);
            break;
        }

        do {
            buf[d_indx] = ((data >> ((d_indx & 0x03) * 8)) & 0xFF);
        } while (++d_indx < (indx + 4) && d_indx < len);
    }

    return d_indx;
}

static int i3c_wait_for_intr(uint8_t* buf, int len, bool is_read_cmd) {
    uint64_t timeout = get_time_ms() + I3C_WAIT_INTERRUPT_TIMEOUT;
    int written = 0;
    int nobyte = 0;
    do {
        if (REG_READ_MASK(I3C->mst_intr.isr, I3C_MST_ISR_TX_THR_MASK)) {
            if (len) {
                if (len > fifo_depth.tx_limit) {
                    nobyte = (fifo_depth.tx / 2);
                } else {
                    nobyte = len;
                }
                write_tx_fifo(buf, nobyte);
                written += nobyte;
                len = (len - nobyte);
                buf += nobyte;
                timeout = get_time_ms() + I3C_WAIT_INTERRUPT_TIMEOUT;
            }

            REG_WRITE_MASK(I3C->mst_intr.icr, I3C_MST_ISR_TX_THR_MASK, I3C_MST_ISR_TX_THR_MASK);
        } else if (REG_READ_MASK(I3C->mst_intr.isr, I3C_MST_ISR_CMDD_EMP_MASK)) {
            REG_WRITE_MASK(I3C->mst_intr.icr, I3C_MST_ISR_CMDD_EMP_MASK, I3C_MST_ISR_CMDD_EMP_MASK);
            if (true == is_read_cmd) {
                nobyte += read_rx_fifo(buf + nobyte, (len - nobyte));
                return nobyte;
            }
            return written;
        } else if (REG_READ_MASK(I3C->mst_intr.isr, I3C_MST_ISR_RX_THR_MASK)) {
            if ((len - nobyte) > fifo_depth.rx) {
                nobyte += read_rx_fifo(buf + nobyte, (len - nobyte));
                timeout = get_time_ms() + I3C_WAIT_INTERRUPT_TIMEOUT;
            }
            REG_WRITE_MASK(I3C->mst_intr.icr, I3C_MST_ISR_RX_THR_MASK, I3C_MST_ISR_RX_THR_MASK);
        }

    } while (get_time_ms() < timeout);

    I2C_LOGE("%s DEVICE_ERROR \n", __FUNCTION__);
    I2C_LOGI("%s - I3C->mst_intr.isr: 0x%x\n", __FUNCTION__, I3C->mst_intr.isr);

    /*Clear all the interrupts*/
    REG_WRITE(I3C->mst_intr.icr, I3C_MST_ICR_ALL);
    return EFAILURE;
}
#ifdef CONFIG_I3C_SW_DEV_TABLE
static void i3c_select_slave(uint32_t addr) {
    static uint32_t i2c_current_addr = 0xffff;

    if (i2c_current_addr == addr) {
        /* No need to configure if accessing the same address again. */
        return;
    }

    i2c_current_addr = addr;

    /*I2C slave configuration*/
    REG_WRITE(I3C->device_id[I3C_SLAVE_INDX].rr0, (addr << 1));

    if (((addr & 0x3ff) >> 7) & 0x7) {
        REG_WRITE_MASK(I3C->device_id[I3C_SLAVE_INDX].rr0,
                       I3C_DEV_IDX_RR0_LVR_SA_MSB_SET(((addr & 0x3ff) >> 7) & 0x7) |
                           I3C_DEV_IDX_RR0_LVR_EXT_ADDR_MASK,
                       I3C_DEV_IDX_RR0_LVR_SA_MSB_MASK | I3C_DEV_IDX_RR0_LVR_EXT_ADDR_MASK);
    }

    REG_WRITE_MASK(I3C->devs_ctrl, I3C_DEVS_CTRL_DEV_CLR(I3C_SLAVE_INDX),
                   I3C_DEVS_CTRL_DEV_CLR(I3C_SLAVE_INDX));
}
#else
int i3c_slave_config(uint32_t addr, i3c_dev_conf_t* i3c_conf) {
    uint32_t reg_r2 = 0;
    int dev_indx;

    for (dev_indx = 1; dev_indx <= I3C_MAX_DEVICES; dev_indx++) {
        uint32_t rr0 = REG_READ(I3C->device_id[dev_indx].rr0);
        I2C_LOGD("dev %d %08x", dev_indx, rr0);
        if ((rr0 & 0xff) == 0) {
            break; /* found free item */
        }
    }

    if (dev_indx > I3C_MAX_DEVICES) {
        return EFAILURE;
    }
    /*I2C slave configuration*/
    REG_WRITE(I3C->device_id[dev_indx].rr0, (addr << 1));

    if (((addr & 0x3ff) >> 7) & 0x7) {
        REG_WRITE_MASK(I3C->device_id[dev_indx].rr0,
                       I3C_DEV_IDX_RR0_LVR_SA_MSB_SET(((addr & 0x3ff) >> 7) & 0x7) |
                           I3C_DEV_IDX_RR0_LVR_EXT_ADDR_MASK,
                       I3C_DEV_IDX_RR0_LVR_SA_MSB_MASK | I3C_DEV_IDX_RR0_LVR_EXT_ADDR_MASK);
    }

    if (i3c_conf) {
        if (i3c_conf->pid) {
            REG_WRITE(I3C->device_id[dev_indx].rr1, ((i3c_conf->pid >> 16) & 0xffffffff));
            reg_r2 |= I3C_DEV_IDX_RR2_PID_LSB_SET((i3c_conf->pid & 0xffff));
        }

        if (i3c_conf->is_i3c) {
            if (i3c_conf->bcr) {
                reg_r2 |= I3C_DEV_IDX_RR2_BCR_SET(i3c_conf->bcr);
            }
            if (i3c_conf->dcr_lvr.dcr) {
                reg_r2 |= I3C_DEV_IDX_RR2_DCR_LVR_SET(i3c_conf->dcr_lvr.dcr);
            }
        } else {
            if (i3c_conf->dcr_lvr.lvr.code) {
                reg_r2 |= (i3c_conf->dcr_lvr.lvr.code << 4);
            }
            if (i3c_conf->dcr_lvr.lvr.mode) {
                reg_r2 |= ((i3c_conf->dcr_lvr.lvr.mode << 3) & 0x8);
            }
            if (i3c_conf->dcr_lvr.lvr.index) {
                reg_r2 |= (i3c_conf->dcr_lvr.lvr.index & 0x7);
            }
        }
        if (reg_r2) {
            REG_WRITE_MASK(I3C->device_id[dev_indx].rr2, reg_r2, reg_r2);
        }
    }

    REG_WRITE_MASK(I3C->devs_ctrl, I3C_DEVS_CTRL_DEV_CLR(dev_indx),
                   I3C_DEVS_CTRL_DEV_CLR(dev_indx));

    I2C_LOGI("dev_indx : %d slave_addr : 0x%x", dev_indx, addr);

    return SUCCESS;
}

int i3c_slave_deconfig(uint32_t addr) {
    for (int i = 1; i <= I3C_MAX_DEVICES; i++) {
        uint32_t rr0 = REG_READ(I3C->device_id[i].rr0);
        if ((rr0 & 0xff) == (addr << 1)) {
            REG_WRITE_MASK(I3C->devs_ctrl, I3C_DEVS_CTRL_DEV_CLR(i), I3C_DEVS_CTRL_DEV_CLR(i));
            REG_WRITE(I3C->device_id[i].rr0, 0);
            return 0;
        }
    }

    return EFAILURE;
}
#endif

static int i3c_transfer(uint16_t addr, uint8_t* buf, int nobyte, i3c_cmd_conf_t cmd_conf) {
    uint32_t reg = 0;
    int sent_bytes = 0;

    REG_WRITE(I3C->mst_intr.icr, I3C_MST_ICR_ALL);

    if (cmd_conf.is_cmd_code) {
        /*validation of code is required*/
        REG_WRITE(I3C->cmd1_fifo,
                  (cmd_conf.code & 0XFFFF) | I3C_CMD1_FIFO_CMD_ID_SET(cmd_conf.cmd_id));

        if (cmd_conf.code & 0xFF00) {
            reg |= I3C_CMD0_FIFO_SBCA_MASK;
        }
    } else {
        REG_WRITE(I3C->cmd1_fifo, I3C_CMD1_FIFO_CMD_ID_SET(cmd_conf.cmd_id));
    }

    if (addr & 0x380) {
        reg |= I3C_CMD0_FIFO_IS_10B_MASK;
    }

#if 0
    /*Below commented code is not tested and it is used for i3c bus communication and it is helpful during I3C bus driver development*/
    if (true == cmd_conf.is_ddr) {
        reg |= I3C_CMD0_FIFO_IS_DDR_SET(1);
    }

    if (true == cmd_conf.is_ccc) {
        reg |= I3C_CMD0_FIFO_IS_CCC_SET(1);
    }

    if (true == cmd_conf.is_bch) {
        reg |= I3C_CMD0_FIFO_BCH_SET(1);
    }
#endif

    if (true == cmd_conf.rsbc) {
        reg |= I3C_CMD0_FIFO_SBCA_SET(1);
    }

    reg |= I3C_CMD0_FIFO_XMIT_MODE_SET(cmd_conf.mode) |
           (I3C_CMD0_FIFO_PL_LEN_SET(nobyte) | I3C_CMD0_FIFO_DEV_ADDR_SET(addr));

    if (cmd_conf.is_read_cmd) {
        reg |= I3C_CMD0_FIFO_RNW_READ_XFER;
    }

    if (buf && false == cmd_conf.is_read_cmd) {
        int data_len = nobyte;

        /*data more than fifo depth*/
        if (data_len > fifo_depth.tx) {
            data_len = fifo_depth.tx;
        }

        write_tx_fifo(buf, data_len);

        nobyte -= data_len;

        buf += data_len;
        sent_bytes += data_len;
    }

    REG_WRITE(I3C->cmd0_fifo, reg);

    REG_WRITE_MASK(I3C->ctrl, I3C_CTRL_MCS_MASK, I3C_CTRL_MCS_MASK);

    int status = i3c_wait_for_intr(buf, nobyte, cmd_conf.is_read_cmd);

    int error_code = I3C_CMDR_ERROR_GET(I3C->cmdr);

    if (error_code) {
        /* return failure for non zero value in cmdr register*/
        status = EFAILURE;
    }

    if (status > 0)
        sent_bytes += status;

    return status < 0 ? status : sent_bytes;
}

static int i3c_write(uint8_t addr, uint8_t* buf, int nobyte, i3c_cmd_conf_t cmd_conf) {
    int status = SUCCESS;

    if (nobyte) {
        status = i3c_transfer(addr, buf, nobyte, cmd_conf);
    } else {
        status = i3c_transfer(addr, NULL, nobyte, cmd_conf);
    }

    return status;
}

static int akida_i3c_read_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len) {
    i3c_cmd_conf_t cmd_conf;

#ifdef CONFIG_I3C_SW_DEV_TABLE
    i3c_select_slave(drv->address);
#endif

    memset(&cmd_conf, 0, sizeof(i3c_cmd_conf_t));

    cmd_conf.cmd_id = 0xF; /*As there is only one slave the cmd_id is config to 0xF*/
    cmd_conf.rsbc = false;
    cmd_conf.mode = I3C_BURST_INC;
    cmd_conf.is_cmd_code = true;
    cmd_conf.code = reg;
    cmd_conf.is_read_cmd = true;

    return i3c_transfer(drv->address, buf, len, cmd_conf);
}

static int akida_i3c_write_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len) {
    int status = SUCCESS;
    i3c_cmd_conf_t cmd_conf;

#ifdef CONFIG_I3C_SW_DEV_TABLE
    i3c_select_slave(drv->address);
#endif

    memset(&cmd_conf, 0, sizeof(i3c_cmd_conf_t));

    cmd_conf.cmd_id = 0xF;
    cmd_conf.rsbc = false;
    cmd_conf.mode = I3C_BURST_INC;
    cmd_conf.is_read_cmd = false;
    cmd_conf.is_cmd_code = true;
    cmd_conf.code = reg;

    if (buf) {
        status = i3c_write(drv->address, buf, len, cmd_conf);

    } else {
        status = i3c_transfer(drv->address, NULL, 0, cmd_conf);
    }

    return status;
}

static int akida_i3c_write(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb,
                           void* cb_param) {
    (void)cb;
    (void)cb_param;

    i3c_cmd_conf_t cmd_conf;
#ifdef CONFIG_I3C_SW_DEV_TABLE
    i3c_select_slave(drv->address);
#endif

    memset(&cmd_conf, 0, sizeof(i3c_cmd_conf_t));

    cmd_conf.cmd_id = 0xF;
    cmd_conf.rsbc = false;
    cmd_conf.mode = I3C_BURST;
    cmd_conf.is_read_cmd = false;
    cmd_conf.is_cmd_code = false;

    return i3c_write(drv->address, buf, len, cmd_conf);
}

static int akida_i3c_read(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb,
                          void* cb_param) {
    (void)cb;
    (void)cb_param;

    i3c_cmd_conf_t cmd_conf;

#ifdef CONFIG_I3C_SW_DEV_TABLE
    i3c_select_slave(drv->address);
#endif
    memset(&cmd_conf, 0, sizeof(i3c_cmd_conf_t));

    cmd_conf.cmd_id = 0xF;
    cmd_conf.rsbc = false;
    cmd_conf.mode = I3C_BURST;
    cmd_conf.is_read_cmd = true;
    cmd_conf.is_cmd_code = false;

    return i3c_transfer(drv->address, buf, len, cmd_conf);
}

static int akida_i3c_attach(struct i2c_bus_driver* drv, struct i2c_device* dev) {
    (void)drv;
#ifdef CONFIG_I3C_SW_DEV_TABLE
    (void)dev;
    return 0;
#else
    return i3c_slave_config(dev->address, NULL);
#endif
}

static int akida_i3c_detach(struct i2c_bus_driver* drv, struct i2c_device* dev) {
    (void)drv;
#ifdef CONFIG_I3C_SW_DEV_TABLE
    (void)dev;
    return 0;
#else
    return i3c_slave_deconfig(dev->address);
#endif
}

static int akida_i3c_init(struct driver_common* drv, void* priv) {
    (void)priv;

    uint32_t tx_thr_val = 0, rx_thr_val = 0;

    uint32_t clock = clock_get_freq(DRIVER_GET_CLK_ID(drv));

    clock_acquire(DRIVER_GET_CLK_ID(drv));

    uint32_t i2c_prescaler = ((clock / (I2C_OPERATING_FRQ * 5)) - 1);

    uint32_t i3c_prescaler = ((clock / (I3C_OPERATING_FRQ * 4)) - 1);

    uint32_t i3c_od_scl = (((220 * (clock / 10000000) / 100) / (i3c_prescaler + 1)) - 2);

    /*I3C Controller disable*/
    REG_CLEAR_BIT_FIELD(I3C->ctrl, I3C_CTRL_DEV_EN_MASK);

    REG_WRITE_MASK(I3C->flush_ctrl,
                   (I3C_FLUSH_CTRL_TX_FIFO_FLUSH_MASK | I3C_FLUSH_CTRL_RX_FIFO_FLUSH_MASK),
                   (I3C_FLUSH_CTRL_TX_FIFO_FLUSH_MASK | I3C_FLUSH_CTRL_RX_FIFO_FLUSH_MASK));

    /*I2C and I3C SCL clock configuration*/
    REG_WRITE(I3C->prescl_ctrl0,
              I3C_PRESCL_CTRL0_I2C_SET(i2c_prescaler) | I3C_PRESCL_CTRL0_I3C_SET(i3c_prescaler));

    /*od_low 1 && pp_low 1*/
    REG_WRITE(I3C->prescl_ctrl1,
              I3C_PRESCL_CTRL1_PP_LOW_SET(i3c_od_scl) | I3C_PRESCL_CTRL1_OD_LOW_SET(i3c_od_scl));

    // Enable cmdd_emp interrupt
    REG_WRITE_MASK(I3C->mst_intr.ier, I3C_MST_IER_CMDD_EMP_MASK, I3C_MST_IER_CMDD_EMP_MASK);
    REG_WRITE_MASK(I3C->mst_intr.ier, I3C_MST_IER_TX_OVF_MASK, I3C_MST_IER_TX_OVF_MASK);

    REG_WRITE_MASK(I3C->ctrl, I3C_CTRL_BUS_MODE_BUS_MODE_MIXED_SLOW | I3C_CTRL_MCS_EN_MASK,
                   I3C_CTRL_BUS_MODE_MASK | I3C_CTRL_MCS_EN_MASK);

    REG_WRITE(I3C->device_id[I3C_SELF_INDX].rr0, I3C_MASTER_ID);

    fifo_depth.rx = 1 << (3 + I3C_CONF_STATUS1_RX_MEM_DEPTH_GET(I3C->conf_status1));

    fifo_depth.tx = 1 << (3 + I3C_CONF_STATUS1_TX_MEM_DEPTH_GET(I3C->conf_status1));
    fifo_depth.tx_limit = (3 * fifo_depth.tx) / 4;

    tx_thr_val = (fifo_depth.tx / 16);

    rx_thr_val = ((3 * fifo_depth.rx) / 16);

    REG_WRITE(I3C->tx_rx_thr_ctrl, I3C_TX_RX_THR_CTRL_RX_THR_SET(rx_thr_val) |
                                       I3C_TX_RX_THR_CTRL_TX_THR_SET(tx_thr_val));

    I2C_LOGD("%s - rx_thr: 0x%x conf_status1 0x%x \n fifo_depth.rx 0x%x tx_fifo_depth 0x%x \n",
             __FUNCTION__, rx_thr_val, I3C->conf_status1, fifo_depth.rx, fifo_depth.tx);
    /*I3C Controller Enable*/
    REG_WRITE_MASK(I3C->ctrl, I3C_CTRL_DEV_EN_MASK, I3C_CTRL_DEV_EN_MASK);

    for (int i = 1; i <= I3C_MAX_DEVICES; i++) {
        REG_WRITE(I3C->device_id[i].rr0, 0);
    }

    I2C_LOGI("MST_STATUS0: 0x%x\n", I3C->mst_status0);

    return SUCCESS;
}

struct i2c_bus_ops akida_i3c_ops = {.read_reg = akida_i3c_read_reg,
                                    .write_reg = akida_i3c_write_reg,
                                    .read = akida_i3c_read,
                                    .write = akida_i3c_write,
                                    .attach = akida_i3c_attach,
                                    .detach = akida_i3c_detach};

DECLARE_I2C_BUS(akida_i3c, &akida_i3c_ops, akida_i3c_init, NULL, CRG11_CLK_ID_I3C);
