/**
 * @file     exiu.c
 * @brief    The APIs defines in exiu.c are private APIs which are called by interrupt public APIs
 *           declared in interrupt.h
 */
#include "exiu.h"
#include "common.h"
#include "exiu_reg.h"
#include "nsoc_v1_interrupts.h"

#define EXIU_INTR_OFFSET INTR_AKIDACORE
#define EXIU_NO_INTR (16)

#define IS_EXIU_INTR(_INTR_) (((_INTR_ - EXIU_INTR_OFFSET) < EXIU_NO_INTR))

#define EXIU_GET_INT_BIT(_INTR_) (1 << (_INTR_ - EXIU_INTR_OFFSET))

__STATIC_INLINE void exiu_set_detect_method(const uint8_t intr,
                                            intr_detection_method_t detection_method) {
    switch (detection_method) {
        case INTR_LOWLEVEL:
            REG_CLEAR_BIT_FIELD(EXIU->EIEDG, EXIU_GET_INT_BIT(intr));
            REG_CLEAR_BIT_FIELD(EXIU->EILVL, EXIU_GET_INT_BIT(intr));
            break;
        case INTR_HIGHLEVEL:
            REG_CLEAR_BIT_FIELD(EXIU->EIEDG, EXIU_GET_INT_BIT(intr));
            REG_WRITE_MASK(EXIU->EILVL, EXIU_GET_INT_BIT(intr), EXIU_GET_INT_BIT(intr));
            break;
        case INTR_FALLINGEDGE:
            REG_CLEAR_BIT_FIELD(EXIU->EILVL, EXIU_GET_INT_BIT(intr));
            REG_WRITE_MASK(EXIU->EIEDG, EXIU_GET_INT_BIT(intr), EXIU_GET_INT_BIT(intr));
            break;
        case INTR_RISINGEDGE:
            REG_WRITE_MASK(EXIU->EILVL, EXIU_GET_INT_BIT(intr), EXIU_GET_INT_BIT(intr));
            REG_WRITE_MASK(EXIU->EIEDG, EXIU_GET_INT_BIT(intr), EXIU_GET_INT_BIT(intr));
            break;
    }
}

int exiu_interrupt_conf(const uint8_t intr, intr_detection_method_t detection_method) {
    if (!IS_EXIU_INTR(intr)) {
        return -1;
    }

    /* external interrupt is masked*/
    REG_WRITE_MASK(EXIU->EIMASK, EXIU_GET_INT_BIT(intr), EXIU_GET_INT_BIT(intr));

    /*interrupt cleared*/
    REG_WRITE_MASK(EXIU->EIREQCLR, EXIU_GET_INT_BIT(intr), EXIU_GET_INT_BIT(intr));

    /*interrupt sources as external interrupt*/
    REG_CLEAR_BIT_FIELD(EXIU->EISRCSEL, EXIU_GET_INT_BIT(intr));

    exiu_set_detect_method(intr, detection_method);

    /* external interrupt is unmasked*/
    REG_CLEAR_BIT_FIELD(EXIU->EIMASK, EXIU_GET_INT_BIT(intr));

    return 0;
}

int exiu_interrupt_clear(const uint8_t intr) {
    if (!IS_EXIU_INTR(intr)) {
        return -1;
    }

    REG_WRITE_MASK(EXIU->EIREQCLR, EXIU_GET_INT_BIT(intr), EXIU_GET_INT_BIT(intr));

    return 0;
}

int exiu_interrupt_enable(const uint8_t intr) {
    if (!IS_EXIU_INTR(intr)) {
        return -1;
    }

    REG_WRITE_MASK(EXIU->EIMASK, 0, EXIU_GET_INT_BIT(intr));

    return 0;
}

int exiu_interrupt_disable(const uint8_t intr) {
    if (!IS_EXIU_INTR(intr)) {
        return -1;
    }

    REG_WRITE_MASK(EXIU->EIMASK, 0, EXIU_GET_INT_BIT(intr));

    return 0;
}
