/**
 * @file     interrupt.c
 * @brief    The API functions declared in this header file are called by upper levels (which
 *           include interrupt.h) to configure/use the interrupt services.
 *
 * @note     This interrupt public APIs are callable from LLCS(ex. DMA) and LLDs only.
 *
 * @note     error numbers will be replaced with a macro once error.h is available with master
 *
 * @note     NVIC_* are CMSIS calls
 */

#include "interrupt.h"
#include <stddef.h>
#include "AKD1000.h"
#include "exiu.h"
#include "soc.h"

#define CHECK_INTERRUPT(_INTR_) ((_INTR_) > INTR_GPIO)

int intr_set_priority(int intr, uint8_t priority) {
    if (CHECK_INTERRUPT(intr)) {
        return -1;
    }

    NVIC_SetPriority((IRQn_Type)intr, priority);
    return 0;
}

int intr_get_priority(int intr) {
    if (CHECK_INTERRUPT(intr)) {
        return -1;
    }

    return NVIC_GetPriority((IRQn_Type)intr);
}

int intr_enable(int intr) {
    if (CHECK_INTERRUPT(intr)) {
        return -1;
    }

    exiu_interrupt_enable(intr);

    NVIC_EnableIRQ((IRQn_Type)intr);
    return 0;
}

bool intr_is_enabled(int intr) {
    if (CHECK_INTERRUPT(intr)) {
        return false;
    }

    if (NVIC_GetEnableIRQ((IRQn_Type)intr)) {
        return true;
    }

    return false;
}

int intr_disable(int intr) {
    if (CHECK_INTERRUPT(intr)) {
        return -1;
    }

    exiu_interrupt_disable(intr);

    NVIC_DisableIRQ((IRQn_Type)intr);

    return 0;
}

int intr_clear_pending(int intr) {
    if (CHECK_INTERRUPT(intr)) {
        return -1;
    }

    exiu_interrupt_clear(intr);

    NVIC_ClearPendingIRQ((IRQn_Type)intr);

    return 0;
}

int intr_set_pending(int intr) {
    if (CHECK_INTERRUPT(intr)) {
        return -1;
    }

    NVIC_SetPendingIRQ((IRQn_Type)intr);

    return 0;
}

int intr_set_vector(int intr, intr_handler_t handle) {
    if (CHECK_INTERRUPT(intr)) {
        return -1;
    }

    if (NULL == handle) {
        return -2;
    }

    NVIC_SetVector((IRQn_Type)intr, (uint32_t)handle);

    return 0;
}

intr_handler_t intr_get_vector(int intr) {
    intr_handler_t handle;

    if (CHECK_INTERRUPT(intr)) {
        return NULL;
    }

    handle = (intr_handler_t)NVIC_GetVector((IRQn_Type)intr);

    if (handle) {
        return handle;
    }

    return NULL;
}
