#include <stdint.h>
#include <stdlib.h>

#include "cli_cmd.h"
#include "nsocv1/soc.h"
#include "systick.h"

// **************************************************************************
// CLOCKS and RESET INIT
// **************************************************************************

static int crg11_clock_used[16] = {0};

int crg11_init(uint8_t speed) {
    uint8_t FBMODE_1200 = 24;  // m X 2 -> 48 *25Mhz = 1200 Mhz
    uint8_t PSMODE_1200 = 1;   // n DIV 2 1200/2 = 600 Mhz
                               // uint8_t   PLL_Bypass = 0;
    uint32_t delay = 0;
    uint32_t count = 0;
    int i = 0;
    int ret = 0;

    if (speed)
        PSMODE_1200 = 0;  // n DIV 1 1200/1 = 1200 Mhz
    else
        PSMODE_1200 = 1;  // n DIV 2 1200/2 = 600 Mhz

    // setup PLL bypass and FBMODE PSMODE
    CRG11->PLLICP = 2;

    CRG11->CRPLC = 0x010F0000 | (PSMODE_1200 << 8) | FBMODE_1200;

    // Wait for PLLRDY bit set
    do {
        count++;
        if (count > 0xB10) {
            ret = -2;
            goto out;
        }
    } while (!(CRG11->CRRDY & 1));

    // Remove PLLbypass
    CRG11->CRPLC = 0x000F0000 | (PSMODE_1200 << 8) | FBMODE_1200;  // CRPLR + no Bypass

    // Setup DIV MODEx
    CRG11->CRCDC = 0;
    static int divs[] = {3, 7, 7, 3, 23, 23, 23, 23, 3, 3, 7, 3, 47, 7, 7};
    for (i = 0; i < 15; i++) {
        if (speed)  // 1200
            CRG11->CLK[i].CRDM = divs[i];
        else
            CRG11->CLK[i].CRDM = divs[i] / 2;
    }

    // Assert DCHREQ
    CRG11->CRCDC = 1;

    // Poll for CEN running (NOT STOP)
    count = 0;
    int done;
    do {
        count++;
        if (count > 0x100) {
            ret = -2;
            goto out;
        }
        done = 1;
        for (i = 0; i < 15; i++) {
            if ((CRG11->CLK[i].CRLP >> 24 == 0))
                done = 0;
        }
    } while (!done);

    CRG11->DDRGC = 1;

    delay = 0x100;
    while (delay--)
        ;  // give it some time for Reset hardware to finish

out:
    return ret;
}

int clock_acquire(int clock_id) {
    crg11_clock_used[clock_id]++;
    return 0;
}

int clock_release(int clock_id) {
    crg11_clock_used[clock_id]--;
    return 0;
}

uint32_t clock_get_freq(int clock_id) {
    int div = CRG11->CLK[clock_id].CRDM & 0xff;

    return (2 * CONFIG_SYSTEM_CLOCK_FREQUENCY) / (div + 1);
}

int clock_set_div(int clock_id, int div) {
    // Setup DIV MODEx
    CRG11->CRCDC = 0;
    CRG11->CLK[clock_id].CRDM = div - 1;
    CRG11->CRCDC = 1;

    return 0;
}

int clk_handler(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    const char* _labels[] = {"M4",  "TRACE", "SYSTICK", "---", "APB", "TIMER", "UART", "SAPB",
                             "AXI", "AK",    "AK_APB",  "SCC", "DDR", "I2S",   "RTC",  "---"};
    if (argc > 1) {
        if (!strcmp(argv[1], "setdiv")) {
            if (argc > 3) {
                int idx = atoi(argv[2]);
                int div = atoi(argv[3]);
                CRG11->CRCDC = 0;
                CRG11->CLK[idx].CRDM = div;
                CRG11->CRCDC = 1;
            }
            goto out_display;
        } else if (!strcmp(argv[1], "mips")) {
            volatile int i = 0;
            uint64_t time = get_time_ms();
            while (time == get_time_ms())
                ;
            time = get_time_ms();
            while (time == get_time_ms())
                i++;
            cli_printf(ctx, "loops/ms: %d\r\n", i);
        }
        return 0;
    }

out_display:
    for (int i = 0; i < 16; i++) {
        int div = CRG11->CLK[i].CRDM;
        cli_printf(ctx, "%8s div %02d (%10dHz) lp %08x used: %d\r\n", _labels[i], div,
                   (int)((uint64_t)CONFIG_SYSTEM_CLOCK_FREQUENCY * 2) / (div + 1),
                   CRG11->CLK[i].CRLP, crg11_clock_used[i]);
    }
    return 0;
}
DECLARE_CMD(clk, clk_handler);
