#include "akida_dma.h"
#include "akida_spi_reg.h"
#include "common.h"
#include "interrupt.h"
#include "spi_bus.h"

#ifndef CONFIG_LOG_SPI_BUS_LEVEL
#define CONFIG_LOG_SPI_BUS_LEVEL CONFIG_LOG_DEFAULT_LEVEL
#endif
DECLARE_LOG_MODULE(SPI_BUS, CONFIG_LOG_SPI_BUS_LEVEL);

/**< Macro definitions START */
#define SEL_ALL_INTR_MASK (0x7F)
#define FIFO_WIDTH_8 (0x0)
#define DMA_TRANSFER_WIDTH (0x8)
#define MAX_DMA_TRANSFER_SIZE (0x10000)
#define RX_ONLY (0)
#define TX_ONLY (1)
#define IO_TYPE_POLL (0)
#define IO_TYPE_DMA (1)
#define IO_TYPE_INTR (2)
#define MAP_DIRECTION(x) (x + 1)
/** Default RX FIFO threshold */
#define RX_FIFO_THRESH_DEF 0x0
/** Default TX FIFO threshold */
#define TX_FIFO_THRESH_DEF 0xf
#define DUMMY_BYTE (0xaa)
#define TX_FIFO_DEPTH (16)
#define RX_FIFO_DEPTH (16)
#define RX_FIFO_EMPTY (0)
#define CLEAR_TXFIFO_INT() REG_WRITE(SPI->TXC, BC_HSSPIn_TXC_ALL_SET(0x7F))
#define CLEAR_RXFIFO_INT() REG_WRITE(SPI->RXC, BC_HSSPIn_RXC_ALL_SET(0x7F))
#define WAIT_FOR_EMPTY_TXFIFO() while (!BC_HSSPIn_TXF_TFES_GET(REG_READ(SPI->TXF)))
#define WAIT_FOR_RX()                                                                          \
    do {                                                                                       \
        if (REG_READ(SPI->RXF) & (BC_HSSPIn_RXF_RSSRS_MASK | BC_HSSPIn_RXF_RFFS_MASK |         \
                                  BC_HSSPIn_RXF_RFMTS_MASK | BC_HSSPIn_FIFOCFG_RXFLSH_MASK)) { \
            break;                                                                             \
        }                                                                                      \
    } while (1);
#define CLEAR_FAULTS() REG_WRITE(SPI->FAULTC, 0x1F)
#define ENABLE_HI_Z_TX() \
    REG_WRITE_MASK(SPI->FIFOCFG, BC_HSSPIn_FIFOCFG_TXCTRL_SET(1), BC_HSSPIn_FIFOCFG_TXCTRL_MASK)
#define DISABLE_HI_Z_TX() \
    REG_WRITE_MASK(SPI->FIFOCFG, BC_HSSPIn_FIFOCFG_TXCTRL_SET(0), BC_HSSPIn_FIFOCFG_TXCTRL_MASK)
/**< Macro definitions END */
/**< Local function declarations START */
#ifndef CONFIG_NO_INTERRUPTS
static inline void spi_irq_handler_rx(void);
static inline void spi_irq_handler_tx(void);
#endif
static inline void spi_rx_dma_cb(hdmac_peripheral_id_t p_id, uint8_t dma_status, void* param);
static inline void spi_tx_dma_cb(hdmac_peripheral_id_t p_id, uint8_t dma_status, void* param);
static inline int spi_rx_priv(struct spi_bus_driver* drv, spi_transfer_t* cur_xfer, uint8_t io_type,
                              spi_bus_callback_t cb, void* cb_param);
static inline int spi_tx_priv(struct spi_bus_driver* drv, spi_transfer_t* cur_xfer, uint8_t io_type,
                              spi_bus_callback_t cb, void* cb_param);
/**< Local function declarations END */
typedef struct {
    uint8_t* usr_buff_ptr;
    uint32_t res_data;
    uint8_t xfer_mode;
} spi_data_priv_t;

typedef struct {
    spi_bus_callback_t rx_cb;
    void* rx_cb_param;
    spi_data_priv_t rx_data_conf;
    spi_data_priv_t tx_data_conf;
    spi_bus_callback_t tx_cb;
    void* tx_cb_param;
    struct spi_bus_driver* drv_inst;
    hdmac_request_conf_t rx_dma_req_conf;
    hdmac_request_conf_t tx_dma_req_conf;
} spi_bus_priv_t;

static spi_bus_priv_t priv_data = {
    .rx_cb = NULL,
    .rx_cb_param = NULL,
    .rx_data_conf =
        {
            .usr_buff_ptr = NULL,
            .res_data = 0,
            .xfer_mode = 0,
        },
    .tx_data_conf =
        {
            .usr_buff_ptr = NULL,
            .res_data = 0,
            .xfer_mode = 0,
        },
    .tx_cb = NULL,
    .tx_cb_param = NULL,
    .drv_inst = NULL,
};
/**< Public interface defintions START */
static int akida_spi_bus_lld_init(struct driver_common* cmn, void* priv) {
    (void)priv;
    struct spi_bus_driver* drv = (struct spi_bus_driver*)cmn;
    if (drv->priv) {
        /* Already initialized */
        return 0;
    }
    spi_bus_priv_t* state = NULL;
    drv->priv = priv;
    state = (spi_bus_priv_t*)drv->priv;
    state->drv_inst = drv;
#ifdef CONFIG_DRIVER_DMA
    akida_hdmac_default_request_conf_get(&(state->rx_dma_req_conf));
    state->rx_dma_req_conf.beat_transfer = HDMAC_BEAT_TRANSFER_SINGLE;
    state->rx_dma_req_conf.transfer_width = DMA_TRANSFER_WIDTH;
    state->rx_dma_req_conf.block_count = RX_FIFO_THRESH_DEF + 1;
    state->rx_dma_req_conf.is_src_addr_fixed = true;
    state->rx_dma_req_conf.src_address = (uint32_t) & (SPI->RXFIFO[0]);

    akida_hdmac_default_request_conf_get(&(state->tx_dma_req_conf));
    state->tx_dma_req_conf.beat_transfer = HDMAC_BEAT_TRANSFER_SINGLE;
    state->tx_dma_req_conf.transfer_width = DMA_TRANSFER_WIDTH;
    state->tx_dma_req_conf.block_count = TX_FIFO_DEPTH - TX_FIFO_THRESH_DEF;
    state->tx_dma_req_conf.is_dest_addr_fixed = true;
    state->tx_dma_req_conf.dest_address = (uint32_t) & (SPI->TXFIFO[0]);
#endif
    /** Clear Interrupts */
    REG_WRITE_MASK(SPI->TXC, BC_HSSPIn_TXC_ALL_SET(SEL_ALL_INTR_MASK), BC_HSSPIn_TXC_ALL_MASK);
    REG_WRITE_MASK(SPI->RXC, BC_HSSPIn_RXC_ALL_SET(SEL_ALL_INTR_MASK), BC_HSSPIn_RXC_ALL_MASK);
    /** Configuring FIFO width */
    REG_WRITE_MASK(SPI->FIFOCFG, BC_HSSPIn_FIFOCFG_FWIDTH_SET(FIFO_WIDTH_8),
                   BC_HSSPIn_FIFOCFG_FWIDTH_MASK);
    /** Clear FIFOs and set FIFO thresholds */
    REG_WRITE_MASK(SPI->FIFOCFG,
                   BC_HSSPIn_FIFOCFG_RXFLSH_SET(1) | BC_HSSPIn_FIFOCFG_TXFLSH_SET(1) |
                       BC_HSSPIn_FIFOCFG_RXFTH_SET(RX_FIFO_THRESH_DEF) |
                       BC_HSSPIn_FIFOCFG_TXFTH_SET(TX_FIFO_THRESH_DEF),
                   BC_HSSPIn_FIFOCFG_RXFLSH_MASK | BC_HSSPIn_FIFOCFG_TXFLSH_MASK |
                       BC_HSSPIn_FIFOCFG_RXFTH_MASK | BC_HSSPIn_FIFOCFG_TXFTH_MASK);
#ifndef CONFIG_NO_INTERRUPTS
    /** Enable SFRC controller and interrupts */
    /** Change interrupt vector */
    intr_set_vector(INTR_SPI_RX, spi_irq_handler_rx);
    intr_set_vector(INTR_SPI_TX, spi_irq_handler_tx);
    /** Enable interrupts */
    intr_enable(INTR_SPI_RX);
    intr_enable(INTR_SPI_TX);
#endif
    /** Enable SFRC controller, select internal clock source */
    REG_WRITE_MASK(SPI->MCTRL, BC_HSSPIn_MCTRL_MEN_SET(1) | BC_HSSPIn_MCTRL_CDSS_SET(0),
                   BC_HSSPIn_MCTRL_MEN_MASK | BC_HSSPIn_MCTRL_CDSS_MASK);
    /**  select software counter mode */
    REG_WRITE_MASK(SPI->DM, BC_HSSPIn_DM_DMCFG_SSDC_SET(0), BC_HSSPIn_DM_DMCFG_SSDC_MASK);

    return SUCCESS;
}

static int akida_spi_lld_slave_config_set(struct spi_bus_driver* drv, const uint8_t slv_id,
                                          const spi_slave_conf_t* conf) {
    (void)drv;
    BCFW_VALIDATE_NULL(conf);
    /** Configure shift direction, endianness, clock divider, slave select polarity */
    REG_WRITE_MASK(
        SPI->PCC[slv_id],
        (true == conf->shift_lsb_first ? BC_HSSPIn_PCCn_SDIR_SET(1) : BC_HSSPIn_PCCn_SDIR_SET(0)) |
            BC_HSSPIn_PCCn_SENDIAN_SET(1) | BC_HSSPIn_PCCn_CDRS_SET(conf->clk_div) |
            (true == conf->slave_sel_pol ? BC_HSSPIn_PCCn_SSPOL_SET(1)
                                         : BC_HSSPIn_PCCn_SSPOL_SET(0)),
        BC_HSSPIn_PCCn_SDIR_MASK | BC_HSSPIn_PCCn_SENDIAN_MASK | BC_HSSPIn_PCCn_CDRS_MASK |
            BC_HSSPIn_PCCn_SSPOL_MASK);
    return SUCCESS;
}

static int akida_spi_bus_lld_read(struct spi_bus_driver* drv, const uint8_t slv_id,
                                  const spi_xfer_mode_t transfer_mode, void* buf, uint32_t size,
                                  spi_bus_callback_t cb, void* cb_param) {
    /* TODO */
    (void)drv;
    (void)slv_id;
    (void)transfer_mode;
    (void)buf;
    (void)size;
    (void)cb;
    (void)cb_param;
    return ENOT_IMPLEMENTED;
}

static int akida_spi_bus_lld_write(struct spi_bus_driver* drv, const uint8_t slv_id,
                                   const spi_xfer_mode_t transfer_mode, void* buf, uint32_t size,
                                   spi_bus_callback_t cb, void* cb_param) {
    /* TODO */
    (void)drv;
    (void)slv_id;
    (void)transfer_mode;
    (void)buf;
    (void)size;
    (void)cb;
    (void)cb_param;
    return ENOT_IMPLEMENTED;
}

static int akida_spi_bus_lld_transceive(struct spi_bus_driver* drv, const uint8_t slv_id,
                                        const spi_xfer_mode_t transfer_mode, void* txb, void* rxb,
                                        uint32_t tx_size, uint32_t rx_size, spi_bus_callback_t cb,
                                        void* cb_param) {
    /* TODO */
    (void)drv;
    (void)slv_id;
    (void)transfer_mode;
    (void)txb;
    (void)rxb;
    (void)tx_size;
    (void)rx_size;
    (void)cb;
    (void)cb_param;
    return ENOT_IMPLEMENTED;
}

static int akida_spi_bus_lld_transact(struct spi_bus_driver* drv, spi_transaction_t* transact,
                                      spi_bus_callback_t cb, void* cb_param) {
    uint8_t io_type = IO_TYPE_POLL, avail_slots = 0;
    int loop_var = 0, rc = 0;
    spi_transfer_t* cur_xfer = NULL;
    BCFW_VALIDATE_NULL(transact);

    uint32_t count = 0;
    for (loop_var = 0; loop_var < transact->xfer_count; loop_var++) {
        count += transact->xfer_list[loop_var].size;
    }
    if (count > 0xffff) {
        SPI_BUS_LOGE("spi transactions limited to 64k");

        return EINVALID_PARM;
    }
    REG_WRITE_MASK(SPI->DM, BC_HSSPIn_DM_DMCFG_SSDC_SET(1) | BC_HSSPIn_DM_DMCFG_MSTARTEN_SET(1),
                   BC_HSSPIn_DM_DMCFG_SSDC_MASK | BC_HSSPIn_DM_DMCFG_MSTARTEN_MASK);
    REG_WRITE_MASK(SPI->FIFOCFG, BC_HSSPIn_FIFOCFG_RXFLSH_SET(1) | BC_HSSPIn_FIFOCFG_TXFLSH_SET(1),
                   BC_HSSPIn_FIFOCFG_RXFLSH_MASK | BC_HSSPIn_FIFOCFG_TXFLSH_MASK);

    /** Set Default TRP mode, controller expects it to be configured in TX_ONLY */
    REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMTRP_OPERATION_SET(MAP_DIRECTION(TX_ONLY)),
                   BC_HSSPIn_DMCTRL_DMTRP_OPERATION_MASK);
    REG_WRITE_MASK(SPI->COUNT, count, 0xffff);
    REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTART_START_SET(1),
                   BC_HSSPIn_DMCTRL_DMSTART_START_MASK);
    REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMPSEL_PSEL_SET(transact->slave_id),
                   BC_HSSPIn_DMCTRL_DMPSEL_PSEL_MASK);

    for (loop_var = 0; loop_var < transact->xfer_count; loop_var++) {
        io_type = IO_TYPE_POLL;
        cur_xfer = &transact->xfer_list[loop_var];
        BCFW_VALIDATE_NULL(cur_xfer);
        /** Check if the transfer size qualifies for DMA/interrupt */
        if (cur_xfer->txd || (NULL == cur_xfer->txd && NULL == cur_xfer->rxd)) {
            if (NULL != cb && loop_var == (transact->xfer_count - 1)) {
                /* Get available free slots in in TxFIFO */
                avail_slots =
                    (TX_FIFO_DEPTH - BC_HSSPIn_DMSTATUS_TXFLEVEL_GET(REG_READ(SPI->DMSTATUS)));
                if (cur_xfer->size <= avail_slots) {
                    io_type = IO_TYPE_INTR;
                } else {
                    io_type = IO_TYPE_DMA;
                }
            }
            if ((rc = spi_tx_priv(drv, cur_xfer, io_type, cb, cb_param))) {
                SPI_BUS_LOGE("spi_tx_priv() failed rc = %u , status  = 0x%08x", rc,
                             REG_READ(SPI->DMSTATUS));
                return rc;
            }
        } else if (cur_xfer->rxd) {
            if (NULL != cb && loop_var == (transact->xfer_count - 1)) {
                /* Get available occupied slots in in RxFIFO */
                avail_slots =
                    RX_FIFO_DEPTH - BC_HSSPIn_DMSTATUS_RXFLEVEL_GET(REG_READ(SPI->DMSTATUS));
                if (cur_xfer->size <= avail_slots) {
                    io_type = IO_TYPE_INTR;
                } else {
                    io_type = IO_TYPE_DMA;
                }
            }
            if ((rc = spi_rx_priv(drv, cur_xfer, io_type, cb, cb_param))) {
                SPI_BUS_LOGE("spi_rx_priv() failed rc = %u , status  = 0x%08x", rc,
                             REG_READ(SPI->DMSTATUS));
                return rc;
            }
        }
    }

    if (IO_TYPE_POLL == io_type) {
        REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(1),
                       BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
        if (cur_xfer->rxd) {
            while (0 == BC_HSSPIn_RXF_RSSRS_GET(REG_READ(SPI->RXF)))
                ;

        } else {
            while (0 == BC_HSSPIn_TXF_TSSRS_GET(REG_READ(SPI->TXF)))
                ;
        }
        REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(0),
                       BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);

        return SUCCESS;
    }

    return rc;
}
/**< Public interface defintions END */

/**< Local function definitions START */
static inline int spi_tx_priv(struct spi_bus_driver* drv, spi_transfer_t* cur_xfer, uint8_t io_type,
                              spi_bus_callback_t cb, void* cb_param) {
    int loop_var = 0, rc = SUCCESS;
    CLEAR_TXFIFO_INT();
    CLEAR_RXFIFO_INT();
    /** Configure TRP:operating mode, TRP:transfer width */
    REG_WRITE_MASK(SPI->DMCTRL,
                   BC_HSSPIn_DMCTRL_DMTRP_OPERATION_SET(MAP_DIRECTION(TX_ONLY)) |
                       BC_HSSPIn_DMCTRL_DMTRP_WIDTH_SET(cur_xfer->mode),
                   BC_HSSPIn_DMCTRL_DMTRP_OPERATION_MASK | BC_HSSPIn_DMCTRL_DMTRP_WIDTH_MASK);
#ifdef CONFIG_DRIVER_DMA
    if (IO_TYPE_DMA == io_type) {
        /** Using DMA for Transmit operation */
        int size = MAX_DMA_TRANSFER_SIZE;
        if (size > cur_xfer->size) {
            size = cur_xfer->size;
        }
        spi_bus_priv_t* state = (spi_bus_priv_t*)drv->priv;
        state->tx_data_conf.res_data = (cur_xfer->size) - size;
        state->tx_data_conf.xfer_mode = cur_xfer->mode;
        state->tx_data_conf.usr_buff_ptr = (cur_xfer->txd) + size;
        state->tx_cb = cb;
        state->tx_cb_param = cb_param;
        state->tx_dma_req_conf.transfer_count = size;
        state->tx_dma_req_conf.src_address = (uint32_t)cur_xfer->txd;
        SPI_BUS_LOGD("Size = %u, leftout=%u", size, state->tx_data_conf.res_data);

        if (!(rc = akida_hdmac_request_transfer(HDMAC_PERIPHERAL_ID_SPI_TX,
                                                &(state->tx_dma_req_conf), spi_tx_dma_cb, drv))) {
            REG_WRITE_MASK(SPI->DM, BC_HSSPIn_DM_DMDMAEN_TXDMAEN_SET(1),
                           BC_HSSPIn_DM_DMDMAEN_TXDMAEN_MASK);
        } else {
            SPI_BUS_LOGE("DMA request failed rc = %u", rc);
            return rc;
        }
    } else
#endif
#ifndef CONFIG_NO_INTERRUPTS
        if (IO_TYPE_INTR == io_type) {
        spi_bus_priv_t* state = (spi_bus_priv_t*)drv->priv;
        state->tx_data_conf.res_data = cur_xfer->size;
        state->tx_data_conf.xfer_mode = cur_xfer->mode;
        state->tx_data_conf.usr_buff_ptr = cur_xfer->txd;
        state->tx_cb = cb;
        state->tx_cb_param = cb_param;
        /* Enable interrupts that are triggered on writing into TX_FIFO*/
        REG_WRITE_MASK(SPI->TXE, BC_HSSPIn_TXE_TFLETE_SET(1) | BC_HSSPIn_TXE_TFEE_SET(1),
                       BC_HSSPIn_TXE_TFLETE_MASK | BC_HSSPIn_TXE_TFEE_MASK);

    } else
#else
    (void)cb;
    (void)cb_param;
    (void)io_type;
    (void)drv;
#endif
    {
        /** Using Polling for Transmit operation */
        if (NULL == cur_xfer->txd) {
            ENABLE_HI_Z_TX();
            for (loop_var = 0; loop_var < cur_xfer->size; loop_var++) {
                /** DUMMY bytes */
                REG_WRITE(*((uint8_t*)&(SPI->TXFIFO[0])), DUMMY_BYTE);
            }
            /** Wait until FIFO is empty*/
            WAIT_FOR_EMPTY_TXFIFO();
            DISABLE_HI_Z_TX();
        } else {
            for (loop_var = 0; loop_var < cur_xfer->size; loop_var++) {
                while (TX_FIFO_DEPTH == BC_HSSPIn_DMSTATUS_TXFLEVEL_GET(REG_READ(SPI->DMSTATUS)))
                    ;
                /** Write to TXFIFO*/
                REG_WRITE(*((uint8_t*)&(SPI->TXFIFO[0])), cur_xfer->txd[loop_var]);
            }
            /** Wait until FIFO is empty*/
            WAIT_FOR_EMPTY_TXFIFO();
        }
    }
    return rc;
}

static inline int spi_rx_priv(struct spi_bus_driver* drv, spi_transfer_t* cur_xfer, uint8_t io_type,
                              spi_bus_callback_t cb, void* cb_param) {
    int rc = SUCCESS, loop_var = 0;

    CLEAR_TXFIFO_INT();
    CLEAR_RXFIFO_INT();
    REG_WRITE_MASK(SPI->DMCTRL,
                   BC_HSSPIn_DMCTRL_DMTRP_OPERATION_SET(MAP_DIRECTION(RX_ONLY)) |
                       BC_HSSPIn_DMCTRL_DMTRP_WIDTH_SET(cur_xfer->mode),
                   BC_HSSPIn_DMCTRL_DMTRP_OPERATION_MASK | BC_HSSPIn_DMCTRL_DMTRP_WIDTH_MASK);
    WAIT_FOR_RX();
    CLEAR_FAULTS();
#ifdef CONFIG_DRIVER_DMA
    if (IO_TYPE_DMA == io_type) {
        int size = MAX_DMA_TRANSFER_SIZE;
        if (size > cur_xfer->size) {
            size = cur_xfer->size;
        }
        spi_bus_priv_t* state = (spi_bus_priv_t*)drv->priv;
        state->rx_data_conf.xfer_mode = cur_xfer->mode;
        state->rx_data_conf.res_data = (cur_xfer->size) - size;
        state->rx_data_conf.usr_buff_ptr = (cur_xfer->rxd) + size;
        state->rx_cb = cb;
        state->rx_cb_param = cb_param;

        /** Using DMA for Receive operation */
        state->rx_dma_req_conf.transfer_count = size;
        state->rx_dma_req_conf.dest_address = (uint32_t)(cur_xfer->rxd);
        SPI_BUS_LOGD("Size = %u, leftout=%u", size, state->rx_data_conf.res_data);
        if (!(rc = akida_hdmac_request_transfer(HDMAC_PERIPHERAL_ID_SPI_RX,
                                                &(state->rx_dma_req_conf), spi_rx_dma_cb, drv))) {
            REG_WRITE_MASK(SPI->DM, BC_HSSPIn_DM_DMDMAEN_RXDMAEN_SET(1),
                           BC_HSSPIn_DM_DMDMAEN_RXDMAEN_MASK);
        } else {
            SPI_BUS_LOGE("DMA request failed rc = %u", rc);
            return rc;
        }
    } else
#endif
#ifndef CONFIG_NO_INTERRUPTS
        if (IO_TYPE_INTR == io_type) {
        spi_bus_priv_t* state = (spi_bus_priv_t*)drv->priv;
        /* Save size to be read, user buffer in shared data structure*/
        state->rx_data_conf.xfer_mode = cur_xfer->mode;
        state->rx_data_conf.res_data = (cur_xfer->size);
        state->rx_data_conf.usr_buff_ptr = cur_xfer->rxd;
        state->rx_cb = cb;
        state->rx_cb_param = cb_param;
        /* Enable interrupts that are triggered on availability of data in RX_FIFO*/
        REG_WRITE_MASK(
            SPI->RXE,
            BC_HSSPIn_RXE_RSSRE_SET(1) | BC_HSSPIn_RXE_RFFE_SET(1) | BC_HSSPIn_RXE_RFOE_SET(1) |
                BC_HSSPIn_RXE_RFLETE_SET(1) | BC_HSSPIn_RXE_RFMTE_SET(1),
            BC_HSSPIn_RXE_RFFE_MASK | BC_HSSPIn_RXE_RFOE_MASK | BC_HSSPIn_RXE_RFLETE_MASK |
                BC_HSSPIn_RXE_RFMTE_MASK | BC_HSSPIn_RXE_RSSRE_MASK);
    } else
#else
    (void)cb;
    (void)cb_param;
    (void)io_type;
    (void)drv;
#endif
    {
        /** Using polling for Receive operation */
        for (loop_var = 0; loop_var < cur_xfer->size; loop_var++) {
            while (RX_FIFO_EMPTY == BC_HSSPIn_DMSTATUS_RXFLEVEL_GET(REG_READ(SPI->DMSTATUS)))
                ;
            /** Read from RXFIFO*/
            cur_xfer->rxd[loop_var] = REG_READ(*((uint8_t*)&(SPI->RXFIFO[0])));
        }
    }
    return rc;
}

static inline void spi_rx_dma_cb(hdmac_peripheral_id_t p_id, uint8_t dma_status, void* param) {
    (void)p_id;
    int rc = SUCCESS;
    uint8_t avail_slots = 0, io_type = IO_TYPE_DMA;
    struct spi_bus_driver* drv = (struct spi_bus_driver*)param;
    spi_bus_priv_t* state = drv->priv;
    spi_transfer_t cur_xfer;
    if (HDMAC_STATUS_NORMAL_END == dma_status) {
        REG_WRITE_MASK(SPI->DM, BC_HSSPIn_DM_DMDMAEN_RXDMAEN_SET(0),
                       BC_HSSPIn_DM_DMDMAEN_RXDMAEN_MASK);
        cur_xfer.txd = NULL;
        cur_xfer.mode = state->rx_data_conf.xfer_mode;
        cur_xfer.size = state->rx_data_conf.res_data;
        cur_xfer.rxd = state->rx_data_conf.usr_buff_ptr;

        avail_slots = RX_FIFO_DEPTH - BC_HSSPIn_DMSTATUS_RXFLEVEL_GET(REG_READ(SPI->DMSTATUS));
        if (cur_xfer.size) {
            if (cur_xfer.size <= avail_slots) {
                io_type = IO_TYPE_INTR;
            } else {
                io_type = IO_TYPE_DMA;
            }
            if ((rc = spi_rx_priv(drv, &cur_xfer, io_type, state->rx_cb, state->rx_cb_param))) {
                if (state->rx_cb) {
                    state->rx_cb(drv, EFAILURE, state->rx_data_conf.res_data, state->rx_cb_param);
                }
            }
        } else {
            REG_WRITE_MASK(SPI->COUNT, BC_HSSPIn_COUNT_DMBCC_BCC_SET(0),
                           BC_HSSPIn_COUNT_DMBCC_BCC_MASK);
            REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(1),
                           BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
            while (0 == BC_HSSPIn_RXF_RSSRS_GET(REG_READ(SPI->RXF)))
                ;
            REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(0),
                           BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
            if (state->rx_cb) {
                state->rx_cb(drv, rc, state->rx_data_conf.res_data, state->rx_cb_param);
            }
        }
    } else {
        if (state->rx_cb) {
            state->rx_cb(drv, EFAILURE, state->rx_data_conf.res_data, state->rx_cb_param);
        }
    }
}

static inline void spi_tx_dma_cb(hdmac_peripheral_id_t p_id, uint8_t dma_status, void* param) {
    (void)p_id;
    int rc = SUCCESS;
    uint8_t avail_slots = 0, io_type = IO_TYPE_DMA;
    struct spi_bus_driver* drv = (struct spi_bus_driver*)param;
    spi_bus_priv_t* state = drv->priv;
    spi_transfer_t cur_xfer;
    if (HDMAC_STATUS_NORMAL_END == dma_status) {
        REG_WRITE_MASK(SPI->DM, BC_HSSPIn_DM_DMDMAEN_TXDMAEN_SET(0),
                       BC_HSSPIn_DM_DMDMAEN_TXDMAEN_MASK);
        cur_xfer.rxd = NULL;
        cur_xfer.mode = state->tx_data_conf.xfer_mode;
        cur_xfer.size = state->tx_data_conf.res_data;
        cur_xfer.txd = state->tx_data_conf.usr_buff_ptr;

        avail_slots = (TX_FIFO_DEPTH - BC_HSSPIn_DMSTATUS_TXFLEVEL_GET(REG_READ(SPI->DMSTATUS)));
        if (cur_xfer.size) {
            if (cur_xfer.size <= avail_slots) {
                io_type = IO_TYPE_INTR;
            } else {
                io_type = IO_TYPE_DMA;
            }
            if ((rc = spi_tx_priv(drv, &cur_xfer, io_type, state->tx_cb, state->tx_cb_param))) {
                if (state->tx_cb) {
                    state->tx_cb(drv, EFAILURE, state->tx_data_conf.res_data, state->tx_cb_param);
                }
            }
        } else {
            REG_WRITE_MASK(SPI->COUNT, BC_HSSPIn_COUNT_DMBCC_BCC_SET(0),
                           BC_HSSPIn_COUNT_DMBCC_BCC_MASK);
            REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(1),
                           BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
            while (0 == BC_HSSPIn_TXF_TSSRS_GET(REG_READ(SPI->TXF)))
                ;
            REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(0),
                           BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
            if (state->tx_cb) {
                state->tx_cb(drv, rc, state->tx_data_conf.res_data, state->tx_cb_param);
            }
        }
    } else {
        if (state->tx_cb) {
            state->tx_cb(drv, EFAILURE, state->tx_data_conf.res_data, state->tx_cb_param);
        }
    }
}

#ifndef CONFIG_NO_INTERRUPTS
static inline void spi_irq_handler_rx(void) {
    uint8_t rx_level = BC_HSSPIn_DMSTATUS_RXFLEVEL_GET(REG_READ(SPI->DMSTATUS)), loop_var = 0,
            intr_status = 0, cur_size = 0;
    spi_bus_priv_t* state = &priv_data;
    uint8_t* rxd = NULL;
    intr_status = REG_READ(SPI->RXF);

    /* Disable interrupts */
    REG_WRITE(SPI->RXC, intr_status);
    REG_WRITE(SPI->RXE, 0);
    intr_clear_pending(INTR_SPI_RX);

    if ((intr_status &
         (BC_HSSPIn_RXF_RFFS_MASK | BC_HSSPIn_RXF_RFOS_MASK | BC_HSSPIn_RXF_RFLETS_MASK |
          BC_HSSPIn_RXF_RFMTS_MASK | BC_HSSPIn_RXF_RSSRS_MASK)) &&
        state->rx_data_conf.res_data) {
        /* Interrupt for data availablity */

        cur_size = rx_level;
        if (rx_level > state->rx_data_conf.res_data) {
            cur_size = state->rx_data_conf.res_data;
        }
        rxd = state->rx_data_conf.usr_buff_ptr;

        for (loop_var = 0; loop_var < cur_size; loop_var++) {
            rxd[loop_var] = REG_READ(*((uint8_t*)&(SPI->RXFIFO[0])));
        }
        state->rx_data_conf.res_data -= cur_size;
        if (state->rx_data_conf.res_data) {
            state->rx_data_conf.usr_buff_ptr += cur_size;
            /* Enable interrupts */
            REG_WRITE_MASK(SPI->RXE,
                           BC_HSSPIn_RXE_RSSRE_SET(1) | BC_HSSPIn_RXE_RFFE_SET(1) |
                               BC_HSSPIn_RXE_RFOE_SET(1) | BC_HSSPIn_RXE_RFLETE_SET(1) |
                               BC_HSSPIn_RXE_RFMTE_SET(1),
                           BC_HSSPIn_RXE_RFFE_MASK | BC_HSSPIn_RXE_RFOE_MASK |
                               BC_HSSPIn_RXE_RFLETE_MASK | BC_HSSPIn_RXE_RFMTE_MASK);

        } else {
            REG_WRITE_MASK(SPI->RXE, BC_HSSPIn_RXE_RSSRE_SET(1), BC_HSSPIn_RXE_RSSRE_MASK);
            REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(1),
                           BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
        }
    }
    if (intr_status & (BC_HSSPIn_RXE_RSSRE_MASK)) { /* Interrupt for slave select relase */
        /* Clear stop condition */
        REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(0),
                       BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
        /* Invoke user call back */
        if (!state->rx_data_conf.res_data) {
            state->rx_cb(state->drv_inst, SUCCESS, 0, state->rx_cb_param);
        } else {
            state->rx_cb(state->drv_inst, EFAILURE, state->rx_data_conf.res_data,
                         state->rx_cb_param);
        }
    }
}

static inline void spi_irq_handler_tx(void) {
    uint8_t tx_level = BC_HSSPIn_DMSTATUS_TXFLEVEL_GET(REG_READ(SPI->DMSTATUS)), loop_var = 0,
            space_aval = 0, intr_status = 0;
    spi_bus_priv_t* state = &priv_data;
    uint8_t* txd = NULL;
    intr_status = REG_READ(SPI->TXF);
    /* Disable interrupts */
    REG_WRITE(SPI->TXC, intr_status);
    REG_WRITE(SPI->TXE, 0);
    intr_clear_pending(INTR_SPI_TX);

    if ((intr_status & (BC_HSSPIn_TXF_TFLETS_MASK | BC_HSSPIn_TXF_TFES_MASK)) &&
        state->tx_data_conf.res_data) {
        space_aval = TX_FIFO_DEPTH - tx_level;
        tx_level = state->tx_data_conf.res_data;
        if (state->tx_data_conf.res_data > space_aval) {
            tx_level = space_aval;
        }
        txd = state->tx_data_conf.usr_buff_ptr;

        for (loop_var = 0; loop_var < tx_level; loop_var++) {
            /** Write to TXFIFO*/
            REG_WRITE(*((uint8_t*)&(SPI->TXFIFO[0])), txd[loop_var]);
        }

        state->tx_data_conf.res_data -= tx_level;
        if (state->tx_data_conf.res_data) {
            state->tx_data_conf.usr_buff_ptr += tx_level;
            /* Enable interrupts */
            REG_WRITE_MASK(SPI->TXE, BC_HSSPIn_TXE_TFLETE_SET(1) | BC_HSSPIn_TXE_TFEE_SET(1),
                           BC_HSSPIn_TXE_TFLETE_MASK | BC_HSSPIn_TXE_TFEE_MASK);

        } else {
            REG_WRITE_MASK(SPI->TXE, BC_HSSPIn_TXE_TSSRE_SET(1), BC_HSSPIn_TXE_TSSRE_MASK);
            REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(1),
                           BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
        }
    } else if (intr_status & (BC_HSSPIn_TXE_TSSRE_MASK)) {
        /* Clear stop condition */
        REG_WRITE_MASK(SPI->DMCTRL, BC_HSSPIn_DMCTRL_DMSTOP_STOP_SET(0),
                       BC_HSSPIn_DMCTRL_DMSTOP_STOP_MASK);
        /* Invoke user call back */
        if (!state->tx_data_conf.res_data) {
            state->tx_cb(state->drv_inst, SUCCESS, 0, state->tx_cb_param);
        } else {
            state->tx_cb(state->drv_inst, EFAILURE, state->tx_data_conf.res_data,
                         state->tx_cb_param);
        }
    }
}
#endif

/**< Local function definitions END */
struct spi_bus_ops akida_spi_bus_lld_ops = {
    .slave_config_set = akida_spi_lld_slave_config_set,
    .read = akida_spi_bus_lld_read,
    .write = akida_spi_bus_lld_write,
    .transceive = akida_spi_bus_lld_transceive,
    .transact = akida_spi_bus_lld_transact,
};

DECLARE_SPI_BUS(spi_bus_lld_0, &akida_spi_bus_lld_ops, akida_spi_bus_lld_init, (void*)&priv_data);
