#include <stdint.h>
#include "common.h"
#include "interrupt.h"
#include "soc.h"
#include "systimer.h"
#include "uart.h"
#include "uart_reg.h"

/*
 * FIFO trigger levels: 1/8, 1/4, 1/2, 3/4, and 7/8
 */
#define UART_FIFO_LVL_1BY8 0
#define UART_FIFO_LVL_1BY4 1
#define UART_FIFO_LVL_1BY2 2
#define UART_FIFO_LVL_3BY4 3
#define UART_FIFO_LVL_7BY8 4

#define UART_FIFO_SIZE 32
#define UART_RX_FIFO_LVL_DEFAULT UART_FIFO_LVL_1BY8
#define UART_TX_FIFO_LVL_DEFAULT UART_FIFO_LVL_1BY2

/*
 *  UART private data strcture
 */
typedef union {
    uint32_t raw;
    struct {
        uint32_t tx : 1;
        uint32_t rx : 1;
    } bits;
} uart_status_t;

typedef struct {
    volatile uint32_t len;
    volatile uint32_t req_len;
    volatile uint32_t index;
    volatile uint64_t bailout_time;
    uart_xfer_complete_callback cb;
    void* cb_param;
    volatile const uint8_t* buf;
    struct systimer tx_timer;
} uart_int_tctx_t;

typedef struct {
    volatile uint32_t len;
    volatile uint32_t req_len;
    volatile uint32_t index;
    volatile uint64_t bailout_time;
    uart_xfer_complete_callback cb;
    void* cb_param;
    volatile uint8_t* buf;
    struct systimer rx_timer;
} uart_int_rctx_t;

typedef struct {
    /* UART instance register address*/
    uart_reg_t* regs;
    volatile uart_status_t uart_status;
    uart_int_tctx_t intr_tx_ctx;
    uart_int_rctx_t intr_rx_ctx;
    int irq_no;
    void (*irq_handler)(void);
    void* drv;
} uart_instance_t;

void uart_intr_handler(uart_instance_t* uart);
void uart0_intr_handler(void);
void uart1_intr_handler(void);

void tx_timer_cb(struct systimer* timer);
void rx_timer_cb(struct systimer* timer);

uart_instance_t uart0 = {
    (uart_reg_t*)UART0_BASE,
    {0},
    {0, 0, 0, 0, NULL, NULL, NULL, {.cb = tx_timer_cb}},
    {0, 0, 0, 0, NULL, NULL, NULL, {.cb = rx_timer_cb}},
    INTR_UART0,
    &uart0_intr_handler,
    NULL,
};

uart_instance_t uart1 = {
    (uart_reg_t*)UART1_BASE,
    {0},
    {0, 0, 0, 0, NULL, NULL, NULL, {.cb = tx_timer_cb}},
    {0, 0, 0, 0, NULL, NULL, NULL, {.cb = rx_timer_cb}},
    INTR_UART1,
    &uart1_intr_handler,
    NULL,
};

void uart0_intr_handler(void) {
    uart_intr_handler(&uart0);
}
void uart1_intr_handler(void) {
    uart_intr_handler(&uart1);
}

void uart_intr_handler(uart_instance_t* uart) {
    uart_reg_t* uart_reg_base = uart->regs;

    /* Read the interrupt status, clear the interrupts */
    volatile uint32_t raw_intr = uart_reg_base->RIS;

    /*Process the RX interrupt when rx is active status */
    if ((raw_intr & (UART_RIS_RXRIS | UART_RIS_RTRIS)) && (uart->uart_status.bits.rx)) {
        while (!(uart_reg_base->FR & UART_FR_RXFE) && (uart->intr_rx_ctx.len > 0)) {
            if (uart_reg_base->RSRECR & (UART_RSR_BE | UART_RSR_PE | UART_RSR_FE)) {
                /* Receive character has met an error, break now */
                uart_reg_base->RSRECR &= 0xf;
                break;
            }
            uart->intr_rx_ctx.len--;
            uart->intr_rx_ctx.buf[uart->intr_rx_ctx.index++] = uart_reg_base->DR & 0xFF;
        }

        if (uart->intr_rx_ctx.len == 0) {
            /* Clear the RX interrupt from IMSC*/
            REG_CLEAR_BIT_FIELD(uart_reg_base->IMSC, (UART_IMSC_RXIM | UART_IMSC_RTIM));

            /*Deregister the rx timer */
            systimer_deregister(&uart->intr_rx_ctx.rx_timer);
            uart_xfer_complete_callback lcb = uart->intr_rx_ctx.cb;
            void* lcb_param = uart->intr_rx_ctx.cb_param;
            uart->uart_status.bits.rx = 0;
            uart->intr_rx_ctx.cb = NULL;
            uart->intr_rx_ctx.cb_param = NULL;
            /* Received the expected length of characters, invoke the callback */
            if (NULL != lcb) {
                lcb(uart->drv, lcb_param, 0, (uart->intr_rx_ctx.req_len - uart->intr_rx_ctx.len));
            }
        }
    }

    if ((raw_intr & UART_RIS_TXRIS) && (uart->uart_status.bits.tx)) {
        while (uart->intr_tx_ctx.len > 0) {
            /* Break when Transmission FIFO got Full */
            if (uart_reg_base->FR & UART_FR_TXFF) {
                break;
            }
            uart->intr_tx_ctx.len--;
            uart_reg_base->DR = uart->intr_tx_ctx.buf[uart->intr_tx_ctx.index++];
        }

        if (uart->intr_tx_ctx.len == 0) {
            /* Clear the TX interrupt from IMSC*/
            REG_CLEAR_BIT_FIELD(uart_reg_base->IMSC, UART_RIS_TXRIS_Msk);
            /*Deregister the tx timer */
            systimer_deregister(&uart->intr_tx_ctx.tx_timer);

            uart_xfer_complete_callback lcb = uart->intr_tx_ctx.cb;
            void* lcb_param = uart->intr_tx_ctx.cb_param;
            /* Completed the transfer, remove tx status and invoke the callback*/
            uart->uart_status.bits.tx = 0;
            uart->intr_tx_ctx.cb = NULL;
            uart->intr_tx_ctx.cb_param = NULL;
            if (NULL != lcb) {
                lcb(uart->drv, lcb_param, 0, (uart->intr_tx_ctx.req_len - uart->intr_tx_ctx.len));
            }
        }
    }

    REG_CLEAR_BIT_FIELD(uart_reg_base->ICR, raw_intr);
}

void tx_timer_cb(struct systimer* timer) {
    uart_instance_t* uart = (uart_instance_t*)timer->cb_param;

    if (uart->uart_status.bits.tx) {
        /* Clear the TX interrupt from IMSC*/
        REG_CLEAR_BIT_FIELD(uart->regs->IMSC, UART_RIS_TXRIS_Msk);
        uart_xfer_complete_callback lcb = uart->intr_tx_ctx.cb;
        void* lcb_param = uart->intr_tx_ctx.cb_param;
        /* Completed the transfer, remove tx status and invoke the callback*/
        uart->uart_status.bits.tx = 0;
        uart->intr_tx_ctx.cb = NULL;
        uart->intr_tx_ctx.cb_param = NULL;
        uart->intr_tx_ctx.bailout_time = 0;
        if (NULL != lcb) {
            lcb(uart->drv, lcb_param, 1, (uart->intr_tx_ctx.req_len - uart->intr_tx_ctx.len));
        }
    }
}

void rx_timer_cb(struct systimer* timer) {
    uart_instance_t* uart = (uart_instance_t*)timer->cb_param;

    if (uart->uart_status.bits.rx) {
        /* Clear the RX interrupt from IMSC*/
        REG_CLEAR_BIT_FIELD(uart->regs->IMSC, (UART_IMSC_RXIM | UART_IMSC_RTIM));
        uart_xfer_complete_callback lcb = uart->intr_rx_ctx.cb;
        void* lcb_param = uart->intr_rx_ctx.cb_param;
        uart->uart_status.bits.rx = 0;
        uart->intr_rx_ctx.cb = NULL;
        uart->intr_rx_ctx.cb_param = NULL;
        uart->intr_rx_ctx.bailout_time = 0;
        /* Received the expected length of characters, invoke the callback */
        if (NULL != lcb) {
            lcb(uart->drv, lcb_param, 1, (uart->intr_rx_ctx.req_len - uart->intr_rx_ctx.len));
        }
    }
}

static int uart_calc_baudrate_divisior(uint32_t clock, uint32_t baud_rate, int32_t* ibrd,
                                       int32_t* fbrd) {
    /*
     *  Baud rate divisor BAUDDIV = (FUARTCLK/(16�Baud rate))
     *  where FUARTCLK is the UART reference clock frequency.
     *  For clock freq @ 100MHZ to the UART peripheral.
     *  BAUDDIV = 100000000 / (16 * 9600) ans = 651.0416667. UARTIBRD = 651
     *  UARTFBRD = (0.0416 * 64)+0.5 = 3.1624 ~ 3
     *  The following integerDivisors and fractiondivisor arrays are derived suing above
     *  caliculations. Dont change these values unless you are sure about it.
     *
     */

    uint32_t baud_int_div;
    float baud_rate_div = (float)clock / (16 * baud_rate);

    /* Calculate integer baud rate register value */
    baud_int_div = (uint32_t)baud_rate_div;
    *ibrd = baud_int_div;

    /* Calculate fractional baud rate register value */
    baud_rate_div -= baud_int_div;
    *fbrd = (uint32_t)((baud_rate_div * 64) + 0.5);

    return 0;
}

static int uart_read_char(struct uart_driver* drv) {
    uint8_t ch = 0;

    uart_instance_t* uart_instance = (uart_instance_t*)drv->priv;
    volatile uart_reg_t* uart_reg_base = (uart_reg_t*)uart_instance->regs;

    /* Return try again when rx fifo is empty */
    if (uart_reg_base->FR & UART_FR_RXFE) {
        return EAGAIN;
    }

    if (uart_reg_base->RSRECR & (UART_RSR_BE | UART_RSR_PE | UART_RSR_FE)) {
        /* Receive character has met an error, return failure */
        uart_reg_base->RSRECR &= 0xf;
        return EFAILURE;
    }
    ch = uart_reg_base->DR & 0xFF;

    return (ch);
}

static void uart_setconfig(struct uart_driver* drv, const uart_config_t* uart_cfg) {
    int32_t ibrd = 0, fbrd = 0;
    uint32_t temp_reg_val = 0, oddEven = 0;
    bool parity = 0;

    uart_instance_t* uart_instance = (uart_instance_t*)drv->priv;
    volatile uart_reg_t* uart_reg_base = (uart_reg_t*)uart_instance->regs;
    bool fifo_enable = true;
    uint32_t rx_fifo_lvl = UART_RX_FIFO_LVL_DEFAULT;
    uint32_t tx_fifo_lvl = UART_TX_FIFO_LVL_DEFAULT;

    uart_calc_baudrate_divisior(clock_get_freq(DRIVER_GET_CLK_ID(&drv->driver)),
                                uart_cfg->baud_rate, &ibrd, &fbrd);

    REG_WRITE_MASK(uart_reg_base->IBRD, ibrd, UART_IBRD_BD_Msk);
    REG_WRITE_MASK(uart_reg_base->FBRD, fbrd, UART_FBRF_BD_Msk);

    if (UART_PARITY_NO != uart_cfg->parity) {
        parity = true;
        oddEven = uart_cfg->parity;
    }

    temp_reg_val = ((uart_cfg->sticky_parity << UART_LCRH_SPS_Pos) |
                    ((uart_cfg->word_len - UART_WORDLEN_5) << UART_LCRH_WLEN_Pos) |
                    (fifo_enable << UART_LCRH_FEN_Pos) |
                    ((uart_cfg->stop_bits - UART_STOP_BIT_ONE) << UART_LCRH_STP2_Pos) |
                    (oddEven << UART_LCRH_EPS_Pos) | (parity << UART_LCRH_PEN_Pos));

    REG_WRITE_MASK(uart_reg_base->LCR_H, temp_reg_val,
                   (UART_LCRH_SPS | UART_LCRH_WLEN | UART_LCRH_FEN | UART_LCRH_STP2 |
                    UART_LCRH_EPS | UART_LCRH_PEN));

    /*Clear all the interrupts */
    REG_WRITE(uart_reg_base->ICR, (UART_ICR_RIMIC | UART_ICR_CTSMIC | UART_ICR_DCDMIC |
                                   UART_ICR_DSRMIC | UART_ICR_RXIC | UART_ICR_TXIC | UART_ICR_RTIC |
                                   UART_ICR_FEIC | UART_ICR_PEIC | UART_ICR_BEIC | UART_ICR_OEIC));

    /* Configure the Tx and Rx fifolevels */
    temp_reg_val =
        ((rx_fifo_lvl << UART_IFLS_RXIFLSEL_Pos) | (tx_fifo_lvl << UART_IFLS_TXIFLSEL_Pos));

    REG_WRITE_MASK(uart_reg_base->IFLS, temp_reg_val,
                   UART_IFLS_RXIFLSEL_Msk | UART_IFLS_TXIFLSEL_Msk);

    /* CTS and RTS flow control is not enabled, incase if it enabled,
    Data will be transmitted only when nUARTCTS is asserted*/

    temp_reg_val = UART_CR_UARTEN | UART_CR_TXE | UART_CR_RXE;
    REG_WRITE_MASK(uart_reg_base->CR, temp_reg_val,
                   (UART_CR_UARTEN_Msk | UART_CR_TXE_Msk | UART_CR_RXE_Msk));
}

static int akida_uart_lld_open(struct uart_driver* drv, const uart_config_t* uart_cfg) {
    uart_instance_t* uart_instance = (uart_instance_t*)drv->priv;
    uart_setconfig(drv, uart_cfg);
    intr_enable(uart_instance->irq_no);

    return 0;
}

static int akida_uart_lld_read(struct uart_driver* drv, uint8_t* buf, const uint32_t len,
                               const uint32_t timeout, uart_xfer_complete_callback cb,
                               void* cb_param) {
    volatile int read_ch = 0, ret = 0;
    uint32_t rlen = 0, flags = 0;
    uart_instance_t* uart_instance = (uart_instance_t*)drv->priv;
    volatile uart_reg_t* uart_reg_base = (uart_reg_t*)uart_instance->regs;
    uint64_t bailout_time;

    flags = interrupt_lock();
    if (uart_instance->uart_status.bits.rx) {
        interrupt_unlock(flags);
        return EBUSY;
    }
    uart_instance->uart_status.bits.rx = 1;
    interrupt_unlock(flags);

    /* if timeout is 0xFFFFFFFF  assign bailout_time to the max of uint64_t*/
    bailout_time = (0xFFFFFFFF == timeout) ? (uint64_t)0xFFFFFFFF : (timeout + get_time_ms());

    if (cb) {
        uart_instance->intr_rx_ctx.cb = cb;
        uart_instance->intr_rx_ctx.cb_param = cb_param;

        uart_instance->intr_rx_ctx.buf = buf;
        uart_instance->intr_rx_ctx.index = 0;
        uart_instance->intr_rx_ctx.len = len;
        uart_instance->intr_rx_ctx.req_len = len;
        uart_instance->intr_rx_ctx.rx_timer.cb_param = uart_instance;
        uart_instance->intr_rx_ctx.bailout_time = bailout_time;
        /* Enable the Rx interrupts */
        REG_WRITE_MASK(uart_reg_base->IMSC, (UART_IMSC_RXIM | UART_IMSC_RTIM),
                       (UART_IMSC_RXIM_Msk | UART_IMSC_RTIM_Msk));

        if (0xFFFFFFFF != timeout) {
            /* Avoid timer registration when the timeout is  0xFFFFFFFF */
            systimer_register(&uart_instance->intr_rx_ctx.rx_timer, timeout);
        }

    } else {
        while ((rlen < len) && (bailout_time > get_time_ms())) {
            /*
             Loop untill character is avial when rx fifo is empty or timeout has met.
            */
            while ((uart_reg_base->FR & UART_FR_RXFE) && (bailout_time > get_time_ms())) {
            };

            read_ch = uart_read_char(drv);
            /* Notify the EFAILURE when frame error, over run or parity errors met*/
            if (EAGAIN == read_ch) {
                continue;
            }
            if (EFAILURE == read_ch) {
                ret = EFAILURE;
                goto out;
            }
            buf[rlen] = read_ch;
            rlen++;
        }
        ret = rlen;
    out:
        flags = interrupt_lock();
        uart_instance->uart_status.bits.rx = 0;
        interrupt_unlock(flags);
    }

    return ret;
}

static int akida_uart_lld_write(struct uart_driver* drv, const uint8_t* buf, const uint32_t len,
                                const uint32_t timeout, uart_xfer_complete_callback cb,
                                void* cb_param) {
    uint32_t wlen = 0, flags = 0, time_elapsed = 0;
    uart_instance_t* uart_instance = (uart_instance_t*)drv->priv;
    volatile uart_reg_t* uart_reg_base = (uart_reg_t*)uart_instance->regs;
    uint64_t bailout_time;

    flags = interrupt_lock();
    if (uart_instance->uart_status.bits.tx) {
        interrupt_unlock(flags);
        return EBUSY;
    }
    uart_instance->uart_status.bits.tx = 1;
    interrupt_unlock(flags);
    /* if timeout is 0xFFFFFFFF  assign bailout_time to the max of uint64_t*/
    bailout_time = (0xFFFFFFFF == timeout) ? (uint64_t)0xFFFFFFFF : (timeout + get_time_ms());

    if (cb) {
        uart_xfer_complete_callback lcb = NULL;
        void* lcb_param;

        uart_instance->intr_tx_ctx.cb = cb;
        uart_instance->intr_tx_ctx.cb_param = cb_param;
        uart_instance->intr_tx_ctx.buf = buf;
        uart_instance->intr_tx_ctx.index = 0;
        uart_instance->intr_tx_ctx.len = len;
        uart_instance->intr_tx_ctx.req_len = len;
        uart_instance->intr_tx_ctx.bailout_time = bailout_time;
        uart_instance->intr_tx_ctx.tx_timer.cb_param = uart_instance;
        REG_CLEAR_BIT_FIELD(uart_reg_base->ICR, UART_ICR_TXIC);

        /*
        Tx interrupt is based on a transition through a level, rather than on the level itself.
        The interrupt is only set, after written data leaves the single location of the transmit
        FIFO and it becomes empty. Hence do transfer the untill the FIFO is FULL or len becomes 0
        */

        while (uart_instance->intr_tx_ctx.len > 0) {
            if (bailout_time < get_time_ms()) {
                time_elapsed = 1;
                break;
            }

            if (uart_reg_base->FR & UART_FR_TXFF) {
                break;
            }
            uart_reg_base->DR = buf[uart_instance->intr_tx_ctx.index++];
            uart_instance->intr_tx_ctx.len--;
        }
        if ((!time_elapsed) && (uart_instance->intr_tx_ctx.len > 0) && (0xFFFFFFFF != timeout)) {
            systimer_register(&uart_instance->intr_tx_ctx.tx_timer, (bailout_time - get_time_ms()));
        }
        /* Enable the Tx interrupts */

        REG_WRITE_MASK(uart_reg_base->IMSC, UART_IMSC_TXIM, UART_IMSC_TXIM_Msk);

        flags = interrupt_lock();
        if (((uart_instance->intr_tx_ctx.len == 0) && (uart_instance->uart_status.bits.tx)) ||
            time_elapsed) {
            REG_CLEAR_BIT_FIELD(uart_reg_base->IMSC, UART_RIS_TXRIS_Msk);
            uart_instance->uart_status.bits.tx = 0;
            lcb = uart_instance->intr_tx_ctx.cb;
            lcb_param = uart_instance->intr_tx_ctx.cb_param;
            uart_instance->intr_tx_ctx.bailout_time = 0;
            uart_instance->intr_tx_ctx.cb = NULL;
            interrupt_unlock(flags);

            /*Deregister the tx timer */
            if (0xFFFFFFFF != timeout) {
                systimer_deregister(&uart_instance->intr_tx_ctx.tx_timer);
            }
            /* Tx is completed invoke the user callback function */
            lcb(uart_instance->drv, lcb_param, time_elapsed,
                (uart_instance->intr_tx_ctx.req_len - uart_instance->intr_tx_ctx.len));
            flags = interrupt_lock();
        }
        interrupt_unlock(flags);
    } else {
        while ((wlen < len) && (bailout_time > get_time_ms())) {
            if (uart_reg_base->FR & UART_FR_TXFF) {
                continue;
            }
            uart_reg_base->DR = buf[wlen++];
        }
        flags = interrupt_lock();
        uart_instance->uart_status.bits.tx = 0;
        interrupt_unlock(flags);
    }

    return wlen;
}

static int akida_uart_lld_close(struct uart_driver* drv) {
    BCFW_VALIDATE_NULL(drv);
    uart_instance_t* uart_instance = (uart_instance_t*)drv->priv;
    volatile uart_reg_t* uart_reg_base = (uart_reg_t*)uart_instance->regs;

    intr_disable(uart_instance->irq_no);
    REG_WRITE(uart_reg_base->IMSC, 0x00);
    REG_WRITE(uart_reg_base->RIS, 0x00);
    REG_WRITE_MASK(uart_reg_base->CR, 0x0, UART_CR_UARTEN_Msk);

    return 0;
}

struct uart_stream_ops akida_uart_lld_stream_ops = {.open = akida_uart_lld_open,
                                                    .read = akida_uart_lld_read,
                                                    .write = akida_uart_lld_write,
                                                    .close = akida_uart_lld_close};

static int akida_uart_lld_init(struct driver_common* drv, void* priv) {
    BCFW_VALIDATE_NULL(drv);
    BCFW_VALIDATE_NULL(priv);

    struct uart_driver* uart_drv = (struct uart_driver*)drv;
    uart_drv->priv = (uart_instance_t*)priv;
    uart_instance_t* uart_instance = (uart_instance_t*)priv;
    uart_instance->drv = drv;
    intr_set_vector(uart_instance->irq_no, uart_instance->irq_handler);

    UART_LOGI("Initialize akida_uart_lld");
    return 0;
}

struct driver_common akida_uart_lld_common = {.init = akida_uart_lld_init};

static int akida_uart_lld_char_set_config(struct uart_driver* drv, const uart_config_t* uart_cfg) {
    BCFW_VALIDATE_NULL(drv);
    uart_setconfig(drv, uart_cfg);
    return 0;
}

static int akida_uart_lld_get_char(struct uart_driver* drv) {
    BCFW_VALIDATE_NULL(drv);
    return uart_read_char(drv);
}

static int akida_uart_lld_put_char(struct uart_driver* drv, const uint8_t ch) {
    BCFW_VALIDATE_NULL(drv);
    uart_instance_t* uart_instance = (uart_instance_t*)drv->priv;
    volatile uart_reg_t* uart_reg_base = (uart_reg_t*)uart_instance->regs;

    if (uart_reg_base->FR & UART_FR_TXFF) {
        return EAGAIN;
    }

    uart_reg_base->DR = ch;

    return 0;
}

struct uart_char_ops akida_uart_lld_char_ops = {.set_config = akida_uart_lld_char_set_config,
                                                .get_char = akida_uart_lld_get_char,
                                                .put_char = akida_uart_lld_put_char};

DECLARE_UART_INTF(uart_lld_0, &akida_uart_lld_stream_ops, &akida_uart_lld_char_ops,
                  akida_uart_lld_init, (void*)&uart0, CRG11_CLK_ID_M4_UART)
DECLARE_UART_INTF(uart_lld_1, &akida_uart_lld_stream_ops, &akida_uart_lld_char_ops,
                  akida_uart_lld_init, (void*)&uart1, CRG11_CLK_ID_M4_UART)
