#include "interrupt.h"
#include "soc.h"
#include "wdog.h"
#include "wdog_reg.h"

#define WDOGCLOCK (clock_get_freq(DRIVER_GET_CLK_ID((struct driver_common*)drv)))

/*Parameter to be read from board specific file*/
#define WDOG_CLKEN_ONE 1
#define WDOG_CLKEN_PULSE 2

#define WDOG_CLKEN WDOG_CLKEN_ONE

#define WDOG_TICK_COUNT(MSEC) ((uint64_t)MSEC * ((WDOGCLOCK / WDOG_CLKEN) / 1000))

#define WDOG_MSEC_TIME(VALUE) ((VALUE / WDOGCLOCK) * (WDOG_CLKEN * 1000))

#ifndef CONFIG_WDOG_DEFAULT_TIMEOUT
#define CONFIG_WDOG_DEFAULT_TIMEOUT 10000
#endif

#define WDOG_DEFAULT_TIMEOUT CONFIG_WDOG_DEFAULT_TIMEOUT

#define WDOG_INTR_PRIORITY 0
#define CONF_ITEM_IDX_WDOG_TIMEOUT 0
#define CONF_ITEM_IDX_WDOG_ENABLED 1

typedef struct {
    wdog_expire_handler_t handle;
    void* param;
    struct wdog_driver* drv;
    uint32_t timeout;
    wdog_reg_t* regs;
} wdog_handler_t;

struct config_item wdt_config_items[] = {{.name = "timeout", .default_value = WDOG_DEFAULT_TIMEOUT},
                                         {.name = "enabled", .default_value = 0}};

static wdog_handler_t wdog_handle = {NULL, NULL, NULL, WDOG_DEFAULT_TIMEOUT,
                                     (wdog_reg_t*)WDOG_BASE};

void wdog_intr_handler() {
    wdog_handle.handle(wdog_handle.drv, wdog_handle.param);

    if (WDOG_RIS_WDOGRIS == REG_READ_MASK(wdog_handle.regs->WdogRIS, WDOG_RIS_WDOGRIS_Msk)) {
        REG_WRITE(wdog_handle.regs->WdogLoad, 0);
    }
}

static int wdog_lld_init(struct driver_common* drv, void* priv) {
    (void)priv;

    /*Unlock Write access to all writable WDOG Registers*/
    REG_WRITE(wdog_handle.regs->WdogLock, WDOG_ENABLE_WRITE);

    /*Enable Watchdog module reset output*/
    REG_WRITE(wdog_handle.regs->WdogControl, WDOG_CNTL_RESEN_Msk);

    /*WDOG Interrupt vector configuration*/
    intr_set_vector(INTR_WDOG, wdog_intr_handler);

    intr_clear_pending(INTR_WDOG);

    intr_set_priority(INTR_WDOG, WDOG_INTR_PRIORITY);

    wdog_handle.timeout = WDOG_TICK_COUNT(WDOG_DEFAULT_TIMEOUT);

    DRIVER_CONFIG_SET_NEW(drv, CONF_ITEM_IDX_WDOG_TIMEOUT, wdog_handle.timeout);
    DRIVER_CONFIG_SET_CURRENT(drv, CONF_ITEM_IDX_WDOG_TIMEOUT, wdog_handle.timeout);
    return SUCCESS;
}

static int wdog_lld_start(struct wdog_driver* drv) {
    (void)drv;
    /*Return failure if wdog is already enabled*/
    if (REG_READ_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk)) {
        return EFAILURE;
    }
    /* clearing any pending interrupt present before setting timeout value to wdog*/
    if (intr_clear_pending(INTR_WDOG)) {
        return EFAILURE;
    }
    if (wdog_handle.handle) {
        /*Load timeout value to the watchdog load register*/
        REG_WRITE(wdog_handle.regs->WdogLoad, wdog_handle.timeout);

        /*Enable the WDOG interrupt*/
        intr_enable(INTR_WDOG);
    } else {
        /*Load timeout value to the watchdog load register*/
        REG_WRITE(wdog_handle.regs->WdogLoad, (wdog_handle.timeout / 2));
    }

    /*Enable the counter and the interrupt*/
    REG_WRITE_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk, WDOG_CNTL_INTEN_Msk);

    DRIVER_CONFIG_SET_CURRENT(
        drv, CONF_ITEM_IDX_WDOG_ENABLED,
        (WDOG_CNTL_INTEN_Msk == REG_READ_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk))
            ? 1
            : 0);

    return SUCCESS;
}

static int wdog_lld_stop(struct wdog_driver* drv) {
    (void)drv;
    /*Return failure if wdog is disabled*/
    if (WDOG_CNTL_INTEN_Msk != REG_READ_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk)) {
        return EFAILURE;
    }

    /*Disable the counter and the interrupt*/
    REG_CLEAR_BIT_FIELD(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk);

    if (wdog_handle.handle) {
        /*Disable the WDOG interrupt*/
        intr_disable(INTR_WDOG);
    }

    DRIVER_CONFIG_SET_CURRENT(drv, CONF_ITEM_IDX_WDOG_ENABLED,
                              !!(WDOG_CNTL_INTEN_Msk == REG_READ_MASK(wdog_handle.regs->WdogControl,
                                                                      WDOG_CNTL_INTEN_Msk)));

    return SUCCESS;
}

static void wdog_lld_kick(struct wdog_driver* drv) {
    (void)drv;

    /*Clear the interrupt and reload the wdog counter*/
    REG_CLEAR_BIT_FIELD(wdog_handle.regs->WdogIntClr, WDOG_RIS_WDOGRIS_Msk);
}

static int wdog_lld_set_timeout(struct wdog_driver* drv, uint32_t timeout) {
    /*Return failure if wdog is enabled*/
    if (WDOG_CNTL_INTEN_Msk == REG_READ_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk)) {
        return EBAD_STATE;
    }

    if (WDOG_TICK_COUNT(timeout) > 0xffffffff) {
        return EINVALID_PARM;
    }

    wdog_handle.timeout = WDOG_TICK_COUNT(timeout);

    return SUCCESS;
}

static int wdog_lld_set_expire_handler(struct wdog_driver* drv, wdog_expire_handler_t handle,
                                       void* param) {
    /*Return failure if wdog is enabled*/
    if (WDOG_CNTL_INTEN_Msk == REG_READ_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk)) {
        return EBAD_STATE;
    }

    wdog_handle.handle = handle;
    wdog_handle.param = param;
    wdog_handle.drv = drv;

    return SUCCESS;
}

int wdog_lld_read_config(struct driver_common* drv) {
    DRIVER_CONFIG_SET_CURRENT(drv, CONF_ITEM_IDX_WDOG_TIMEOUT, wdog_handle.timeout);
    DRIVER_CONFIG_SET_CURRENT(drv, CONF_ITEM_IDX_WDOG_ENABLED,
                              !!(WDOG_CNTL_INTEN_Msk == REG_READ_MASK(wdog_handle.regs->WdogControl,
                                                                      WDOG_CNTL_INTEN_Msk)));

    return SUCCESS;
}

int wdog_lld_apply_config(struct driver_common* drv) {
    bool enable_flag = false;

    if (WDOG_CNTL_INTEN_Msk == REG_READ_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk)) {
        enable_flag = true;
        wdog_lld_stop((struct wdog_driver*)drv);
    }

    wdog_lld_set_timeout((struct wdog_driver*)drv,
                         DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_WDOG_TIMEOUT).new_value);
    DRIVER_CONFIG_APPLY(drv, CONF_ITEM_IDX_WDOG_TIMEOUT);

    if ((DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_WDOG_ENABLED).new_value && enable_flag) ||
        DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_WDOG_ENABLED).new_value) {
        wdog_lld_start((struct wdog_driver*)drv);
        DRIVER_CONFIG_SET_CURRENT(
            drv, CONF_ITEM_IDX_WDOG_ENABLED,
            !!(WDOG_CNTL_INTEN_Msk ==
               REG_READ_MASK(wdog_handle.regs->WdogControl, WDOG_CNTL_INTEN_Msk)));
    }

    return SUCCESS;
}

struct wdog_ops wdog_lld_ops = {
    .start = wdog_lld_start,
    .stop = wdog_lld_stop,
    .kick = wdog_lld_kick,
    .set_timeout = wdog_lld_set_timeout,
    .set_expire_handler = wdog_lld_set_expire_handler,
};

DECLARE_WDOG(wdog_lld, &wdog_lld_ops, wdog_lld_init, wdog_lld_read_config, wdog_lld_apply_config,
             wdt_config_items, (sizeof(wdt_config_items) / sizeof(wdt_config_items[0])), NULL,
             CRG11_CLK_ID_WDOG);
