#include "driver.h"
#include <stdlib.h>
#include "bsp_error.h"
#include "cli_cmd.h"
#include "custom_sections.h"
#include "i2c_bus.h"

DECLARE_CUSTOM_SECTION(drv, struct driver_item, DRV);

struct dynamic_driver {
    struct driver_common* drv;
    struct dynamic_driver* next;
}* dynamic_driver_list = NULL;

int driver_init() {
    struct driver_item* drv = drv_start;
    while (drv < drv_end) {
        int ret;
        drv->driver->reg = drv;
        if (drv->type == DRIVER_TYPE_DEVICE && drv->driver->init) {
            ret = drv->driver->init(drv->driver, drv->priv);
            if (drv->driver->read_config)
                drv->driver->read_config(drv->driver);
            if (ret && ret != ENOT_IMPLEMENTED) {
                return ret;
            }
        }
        drv++;
    }
    return 0;
}

int _strcmp(char* str1, char* str2) {
    /* byte to byte strcmp */
    int ret = 0;
    uint32_t len = strlen(str1);
    uint32_t len2 = strlen(str2);
    if (len2 > len) {
        len = len2;
    }
    for (uint32_t i = 0; i < len; i++) {
        if (str1[i] != str2[i]) {
            ret = str1[i] - str2[i];
            break;
        }
    }

    return ret;
}

/**
 * @brief returns driver matching name.
 * If the driver passed is a bus driver, the items connected to the list are also iterated over.
 * Interface type drivers are excluded from the search as they need an instance to function.
 *
 * @param drv the driver_item to match against
 * @param name the name of the driver to search
 * @return non NULL if found
 */
static struct driver_common* matching(struct driver_item* drv, char* name) {
    if (drv->type == DRIVER_TYPE_DEVICE && !_strcmp(drv->name, name)) {
        if (!drv->driver->reg) {
            drv->driver->reg = drv;
        }
        return drv->driver;
    }
    if (drv->type == DRIVER_TYPE_DEVICE && !_strcmp(drv->class, "i2c")) {
        struct i2c_bus_driver* b = (struct i2c_bus_driver*)drv->driver;
        struct i2c_device* d = b->devices;
        while (d) {
            if (_strcmp(d->name, name) == 0) {
                return &d->driver;
            }
            d = d->next;
        }
    }
    return NULL;
}

struct driver_common* driver_lookup(char* name) {
    struct driver_item* drv = drv_start;
    while (drv < drv_end) {
        struct driver_common* match = matching(drv, name);
        if (match)
            return match;
        drv++;
    }

    struct dynamic_driver* ptr = dynamic_driver_list;
    while (ptr) {
        struct driver_common* match = matching(ptr->drv->reg, name);
        if (match)
            return match;
        ptr = ptr->next;
    }

    return NULL;
}

void driver_foreach(int (*cb)(struct driver_item* drv, void* param), void* param) {
    struct driver_item* drv = drv_start;
    while (drv < drv_end) {
        if (cb(drv, param))
            return;
        drv++;
    }

    struct dynamic_driver* ptr = dynamic_driver_list;
    while (ptr) {
        if (cb(ptr->drv->reg, param))
            return;
        ptr = ptr->next;
    }
}

struct driver_common* interface_lookup(char* name) {
    struct driver_item* drv = drv_start;
    while (drv < drv_end) {
        if (drv->type == DRIVER_TYPE_INTERFACE && !_strcmp(drv->name, name)) {
            return drv->driver;
        }
        drv++;
    }
    struct dynamic_driver* ptr = dynamic_driver_list;
    while (ptr) {
        if (ptr->drv->reg->type == DRIVER_TYPE_INTERFACE && !(_strcmp(drv->name, name))) {
            return ptr->drv;
        }
        ptr = ptr->next;
    }
    return NULL;
}

int driver_add_dynamic(struct driver_common* drv) {
    struct dynamic_driver* d = (struct dynamic_driver*)malloc(sizeof(*d));
    if (d == NULL) {
        return ENOMEM;
    }
    d->drv = drv;
    d->next = dynamic_driver_list;
    dynamic_driver_list = d;
    return 0;
}

int driver_apply_config(struct driver_common* drv) {
    if (drv->apply_config) {
        return drv->apply_config(drv);
    }
    return ENOT_IMPLEMENTED;
}

static void print_config(struct cli_cmd_ctx* ctx, struct config_item* config_items) {
    cli_printf(ctx, "\t%16s: %8d %8d %8d\r\n", config_items->name, config_items->current_value,
               config_items->new_value, config_items->default_value);
}

static void show_config(char* name, struct driver_common* drv, struct cli_cmd_ctx* ctx) {
    cli_printf(ctx, "driver %s:\r\n", name, drv->config_items_count, drv->config_items);
    for (int i = 0; i < drv->config_items_count; i++) {
        print_config(ctx, drv->config_items + i);
    }
}

static void show_config_if_present(struct driver_common* driver, struct driver_item* drv,
                                   struct cli_cmd_ctx* ctx) {
    if (drv->type == DRIVER_TYPE_DEVICE && driver->config_items != NULL) {
        show_config(drv->name, drv->driver, ctx);
    }
    if (drv->type == DRIVER_TYPE_DEVICE && !strcmp(drv->class, "i2c")) {
        struct i2c_bus_driver* b = (struct i2c_bus_driver*)drv->driver;
        struct i2c_device* d = b->devices;
        while (d) {
            if (d->driver.config_items)
                show_config(d->name, &d->driver, ctx);
            d = d->next;
        }
    }
}

int config_cli_handler(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    int ret = SUCCESS;

    if (argc == 1) {
        struct driver_item* drv = drv_start;
        while (drv < drv_end) {
            show_config_if_present(drv->driver, drv, ctx);
            drv++;
        }

        struct dynamic_driver* ptr = dynamic_driver_list;
        while (ptr) {
            show_config_if_present(ptr->drv, ptr->drv->reg, ctx);
            ptr = ptr->next;
        }
    } else if (argc >= 2) {
        struct driver_common* drv = driver_lookup(argv[1]);
        if (drv == NULL) {
            ret = ENOT_IMPLEMENTED;
        }

        if (!ret && drv->config_items != NULL) {
            if (argc < 3) {
                show_config(argv[1], drv, ctx);
            } else {
                if (!strcmp(argv[2], "apply")) {
                    ret = driver_apply_config(drv);
                } else if (argc == 3 && strcmp(argv[2], "apply")) {
                    char* item = argv[2];
                    int indx = 0;
                    for (indx = 0; indx < drv->config_items_count; indx++) {
                        if (!strcmp(DRIVER_CONFIG_GET_ITEM(drv, indx).name, item)) {
                            print_config(ctx, drv->config_items + indx);
                            break;
                        }
                    }

                    if (indx == drv->config_items_count) {
                        ret |= EFAILURE;
                        cli_printf(ctx, "config %s not found for %s\r\n", item, argv[1]);
                    }
                }
                if (strcmp(argv[2], "apply")) {
                    int index = 2;
                    while (index < argc) {
                        char* item = argv[index];
                        int value = strtoul(argv[index + 1], NULL, 0);
                        int indx = 0;
                        for (indx = 0; indx < drv->config_items_count; indx++) {
                            if (!strcmp(DRIVER_CONFIG_GET_ITEM(drv, indx).name, item)) {
                                DRIVER_CONFIG_SET_NEW(drv, indx, value);
                                break;
                            }
                        }
                        if (indx == drv->config_items_count) {
                            ret |= EFAILURE;
                            cli_printf(ctx, "config %s not found for %s\r\n", item, argv[1]);
                        }
                        index += 2;
                    }
                }
            }
        }
    }
    if (ret) {
        cli_printf(ctx, "usage: %s [<driver_name> [{apply} | <item_1>] <value>.... ]\r\n", argv[0]);
    }
    return ret;
}

/**
 * @page Commands CLI
 * @section config CONFIG
 * @ref CONFIG-CLI-1 \n
 */
/** @page Commands
 * 	@subsection CONFIG-CLI-1 config
 *   CLI to set and verify configuration items of driver modules \n
 * 	__syntax			:  config [&lt;driver_name&gt; [{apply} | &lt;item_1&gt;]
 * <value>...]__\n
 * 	__driver_name__		:  Represents driver name, example-> ina1, ina2, humid_sensor,
 * wdog_lld\n
 * 	__item1__			:  Represents config item defined in the driver\n
 * 	__value__			:  Represents new value\n
 */

DECLARE_CMD(config, config_cli_handler);

/**
 * @page Commands CLI
 * @section drv DRV
 */

/** @page Commands
 * 	@subsection DRV drv
 *   List available drivers in the system \n
 * __config list | config__: list available drivers in the system
 */

int drv_cli_handler(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    if (argc == 1 || !strcmp("list", argv[1])) {
        struct driver_item* drv = drv_start;
        int indent = 0;
        while (drv < drv_end) {
            for (int i = 0; i < indent; i++) {
                cli_printf(ctx, "    ");
            }
            cli_printf(ctx, "%s: %s class: %s type: %d\r\n",
                       drv->type == DRIVER_TYPE_INTERFACE ? "intf" : "drv ", drv->name, drv->class,
                       drv->type);
            if (drv->type == DRIVER_TYPE_DEVICE && !strcmp(drv->class, "i2c")) {
                struct i2c_bus_driver* b = (struct i2c_bus_driver*)drv->driver;
                struct i2c_device* d = b->devices;
                indent++;
                while (d) {
                    for (int i = 0; i < indent; i++) {
                        cli_printf(ctx, "    ");
                    }
                    cli_printf(ctx, "drv : %s intf: %s\r\n", d->name,
                               d->driver.reg ? d->driver.reg->name : "none");
                    d = d->next;
                }
                indent--;
            }
            drv++;
        }
        struct dynamic_driver* ptr = dynamic_driver_list;
        while (ptr) {
            cli_printf(ctx, "%s: %s class: %s type: %d\r\n",
                       ptr->drv->reg->type == DRIVER_TYPE_INTERFACE ? "intf" : "drv ",
                       ptr->drv->reg->name, ptr->drv->reg->class, ptr->drv->reg->type);

            if (ptr->drv->reg->type == DRIVER_TYPE_DEVICE && !strcmp(ptr->drv->reg->class, "i2c")) {
                struct i2c_bus_driver* b = (struct i2c_bus_driver*)ptr->drv->reg->driver;
                struct i2c_device* d = b->devices;
                indent++;
                while (d) {
                    for (int i = 0; i < indent; i++) {
                        cli_printf(ctx, "    ");
                    }
                    cli_printf(ctx, "drv : %s intf: %s\r\n", d->name,
                               d->driver.reg ? d->driver.reg->name : "none");
                    d = d->next;
                }
                indent--;
            }
            ptr = ptr->next;
        }
    }
    return 0;
}
DECLARE_CMD(drv, drv_cli_handler);
