#include <stddef.h>
#include <stdint.h>
#include "bsp_error.h"
#include "common.h"
#include "i2c/ina233.h"
#include "i2c_bus.h"
#include "power_monitor.h"

#define INA233_MFR_CALIB_REG 0xd4
#define INA233_VBUS_VOLTAGE_REG 0x88
#define INA233_INPUT_CURRENT_REG 0x89
#define INA233_INPUT_POWER_REG 0x97
#define INA233_DEVICE_CONFIG_REG 0xd5
#define INA233_CLEAR_EIN_REG 0xd6
#define INA233_TI_MFR_ID_REG 0xe0
#define INA233_TI_MFR_MODEL_REG 0xe1
#define INA233_TI_MFR_REVISION_REG 0xe2
#define INA233_MFR_ID_REG 0x99
#define INA233_MFR_MODEL_REG 0x9a
#define INA233_MFR_REVISION_REG 0x9b
#define INA233_MFR_MODEL_REG_RET_SIZE 0x7

/*To be read form board config file*/
#define CURR_LSB_CAL_2_POW_15 32768
#define BUS_VOLTAGE_LSB 1250
#define CURRENT_LSB(max_cur) \
    ((1000 * max_cur) / CURR_LSB_CAL_2_POW_15) /*Value for a 0.001 ohm shunt and a 1mA LSb*/
#define POWER_LSB(cur_lsb) (25 * cur_lsb)

#define CONF_CALIB_MULTPLIER 5120000
#define CONF_ITEM_IDX_INA233_SHUNT 0
#define CONF_ITEM_IDX_INA233_CURRENT 1

struct config_item ina233_config_items[] = {{.name = "shunt_value", .default_value = 1000},
                                            {.name = "max_current", .default_value = 32768}};

int ina233_lld_get_voltage(struct power_monitor_driver* drv, int* volt) {
    int ret = SUCCESS;
    uint8_t buf[2];

    struct i2c_device* d = (struct i2c_device*)drv->priv;

    ret = i2c_read_reg(d, INA233_VBUS_VOLTAGE_REG, buf, 2);

    if (0 < ret) {
        *volt = (((buf[0]) | (buf[1] << 8)) * BUS_VOLTAGE_LSB);
    }

    return ret == 2 ? 0 : ret;
}

int ina233_lld_get_current(struct power_monitor_driver* drv, int* current) {
    int ret = SUCCESS;
    uint8_t buf[2];
    struct i2c_device* d = (struct i2c_device*)drv->priv;

    ret = i2c_read_reg(d, INA233_INPUT_CURRENT_REG, buf, 2);

    if (0 < ret) {
        *current = (((buf[0]) | (buf[1] << 8)) *
                    CURRENT_LSB(DRIVER_CONFIG_GET_ITEM(((struct driver_common*)drv),
                                                       CONF_ITEM_IDX_INA233_CURRENT)
                                    .current_value));
    }
    return ret == 2 ? 0 : ret;
}

int ina233_lld_get_power(struct power_monitor_driver* drv, int* power) {
    int ret = SUCCESS;
    uint8_t buf[2];
    struct i2c_device* d = (struct i2c_device*)drv->priv;

    ret = i2c_read_reg(d, INA233_INPUT_POWER_REG, buf, 2);

    if (0 < ret) {
        *power = (((buf[0]) | (buf[1] << 8)) *
                  POWER_LSB(CURRENT_LSB(DRIVER_CONFIG_GET_ITEM(((struct driver_common*)drv),
                                                               CONF_ITEM_IDX_INA233_CURRENT)
                                            .current_value)));
    }

    return ret == 2 ? 0 : ret;
}

static int ina233_lld_init(struct driver_common* drv, void* priv) {
    uint8_t buf[INA233_MFR_MODEL_REG_RET_SIZE];
    uint16_t cal;

    ((struct power_monitor_driver*)drv)->priv = priv;
    struct i2c_device* i2c_dev = (struct i2c_device*)priv;
    int ret = i2c_read_reg(i2c_dev, INA233_MFR_MODEL_REG, buf, INA233_MFR_MODEL_REG_RET_SIZE);
    /* INA233 replies with number of bytes followed by "INA233" */
    if (INA233_MFR_MODEL_REG_RET_SIZE == ret) {
        if (strncmp((char*)buf + 1, "INA233", 6)) {
            return EFAILURE;
        }
    } else {
        return EFAILURE;
    }

    cal = (((uint64_t)CONF_CALIB_MULTPLIER * 1000) /
           (CURRENT_LSB(DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_CURRENT).default_value) *
            DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_SHUNT).default_value));

    /*LSB First*/
    buf[1] = ((cal >> 8) & 0xFF);
    buf[0] = (cal & 0xFF);

    ret = i2c_write_reg(i2c_dev, INA233_MFR_CALIB_REG, buf, 2);

    return ret != 2 ? ret : 0;
}

struct power_monitor_ops ina233_lld_ops = {
    .read_voltage = ina233_lld_get_voltage,
    .read_current = ina233_lld_get_current,
    .read_power = ina233_lld_get_power,
};

int ina233_read_config(struct driver_common* drv) {
    DRIVER_CONFIG_SET_CURRENT(
        drv, CONF_ITEM_IDX_INA233_SHUNT,
        DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_SHUNT).default_value);
    DRIVER_CONFIG_SET_CURRENT(
        drv, CONF_ITEM_IDX_INA233_CURRENT,
        DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_CURRENT).default_value);

    return SUCCESS;
}

int ina233_apply_config(struct driver_common* drv) {
    struct i2c_device* i2c_dev = (struct i2c_device*)((struct power_monitor_driver*)drv)->priv;
    uint8_t buf[2];
    uint16_t cal;

    if (DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_SHUNT).new_value) {
        DRIVER_CONFIG_APPLY(drv, CONF_ITEM_IDX_INA233_SHUNT);
    }
    if (DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_CURRENT).new_value) {
        DRIVER_CONFIG_APPLY(drv, CONF_ITEM_IDX_INA233_CURRENT);
    }

    cal = (((uint64_t)CONF_CALIB_MULTPLIER * 1000) /
           (CURRENT_LSB(DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_CURRENT).current_value) *
            DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_INA233_SHUNT).current_value));

    /*LSB First*/
    buf[1] = ((cal >> 8) & 0xFF);
    buf[0] = (cal & 0xFF);

    int ret = i2c_write_reg(i2c_dev, INA233_MFR_CALIB_REG, buf, 2);

    return ret != 2 ? ret : 0;
}

struct {
    struct driver_common driver;
    struct power_monitor_ops* ops;
    void* priv;
} ina233_lld_drv = {
    .driver.init = ina233_lld_init,
    .driver.read_config = ina233_read_config,
    .driver.apply_config = ina233_apply_config,
    .driver.config_items = ina233_config_items,
    .driver.config_items_count = sizeof(ina233_config_items) / sizeof(ina233_config_items[0]),
    .ops = &ina233_lld_ops,
};

DECLARE_INTERFACE(ina233, pmon, &ina233_lld_drv.driver, NULL);
