#include <stdlib.h>

#include <string.h>
#include "bsp_error.h"
#include "elog.h"
#include "i2c_bus.h"

#ifdef CONFIG_LOG_DEFAULT_LEVEL /* If not defined, we are in the applib build */
#ifndef CONFIG_LOG_I2C_LEVEL
#define CONFIG_LOG_I2C_LEVEL CONFIG_LOG_DEFAULT_LEVEL
#endif
DECLARE_LOG_MODULE(I2C, CONFIG_LOG_I2C_LEVEL);
#endif

int i2c_read(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb, void* cb_param) {
    if (drv->bus && drv->bus->ops->read) {
        return drv->bus->ops->read(drv, buf, len, cb, cb_param);
    }
    return ENOT_IMPLEMENTED;
}

int i2c_write(struct i2c_device* drv, uint8_t* buf, int len, i2c_callback_t cb, void* cb_param) {
    if (drv->bus && drv->bus->ops->write) {
        return drv->bus->ops->write(drv, buf, len, cb, cb_param);
    }
    return ENOT_IMPLEMENTED;
}

int i2c_read_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len) {
    if (drv->bus && drv->bus->ops->read_reg) {
        return drv->bus->ops->read_reg(drv, reg, buf, len);
    }
    return ENOT_IMPLEMENTED;
}

int i2c_write_reg(struct i2c_device* drv, uint8_t reg, uint8_t* buf, int len) {
    if (drv->bus && drv->bus->ops->write_reg) {
        return drv->bus->ops->write_reg(drv, reg, buf, len);
    }
    return ENOT_IMPLEMENTED;
}

int i2c_detach(struct i2c_device* drv) {
    if (drv->bus && drv->bus->ops->detach) {
        return drv->bus->ops->detach(drv->bus, drv);
    }
    return ENOT_IMPLEMENTED;
}

int i2c_attach(struct i2c_bus_driver* bus, struct i2c_device* dev, struct driver_common* intf) {
    int ret = 0;
    struct config_item* configs;
    /* Copy location of defined config items into temporary pointer variable */
    configs = dev->driver.config_items;

    /* Set underlying bus */
    dev->bus = bus;

    I2C_LOGI("Attach i2c device %s to bus %s intf: %s", dev->name, bus->driver.reg->name,
             intf ? intf->reg->name : "none");

    if (intf) {
        /* Copy the init, deinit pointers, as well as the ops and priv pointer */
        memcpy(&dev->driver, intf, sizeof(struct driver_common) + 2 * sizeof(void*));
        /* Validate the device driver config items count with the interface config items count */
        if (dev->driver.config_items_count == intf->config_items_count) {
            if (NULL != configs) {
                /*Replace the overwritten config items with the temporary stored address value */
                dev->driver.config_items = configs;
                memcpy(dev->driver.config_items, intf->config_items,
                       intf->config_items_count * sizeof(struct config_item));
            }
        } else {
            return EFAILURE;
        }
    }

    if (bus->ops->attach) {
        ret = bus->ops->attach(bus, dev);
        if (ret)
            goto error;
    }
    if (dev->driver.init) {
        ret = dev->driver.init(&dev->driver, dev);
        if (ret) {
            if (bus->ops->detach) {
                bus->ops->detach(bus, dev);
                I2C_LOGI("Detach i2c device %s from bus %s intf: %s", dev->name,
                         bus->driver.reg->name, intf ? intf->reg->name : "none");
            }
            goto error;
        }
    }
    if (0 == ret) {
        if (dev->driver.read_config) {
            dev->driver.read_config(&dev->driver);
        }
        /* Add to list */
        dev->next = bus->devices;
        bus->devices = dev;
    }

error:
    return ret;
}
