#include "common.h"
#include "partition.h"
#include "spi_bus.h"
#include "spi_flash.h"
#ifdef CONFIG_SPI_FLASH_RESET_PIN
#include "gpio.h"
#endif
#include "spi_flash_sfdp.h"

#ifndef CONFIG_LOG_SPI_FLASH_LEVEL
#define CONFIG_LOG_SPI_FLASH_LEVEL CONFIG_LOG_DEFAULT_LEVEL
#endif
DECLARE_LOG_MODULE(SPI_FLASH, CONFIG_LOG_SPI_FLASH_LEVEL);

/** Macro defines START */
#define SPI_BUS_DRV_STR "spi_bus_lld_0"
#define SPI_SLAVE_ID_0 (0)
#define READ (0)
#define WRITE (1)
#define WR_EN_FLAG (0x2)
#define WIP_FLAG (0x1)
#define ERASE_ERROR_FLAG (0x20)
#define PAGE_SIZE (0x00000100)
#define PAGE_ADDR_MASK (0x000000FF)
#define TIMEOUT (5)

/* limit spi transfer to a bit less than 64k (to take into account the cmd and address fields) */
#define SPI_XFER_LIMIT 0xfff0

/* Configuartion item index*/
#define CONF_ITEM_IDX_SPI_FLASH_QUAD_ENABLED (0)
/** Macro defines END */

/** Local variables START */
typedef struct {
    uint8_t* usr_buff_ptr;
    uint32_t res_data;
    uint32_t flash_offset;
    uint32_t togo;
    spi_flash_callback_t cb;
    void* cb_param;
} flash_data_priv_t;

typedef struct {
    struct spi_bus_driver* spi_bus_drv;
    flash_data_priv_t rx_data_conf;
    flash_data_priv_t tx_data_conf;
    sfdp_data_t sfdp_info;
    uint8_t slave_id;
} spi_flash_priv_t;

static spi_flash_priv_t flash_state = {
    .spi_bus_drv = NULL,
    .rx_data_conf =
        {
            .usr_buff_ptr = NULL,
            .res_data = 0,
            .flash_offset = 0,
            .cb = NULL,
            .cb_param = NULL,
        },
    .tx_data_conf =
        {
            .usr_buff_ptr = NULL,
            .res_data = 0,
            .flash_offset = 0,
            .cb = NULL,
            .cb_param = NULL,
        },
    .slave_id = SPI_SLAVE_ID_0,
};

static struct config_item spi_flash_config_items[] = {
    {.name = "allow_quad", .default_value = 1},
};
/** Local variables END */

/** Local function declarations, START */
#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT
static int spi_flash_write_enable(spi_flash_priv_t* state);
static int spi_flash_write_disable(spi_flash_priv_t* state);
static int spi_flash_clear_flag_register(spi_flash_priv_t* state);
static int spi_flash_read_flag_register(spi_flash_priv_t* state, uint8_t* status_reg);
static int spi_flash_write_page_priv(struct spi_flash_driver* drv, uint32_t flash_addr,
                                     const uint8_t* buff, uint32_t size, spi_bus_callback_t cb);
static void spi_flash_tx_cb(struct spi_bus_driver* drv, int err_code, uint32_t size, void* param);
static int spi_flash_poll_status(spi_flash_priv_t* state, uint8_t flag, uint64_t timeout);
static int spi_flash_read_status_register(spi_flash_priv_t* state, uint8_t* status_reg);

#ifdef CONFIG_CYPRESS_FLASH_SUPPORT
static uint8_t cypress_read_reg(struct spi_flash_driver* drv, int reg);
static int cypress_write_reg(struct spi_flash_driver* drv, int reg, uint8_t val);
#endif

#endif

static int spi_flash_read_priv(struct spi_flash_driver* drv, uint32_t flash_addr, uint8_t* rd_buff,
                               uint32_t size, spi_bus_callback_t cb);
static void spi_flash_rx_cb(struct spi_bus_driver* drv, int err_code, uint32_t size, void* param);
/** Local function declarations, END */

/** API interfaces definitions, START */
static int akida_spi_flash_lld_init(struct driver_common* cmn, void* priv) {
    int rc = 0;
    spi_slave_conf_t slv_conf = {
        .clk_div = 0,
        .shift_lsb_first = false,
        .slave_sel_pol = false,
    };
    struct spi_flash_driver* spi_flash_drv = (struct spi_flash_driver*)cmn;
    if (spi_flash_drv->priv) {
        return 0; /* Already init. */
    }
    spi_flash_priv_t* state = NULL;
    spi_flash_drv->priv = priv;
    state = (spi_flash_priv_t*)(spi_flash_drv->priv);

#ifdef CONFIG_SPI_FLASH_RESET_PIN
    struct gpio_driver* gpio = (struct gpio_driver*)driver_lookup("gpio_lld");
    if (gpio) {
        gpio_config_flag_t flags = {0};
        gpio_set_config(gpio, CONFIG_SPI_FLASH_RESET_PORT, CONFIG_SPI_FLASH_RESET_PIN, GPIO_DIR_OUT,
                        flags, NULL, NULL);
        gpio_write(gpio, CONFIG_SPI_FLASH_RESET_PORT, CONFIG_SPI_FLASH_RESET_PIN, 1);
    }
#endif
    if (NULL == (state->spi_bus_drv = (struct spi_bus_driver*)driver_lookup(SPI_BUS_DRV_STR))) {
        SPI_FLASH_LOGE("SPI Controller driver instance %s not found err = %u", SPI_BUS_DRV_STR, rc);
        return EFAILURE;
    }

    if (0 != (rc = state->spi_bus_drv->driver.init(&state->spi_bus_drv->driver,
                                                   state->spi_bus_drv->driver.reg->priv))) {
        SPI_FLASH_LOGE("SPI Controller driver initilization failed err = %u", rc);
        return EFAILURE;
    }

    if ((rc = spi_slave_config_set(state->spi_bus_drv, state->slave_id, &slv_conf))) {
        SPI_FLASH_LOGE("Slave configuration failed, err = %u", rc);
        return EFAILURE;
    }

    if (sfdp_read_data(state->spi_bus_drv, state->slave_id, &(state->sfdp_info))) {
        SPI_FLASH_LOGE("SFDP parsing failed");
        return EFAILURE;
    } else {
        SPI_FLASH_LOGI("SPI flash vendor: %02x dev: %04x", state->sfdp_info.flash_info.vendor_id,
                       state->sfdp_info.flash_info.device_id);
    }

#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT
    /* Specific flash hooks */
    switch (state->sfdp_info.flash_info.vendor_id) {
#ifdef CONFIG_CYPRESS_FLASH_SUPPORT
#define CYPRESS_FLASH_CR1NV_REG_ADDRESS 2
#define CYPRESS_FLASH_CR1NV_QSPI_ENABLED_MASK 2

        case SPI_FLASH_VENDOR_CYPRESS: {
            /* Cypress flash do not enable QSPI by default, which makes romcode failing to load
             * code. Enabling QSPI in the Non Volatile Configuration register fixes the problem.
             */
            uint8_t reg = cypress_read_reg(spi_flash_drv, CYPRESS_FLASH_CR1NV_REG_ADDRESS);
            if (!(reg & CYPRESS_FLASH_CR1NV_QSPI_ENABLED_MASK)) {
                reg |= CYPRESS_FLASH_CR1NV_QSPI_ENABLED_MASK;
                if (!(rc = spi_flash_write_enable(state))) {
                    cypress_write_reg(spi_flash_drv, CYPRESS_FLASH_CR1NV_REG_ADDRESS, reg);
                }
            }
            /* display config registers */
            for (int i = 2; i < 5; i++) {
                SPI_FLASH_LOGD("reg cr%d v:%02x nv:%02x", i - 1,
                               cypress_read_reg(spi_flash_drv, 0x800000 + i),
                               cypress_read_reg(spi_flash_drv, i));
            }
        } break;
#endif
        default:
            break;
    }
#endif
    return SUCCESS;
}

static int akida_spi_flash_lld_get_info(struct spi_flash_driver* drv,
                                        spi_flash_information_t* flash_info) {
    spi_flash_priv_t* state = drv->priv;
    *flash_info = state->sfdp_info.flash_info;
    return SUCCESS;
}

static int akida_spi_flash_lld_read(struct spi_flash_driver* drv, const uint32_t flash_addr,
                                    void* rd_buff, const uint32_t size, spi_flash_callback_t cb,
                                    void* cb_param) {
    int rc = SUCCESS;
    spi_flash_priv_t* state = drv->priv;
    if (!cb) {
        uint32_t left = size;
        uint32_t read = 0;
        while (left) {
            int count = left > SPI_XFER_LIMIT ? SPI_XFER_LIMIT : left;
            if (spi_flash_read_priv(drv, flash_addr + read, ((uint8_t*)rd_buff) + read, count,
                                    NULL)) {
                SPI_FLASH_LOGE("Flash read failed, addr 0x%08x size =0x%08x %err = %u, ",
                               flash_addr, size, rc);
                return EFAILURE;
            }
            read += count;
            left -= count;
        }
    } else {
        uint32_t count = size > SPI_XFER_LIMIT ? SPI_XFER_LIMIT : size;
        state->rx_data_conf.res_data = size;
        state->rx_data_conf.usr_buff_ptr = (uint8_t*)rd_buff;
        state->rx_data_conf.flash_offset = flash_addr;
        state->rx_data_conf.cb = cb;
        state->rx_data_conf.cb_param = cb_param;
        state->rx_data_conf.togo = size - count;

        if (spi_flash_read_priv(drv, flash_addr, (uint8_t*)rd_buff, count, spi_flash_rx_cb)) {
            SPI_FLASH_LOGE("Async flash read failed, addr 0x%08x %err = %u", flash_addr, rc);
            return EFAILURE;
        } else {
            SPI_FLASH_LOGD("Async flash read , addr 0x%08x", flash_addr);
        }
    }
    return rc;
}

#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT
static int akida_spi_flash_lld_write(struct spi_flash_driver* drv, const uint32_t flash_addr,
                                     const void* buff, const uint32_t size, spi_flash_callback_t cb,
                                     void* cb_param) {
    spi_flash_priv_t* state = drv->priv;
    uint32_t offset = 0, nex_tran = 0, rem_tran = 0;
    int rc = 0;

    /** check if address falls with in a page and write from that offset to end of page */
    nex_tran = (PAGE_SIZE - (flash_addr & PAGE_ADDR_MASK)) & PAGE_ADDR_MASK;
    rem_tran = (flash_addr & PAGE_ADDR_MASK);
    nex_tran = rem_tran ? (PAGE_SIZE - rem_tran) : PAGE_SIZE;
    if (nex_tran > size) {
        nex_tran = size;
    }
    rem_tran = size;

    if (cb) {
        state->tx_data_conf.res_data = size - nex_tran;
        state->tx_data_conf.usr_buff_ptr = (uint8_t*)buff + nex_tran;
        state->tx_data_conf.flash_offset = flash_addr + nex_tran;
        state->tx_data_conf.cb = cb;
        state->tx_data_conf.cb_param = cb_param;
        if ((rc = spi_flash_write_page_priv(drv, flash_addr, buff, nex_tran, spi_flash_tx_cb))) {
            SPI_FLASH_LOGE("Error writing data at addr = 0x%08x of bytes = %u rc = %d %s():%u\n",
                           flash_addr + offset, nex_tran, rc);
            return EFAILURE;
        }

    } else {
        while (offset < size) {
            if (offset || !nex_tran) {
                /** If its not first transaction */
                nex_tran = (rem_tran > PAGE_SIZE) ? PAGE_SIZE : rem_tran;
            }
            if ((rc = spi_flash_write_page_priv(drv, flash_addr + offset, buff + offset, nex_tran,
                                                NULL))) {
                SPI_FLASH_LOGE(
                    "Error writing data at addr = 0x%08x of bytes = %u rc = %d %s():%u\n",
                    flash_addr + offset, nex_tran, rc);
                return EFAILURE;
            }
            offset += nex_tran;
            rem_tran -= nex_tran;
        }
    }

    return SUCCESS;
}

static int akida_spi_flash_lld_erase(struct spi_flash_driver* drv, const uint32_t sector_id,
                                     const uint32_t n_sectors, spi_flash_sector_type_t type,
                                     spi_flash_callback_t cb, void* cb_param) {
    (void)cb;
    (void)cb_param;
    int rc = 0;
    uint8_t cmd_addr[4] = {0, 0, 0, 0};
    uint64_t timeout = 0;
    spi_transaction_t spi_trans_conf;
    spi_transfer_t t_list[1];  // To ease the assignment of t_list as a pointer
    uint32_t flash_addr = 0, loop_var = 0;
    uint8_t status_reg;
    spi_flash_priv_t* state = drv->priv;
    switch (type) { /** Switch aids type validation*/
        case SPI_FLASH_SECTOR_TYPE_64KB:
            cmd_addr[0] = state->sfdp_info.cmds.erase_64k;
            timeout = state->sfdp_info.erase_64k_timeout;
            break;
        case SPI_FLASH_SECTOR_TYPE_32KB:
            cmd_addr[0] = state->sfdp_info.cmds.erase_32k;
            timeout = state->sfdp_info.erase_32k_timeout;
            break;
        case SPI_FLASH_SECTOR_TYPE_4KB:
            cmd_addr[0] = state->sfdp_info.cmds.erase_4k;
            timeout = state->sfdp_info.erase_4k_timeout;
            break;
        default:
            return EINVALID_PARM;
    }
    if (!cmd_addr[0]) {
        SPI_FLASH_LOGE("Erase type not supported");
        return EINVALID_PARM;
    }
    if (0 == n_sectors) {
        SPI_FLASH_LOGE("Invalid number of sectors");
        return EINVALID_PARM;
    }

    for (loop_var = 0; loop_var < n_sectors; loop_var++) {
        /** Every Erase operation must precede with Write enable operation */
        if (!(rc = spi_flash_write_enable(state))) {
            flash_addr = (sector_id + loop_var) * ((1 << type));
            SPI_FLASH_LOGD("SPI erase on sector %u, size =%u=0x%08x start addr=0x%08x ",
                           sector_id + loop_var, 1 << type, 1 << type, flash_addr);
            cmd_addr[1] = (flash_addr >> 16) & 0xFF;
            cmd_addr[2] = (flash_addr >> 8) & 0xFF;
            cmd_addr[3] = (flash_addr)&0xFF;
            spi_bus_transaction_init_write(t_list, cmd_addr, 4);
            spi_trans_conf.slave_id = state->slave_id;
            spi_trans_conf.xfer_count = 1;
            spi_trans_conf.xfer_list = t_list;
            if ((rc = spi_bus_transact(state->spi_bus_drv, &spi_trans_conf, NULL, NULL))) {
                SPI_FLASH_LOGE("SPI erase on sector %u, err = %u", sector_id + loop_var, rc);
                return EFAILURE;
            }
            if (spi_flash_poll_status(state, WIP_FLAG, timeout)) {
                SPI_FLASH_LOGE("Write in progress check has timedout");
                return ETIMEDOUT;
            }
            /** Read flash flag status register */
            if (!(rc = spi_flash_read_flag_register(state, &status_reg))) {
                /** Check erase error flag and clear error bit if set */
                if (status_reg & ERASE_ERROR_FLAG) {
                    rc = spi_flash_clear_flag_register(state);
                }
            }
            if (rc) {
                SPI_FLASH_LOGE("SPI erase on sector %u, err = %u", sector_id + loop_var, rc);
                return EFAILURE;
            }

            if (!(rc = spi_flash_write_disable(state))) {
                SPI_FLASH_LOGD("SPI WR Disable success");
            } else {
                SPI_FLASH_LOGE("SPI WR Disable failed, err = %d", rc);
            }
        } else {
            SPI_FLASH_LOGE("SPI WR Enable failed, err = %d", rc);
            return EFAILURE;
        }
    }
    return SUCCESS;
}

#ifdef CONFIG_ENABLE_BULK_ERASE
static int akida_spi_flash_lld_erase_bulk(struct spi_flash_driver* drv, spi_flash_callback_t cb,
                                          void* cb_param) {
    (void)cb;
    (void)cb_param;
    int rc = 0;
    uint64_t timeout = 0;
    spi_transaction_t spi_trans_conf;
    spi_transfer_t t_list[1];  // To ease the assignment of t_list as a pointer
    uint8_t status_reg;
    spi_flash_priv_t* state = drv->priv;

    timeout = state->sfdp_info.chip_erase_timeout;

    /** Every Erase operation must precede with Write enable operation */
    if (!(rc = spi_flash_write_enable(state))) {
        spi_bus_transaction_init_write(t_list, &state->sfdp_info.cmds.chip_erase, 1);
        spi_trans_conf.slave_id = state->slave_id;
        spi_trans_conf.xfer_count = 1;
        spi_trans_conf.xfer_list = t_list;
        if ((rc = spi_bus_transact(state->spi_bus_drv, &spi_trans_conf, NULL, NULL))) {
            SPI_FLASH_LOGE("Erasing whole SPI_FLASH failed, err = %u", rc);
            return EFAILURE;
        }
        if (spi_flash_poll_status(state, WIP_FLAG, timeout)) {
            SPI_FLASH_LOGE("Write in progress check has timedout");
            return ETIMEDOUT;
        }
        /** Read flash flag status register */
        if (!(rc = spi_flash_read_flag_register(state, &status_reg))) {
            /** Check erase error flag and clear error bit if set */
            if (status_reg & ERASE_ERROR_FLAG) {
                rc = spi_flash_clear_flag_register(state);
            }
        }
        if (rc) {
            SPI_FLASH_LOGE("Erasing whole SPI_FLASH failed, err = %u", rc);
            return EFAILURE;
        }

        if (!(rc = spi_flash_write_disable(state))) {
            SPI_FLASH_LOGD("SPI WR Disable success");
        } else {
            SPI_FLASH_LOGE("SPI WR Disable failed, err = %u", rc);
        }
    } else {
        SPI_FLASH_LOGE("SPI WR Enable failed, err = %u", rc);
        return EFAILURE;
    }
    return SUCCESS;
}
#endif
#endif
/** API interfaces definitions, END */

/** Local function definitions, START */

#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT

static int spi_flash_send_cmd(spi_flash_priv_t* state, uint8_t cmd) {
    int rc = 0;

    static spi_transaction_t trct = {0};
    static spi_transfer_t t[1];
    static uint8_t _cmd;

    if (trct.xfer_count == 0) {
        spi_bus_transaction_init_write(t, &_cmd, 1);
        trct.xfer_count = 1;
        trct.xfer_list = t;
        trct.slave_id = state->slave_id;
    }

    _cmd = cmd;

    if ((rc = spi_bus_transact(state->spi_bus_drv, &trct, NULL, NULL))) {
        SPI_FLASH_LOGE("SPI transaction failed, err = %u", rc);
        return EFAILURE;
    }
    return SUCCESS;
}
static int spi_flash_write_enable(spi_flash_priv_t* state) {
    uint8_t status_reg = 0;
    uint64_t exceed_time = 0;

    int rc = spi_flash_send_cmd(state, state->sfdp_info.cmds.wr_enable);

    if (rc) {
        return rc;
    }

    exceed_time = TIMEOUT + get_time_ms();
    while (!(spi_flash_read_status_register(state, &status_reg))) {
        if ((status_reg & WR_EN_FLAG) || (get_time_ms() > exceed_time)) {
            break;
        }
    }
    if (!(status_reg & WR_EN_FLAG)) {
        SPI_FLASH_LOGE("Write enable check has timedout %x", status_reg);
        return EBAD_STATE;
    }
    return SUCCESS;
}

static int spi_flash_write_disable(spi_flash_priv_t* state) {
    int rc = spi_flash_send_cmd(state, state->sfdp_info.cmds.wr_disable);

    if (rc) {
        return rc;
    }

    if (spi_flash_poll_status(state, WR_EN_FLAG, TIMEOUT)) {
        SPI_FLASH_LOGE("Write disable check has timedout");
        return ETIMEDOUT;
    }
    return SUCCESS;
}

static int spi_flash_read_register(spi_flash_priv_t* state, uint8_t* status_reg, uint8_t reg) {
    int rc = 0;
    static spi_transaction_t trct;
    static spi_transfer_t t[2];
    static uint8_t _reg;

    if (trct.xfer_count == 0) {
        trct.xfer_count = 2;
        trct.xfer_list = t;
        trct.slave_id = state->slave_id;
        spi_bus_transaction_init_write(t, &_reg, 1);
        spi_bus_transaction_init_read(t + 1, status_reg, 1);
    }
    _reg = reg;
    t[1].rxd = status_reg;

    if ((rc = spi_bus_transact(state->spi_bus_drv, &trct, NULL, NULL))) {
        SPI_FLASH_LOGE("SPI transaction failed, err = %u", rc);
        return EFAILURE;
    }

    return SUCCESS;
}

static int spi_flash_read_flag_register(spi_flash_priv_t* state, uint8_t* status_reg) {
    return spi_flash_read_register(state, status_reg, state->sfdp_info.cmds.read_flag_status_reg);
}

/** writing the clear flag register command onto the SPI BUS will auto reset the erase error bits */
static int spi_flash_clear_flag_register(spi_flash_priv_t* state) {
    return spi_flash_send_cmd(state, state->sfdp_info.cmds.clear_flag_status_reg);
}

/** Reads the data of a given register command */

static int spi_flash_read_status_register(spi_flash_priv_t* state, uint8_t* status_reg) {
    return spi_flash_read_register(state, status_reg, state->sfdp_info.cmds.read_status_reg);
}
#endif

static int spi_flash_read_priv(struct spi_flash_driver* drv, uint32_t flash_addr, uint8_t* buff,
                               uint32_t size, spi_bus_callback_t cb) {
    uint8_t cmd_addr[4] = {0, 0, 0, 0};
    int rc = 0;
    spi_transaction_t spi_trans_conf;
    spi_transfer_t t_list[3];
    spi_flash_priv_t* state = ((spi_flash_priv_t*)(drv->priv));
    cmd_addr[0] = state->sfdp_info.cmds.read;
    cmd_addr[1] = (flash_addr >> 16) & 0xFF;
    cmd_addr[2] = (flash_addr >> 8) & 0xFF;
    cmd_addr[3] = (flash_addr)&0xFF;

    spi_bus_transaction_init_write(t_list, cmd_addr, 4);

    if (DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_SPI_FLASH_QUAD_ENABLED).current_value &&
        state->sfdp_info.flash_info.is_quad_rd_supp) {
        spi_bus_transaction_init_write(t_list + 1, NULL, state->sfdp_info.read_dummy_cycles);
        spi_bus_transaction_init_read(t_list + 2, buff, size);
        t_list[2].mode = SPI_XFER_QUAD_SPI_MODE;
        cmd_addr[0] = state->sfdp_info.cmds.quad_read;
        spi_trans_conf.xfer_count = 3;
    } else {
        spi_bus_transaction_init_read(t_list + 1, buff, size);
        spi_trans_conf.xfer_count = 2;
    }

    SPI_FLASH_LOGD("flash read add = 0x%08x, buf_ptr = 0x%08x size =%u cmd=0x%02x", flash_addr,
                   buff, size, cmd_addr[0]);
    spi_trans_conf.slave_id = state->slave_id;
    spi_trans_conf.xfer_list = t_list;
    if (cb) {
        if ((rc = spi_bus_transact(state->spi_bus_drv, &spi_trans_conf, cb, drv))) {
            SPI_FLASH_LOGE("Async SPI transaction failed, err = %u", rc);
            return EFAILURE;
        }
    } else {
        if ((rc = spi_bus_transact(state->spi_bus_drv, &spi_trans_conf, NULL, NULL))) {
            SPI_FLASH_LOGE("Sync SPI transaction failed, err = %u", rc);
            return EFAILURE;
        }
    }
    return SUCCESS;
}

#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT
static int spi_flash_write_page_priv(struct spi_flash_driver* drv, uint32_t flash_addr,
                                     const uint8_t* buff, uint32_t size, spi_bus_callback_t cb) {
    uint8_t cmd_addr[4] = {0, 0, 0, 0};
    int rc = 0;
    spi_transaction_t spi_trans_conf;
    spi_transfer_t t_list[2];
    spi_flash_priv_t* state = ((spi_flash_priv_t*)(drv->priv));
    if (spi_flash_poll_status(state, WIP_FLAG, TIMEOUT)) {
        SPI_FLASH_LOGE("Write page check has timedout");
        return ETIMEDOUT;
    }

    if (!(rc = spi_flash_write_enable(state))) {
        cmd_addr[0] = state->sfdp_info.cmds.write;
        cmd_addr[1] = (flash_addr >> 16) & 0xFF;
        cmd_addr[2] = (flash_addr >> 8) & 0xFF;
        cmd_addr[3] = (flash_addr)&0xFF;
        spi_bus_transaction_init_write(t_list, cmd_addr, 4);
        spi_bus_transaction_init_write(t_list + 1, buff, size);
        if (DRIVER_CONFIG_GET_ITEM(drv, CONF_ITEM_IDX_SPI_FLASH_QUAD_ENABLED).current_value &&
            state->sfdp_info.flash_info.is_quad_rd_supp) {
            cmd_addr[0] = state->sfdp_info.cmds.quad_write;
            t_list[1].mode = SPI_XFER_QUAD_SPI_MODE;
        }
        SPI_FLASH_LOGD(
            "Flash write attempted on addr 0x%08x, size 0x%08x = %u, cmd=0x%02x, timeout=%ums",
            flash_addr, size, size, cmd_addr[0], state->sfdp_info.page_write_timeout);
        spi_trans_conf.slave_id = state->slave_id;
        spi_trans_conf.xfer_count = 2;
        spi_trans_conf.xfer_list = t_list;
        if ((rc = spi_bus_transact(state->spi_bus_drv, &spi_trans_conf, cb, drv))) {
            SPI_FLASH_LOGE("SPI transaction failed, err = %u", rc);
            return EFAILURE;
        }
        if (cb == NULL) {
            if (spi_flash_poll_status(state, WIP_FLAG, state->sfdp_info.page_write_timeout)) {
                SPI_FLASH_LOGE("Write page check has timedout");
                return ETIMEDOUT;
            }
        }
    }

    return SUCCESS;
}
#endif

static void spi_flash_rx_cb(struct spi_bus_driver* drv, int err_code, uint32_t size, void* param) {
    struct spi_flash_driver* flash_drv = NULL;
    flash_drv = param;
    (void)drv;
    spi_flash_priv_t* state = flash_drv->priv;
    if (state->rx_data_conf.togo) {
        uint32_t count =
            state->rx_data_conf.togo < SPI_XFER_LIMIT ? SPI_XFER_LIMIT : state->rx_data_conf.togo;
        state->rx_data_conf.usr_buff_ptr += SPI_XFER_LIMIT;
        state->rx_data_conf.flash_offset += SPI_XFER_LIMIT;
        state->rx_data_conf.togo -= count;
        if (spi_flash_read_priv(flash_drv, state->rx_data_conf.flash_offset,
                                state->rx_data_conf.usr_buff_ptr, count, spi_flash_rx_cb)) {
            state->rx_data_conf.cb(flash_drv, EFAILURE, size + state->rx_data_conf.res_data,
                                   state->rx_data_conf.cb_param);
        }
        return;
    }
    if (!err_code && !size) {
        state->rx_data_conf.cb(flash_drv, 0, 0, state->rx_data_conf.cb_param);
    } else {
        state->rx_data_conf.cb(flash_drv, EFAILURE, size + state->rx_data_conf.res_data,
                               state->rx_data_conf.cb_param);
    }
}

#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT
static void spi_flash_tx_cb(struct spi_bus_driver* drv, int err_code, uint32_t size, void* param) {
    int rc = 0;
    uint32_t cur_size = 0, offset = 0;
    uint8_t* buff_ptr = NULL;
    struct spi_flash_driver* flash_drv = NULL;
    (void)drv;
    flash_drv = param;
    spi_flash_priv_t* state = flash_drv->priv;
    if (spi_flash_poll_status(state, WIP_FLAG, TIMEOUT)) {
        SPI_FLASH_LOGE("Write page check has timedout");
        state->tx_data_conf.cb(flash_drv, EFAILURE, size + state->tx_data_conf.res_data,
                               state->tx_data_conf.cb_param);
    }
    if (!err_code && !size) {
        if (state->tx_data_conf.res_data) {
            cur_size = (state->tx_data_conf.res_data > PAGE_SIZE) ? PAGE_SIZE
                                                                  : state->tx_data_conf.res_data;
            state->tx_data_conf.res_data -= cur_size;
            offset = state->tx_data_conf.flash_offset;
            buff_ptr = state->tx_data_conf.usr_buff_ptr;
            state->tx_data_conf.usr_buff_ptr = state->tx_data_conf.usr_buff_ptr + cur_size;
            state->tx_data_conf.flash_offset = state->tx_data_conf.flash_offset + cur_size;
            if ((rc = spi_flash_write_page_priv(param, offset, buff_ptr, cur_size,
                                                spi_flash_tx_cb))) {
                state->tx_data_conf.cb(flash_drv, EFAILURE, cur_size + state->tx_data_conf.res_data,
                                       state->tx_data_conf.cb_param);
            }
        } else {
            state->tx_data_conf.cb(flash_drv, 0, 0, state->tx_data_conf.cb_param);
        }
    } else {
        state->tx_data_conf.cb(flash_drv, EFAILURE, size + state->tx_data_conf.res_data,
                               state->tx_data_conf.cb_param);
    }
}

static int spi_flash_poll_status(spi_flash_priv_t* state, uint8_t flag, uint64_t timeout) {
    uint64_t exceed_time = timeout + get_time_ms();
    uint8_t status_reg;
    do {
        int ret = spi_flash_read_status_register(state, &status_reg);
        if (ret)
            return ret;
        if (get_time_ms() > exceed_time)
            return ETIMEDOUT;
    } while ((status_reg & flag));
    return 0;
}

#endif

static int spi_flash_read_config(struct driver_common* spi_flash_drv) {
    DRIVER_CONFIG_SET_CURRENT(
        spi_flash_drv, CONF_ITEM_IDX_SPI_FLASH_QUAD_ENABLED,
        DRIVER_CONFIG_GET_ITEM(spi_flash_drv, CONF_ITEM_IDX_SPI_FLASH_QUAD_ENABLED).default_value);
    return SUCCESS;
}

static int spi_flash_apply_config(struct driver_common* cmn) {
    struct spi_flash_driver* spi_flash_drv = (struct spi_flash_driver*)cmn;
    spi_flash_priv_t* state = NULL;
    state = (spi_flash_priv_t*)(spi_flash_drv->priv);

    if (state->sfdp_info.flash_info.is_quad_rd_supp) {
        DRIVER_CONFIG_APPLY(spi_flash_drv, CONF_ITEM_IDX_SPI_FLASH_QUAD_ENABLED);
    }
    return SUCCESS;
}

#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT
#ifdef CONFIG_CYPRESS_FLASH_SUPPORT
#define CYPRESS_FLASH_CMD_WRAR 0x71 /* WRAR Write Any Register */
#define CYPRESS_FLASH_CMD_RDAR 0x65 /* RDAR Read Any Register */
static uint8_t cypress_read_reg(struct spi_flash_driver* drv, int reg) {
    spi_flash_priv_t* state = drv->priv;
    uint8_t cmd = CYPRESS_FLASH_CMD_RDAR;
    uint8_t val = 0;
    spi_transfer_t t_list[3];
    spi_transaction_t spi_trans_conf;
    uint8_t cmd_addr[4];
    cmd_addr[0] = (reg >> 16) & 0xff;
    cmd_addr[1] = (reg >> 8) & 0xff;
    cmd_addr[2] = reg & 0xff;
    cmd_addr[3] = 0;

    spi_bus_transaction_init_write(t_list, &cmd, 1);
    spi_bus_transaction_init_write(t_list + 1, cmd_addr, 4);
    spi_bus_transaction_init_read(t_list + 2, &val, 1);

    spi_trans_conf.slave_id = state->slave_id;
    spi_trans_conf.xfer_count = 3;
    spi_trans_conf.xfer_list = t_list;
    int rc;
    if ((rc = spi_bus_transact(state->spi_bus_drv, &spi_trans_conf, NULL, NULL))) {
        SPI_FLASH_LOGE("SFDP read ID failed, err = %u", rc);
        return EFAILURE;
    }
    return val;
}

static int cypress_write_reg(struct spi_flash_driver* drv, int reg, uint8_t val) {
    spi_flash_priv_t* state = drv->priv;
    uint8_t cmd = CYPRESS_FLASH_CMD_WRAR;
    spi_transfer_t t_list[3];
    spi_transaction_t spi_trans_conf;
    uint8_t cmd_addr[4];
    cmd_addr[0] = (reg >> 16) & 0xff;
    cmd_addr[1] = (reg >> 8) & 0xff;
    cmd_addr[2] = reg & 0xff;

    spi_bus_transaction_init_write(t_list, &cmd, 1);
    spi_bus_transaction_init_write(t_list + 1, cmd_addr, 3);
    spi_bus_transaction_init_write(t_list + 2, &val, 1);

    spi_trans_conf.slave_id = state->slave_id;
    spi_trans_conf.xfer_count = 3;
    spi_trans_conf.xfer_list = t_list;
    int rc;
    if ((rc = spi_bus_transact(state->spi_bus_drv, &spi_trans_conf, NULL, NULL))) {
        SPI_FLASH_LOGE("SFDP read ID failed, err = %u", rc);
        return EFAILURE;
    }
    return val;
}
#endif
#endif

/** Local function definitions, END */

struct spi_flash_ops akida_spi_flash_lld_ops = {
    .get_info = akida_spi_flash_lld_get_info,
    .read = akida_spi_flash_lld_read,
#ifndef CONFIG_FLASH_DISABLE_WRITE_SUPPORT
    .write = akida_spi_flash_lld_write,
    .erase = akida_spi_flash_lld_erase,
#ifdef CONFIG_ENABLE_BULK_ERASE
    .erase_bulk = akida_spi_flash_lld_erase_bulk,
#endif
#endif
};

DECLARE_SPI_FLASH(spi_flash_lld_0, &akida_spi_flash_lld_ops, akida_spi_flash_lld_init,
                  spi_flash_read_config, spi_flash_apply_config, spi_flash_config_items,
                  (sizeof(spi_flash_config_items) / sizeof(spi_flash_config_items[0])),
                  &flash_state);

#ifdef CONFIG_PARTITIONS
static int spi_flash_disk_read(struct partition* p, uint32_t start, uint8_t* buffer,
                               uint32_t size) {
    elog_msg(NULL, 0, "read @ %x (offset %x)", p->offset + start, start);
    struct spi_flash_driver* f = (struct spi_flash_driver*)p->disk->priv;
    int ret = akida_spi_flash_lld_read(f, p->offset + start, buffer, size, NULL, NULL);
    return ret;
}

static int spi_flash_disk_write(struct partition* p, uint32_t start, uint8_t* buffer,
                                uint32_t size) {
    int err = 0;
    struct spi_flash_driver* f = (struct spi_flash_driver*)p->disk->priv;
    uint32_t left = size;
    uint32_t addr = start;
    elog_msg(NULL, 0, "write @ %x", p->offset + start);

    while (left) {
        uint32_t towrite = left > p->write_size ? p->write_size : left;
        err = akida_spi_flash_lld_write(f, p->offset + addr, buffer, towrite, NULL, NULL);
        left -= towrite;
        buffer += towrite;
        addr += towrite;
    }

    return err;
}

static int spi_flash_disk_erase(struct partition* p, uint32_t start, uint32_t count) {
    int phy_sector = (p->offset + start) / p->block_size;
    int phy_count = (count / p->erase_size + (count % p->erase_size ? 1 : 0));
    int err = 0;
    struct spi_flash_driver* f = (struct spi_flash_driver*)p->disk->priv;

    for (int i = 0; i < phy_count; i++) {
        err = akida_spi_flash_lld_erase(f, phy_sector, 1, SPI_FLASH_SECTOR_TYPE_4KB, NULL, NULL);
    }
    return err;
}

static int spi_flash_disk_init(struct driver_common* drv, void* priv) {
    struct disk* d = (struct disk*)drv;
    int err = 0;
    if (d->priv)
        return err;

    struct spi_flash_driver* f = (struct spi_flash_driver*)driver_lookup((char*)priv);
    if (f) {
        err = f->driver.init(&f->driver, f->driver.reg->priv);
        d->priv = f;
        for (int i = 0; i < d->partition_count; i++) {
            d->partitions[i].disk = d;
        }
    } else {
        err = E_NORESOURCEAVAILABLE;
    }

    return err;
}

struct partition spi_flash_partitions[] = {
    {.block_size = 0x1000,
     .write_size = 4,
     .erase_size = 0x1000,
     .offset = 0,
     .size = 0x2000,
     .name = "fsbl"},
    {.block_size = 0x1000,
     .write_size = 4,
     .erase_size = 0x1000,
     .offset = 0x2000,
     .size = (CONFIG_PROPERTIES_BLOCK_START - 2) * 0x1000,
     .name = "ssbl"},
    {.block_size = 0x1000,
     .write_size = 4,
     .erase_size = 0x1000,
     .offset = CONFIG_PROPERTIES_BLOCK_START * 0x1000,
     .size = CONFIG_PROPERTIES_BLOCK_COUNT * 0x1000,
     .name = "properties"},
    {.block_size = 0x1000,
     .write_size = 4,
     .erase_size = 0x1000,
     .offset = CONFIG_FILESYSTEM_START_SECTOR * 0x1000,
     .size = CONFIG_FILESYSTEM_SECTOR_COUNT * 0x1000,
     .name = "fatfs"},
};

struct disk spi_flash_disk = {
    .drv.init = spi_flash_disk_init,
    .read = spi_flash_disk_read,
    .write = spi_flash_disk_write,
    .erase = spi_flash_disk_erase,
    .partition_count = sizeof(spi_flash_partitions) / sizeof(spi_flash_partitions[0]),
    .partitions = spi_flash_partitions,
};

DECLARE_DISK(spi_flash, &spi_flash_disk.drv, "spi_flash_lld_0");
#endif