#include "spi_flash_sfdp.h"
#include "common.h"
#include "spi_bus.h"
#include "spi_flash.h"

#define DUMMY_CYCLES_TO_BYTES(trp, cycles) (cycles >> (3 - trp))

/**
 * Flash Command definitions
 **/
#define FLASH_CMD_RD_STAT_REG (0x5)
#define FLASH_CMD_RD_FLAG_STAT_REG (0x70)
#define FLASH_CMD_CLR_FLAG_STAT_REG (0x50)
#define FLASH_CMD_WR_ENABLE (0x6)
#define FLASH_CMD_WR_DISABLE (0x4)
#define FLASH_CMD_RD_ID (0x9F)
#define FLASH_CMD_QUAD_RD (0x6B)
#define FLASH_CMD_QUAD_WR (0x32)
#define FLASH_CMD_RD (0x3)
#define FLASH_CMD_WR (0x2)
#define FLASH_CMD_4KB_ERASE (0x20)
#define FLASH_CMD_32KB_ERASE (0x52)
#define FLASH_CMD_64KB_ERASE (0xD8)
#if CONFIG_ENABLE_BULK_ERASE
#define FLASH_CMD_WHOLE_FLASH_ERASE (0XC7)
#endif

/* Read SFDP command ID*/
#define FLASH_CMD_RD_SFDP (0x5A)
/* Double word length*/
#define DW_LENGTH (4)
/* Number of parameter header index in SFDP header */
#define NPH_INDEX (6)
/* SFDP parameter header 0 offset*/
#define SFDP_PARAM_0_OFFSET (2 * DW_LENGTH)
#define N_SECTOR_TYPES (3)
/* Byte and Bit offsets of various information in parameter 0 table */
#define OFFSET_FLASH_SIZE (0x4)
#define OFFSET_READ_SUPPORT (0x2)
#define OFFSET_QUAD_READ_CMD (0xB)
#define OFFSET_SECTOR_1_SIZE (0x1C)

#define OFFSET_ERASE_TIME (0x24)
#define OFFSET_BIT_SECTOR_1_TYPICAL_TIME (4)
#define OFFSET_BIT_SECTOR_1_TIME_UNIT (9)
#define OFFSET_BIT_SECTOR_2_TYPICAL_TIME (11)
#define OFFSET_BIT_SECTOR_2_TIME_UNIT (16)
#define OFFSET_BIT_SECTOR_3_TYPICAL_TIME (18)
#define OFFSET_BIT_SECTOR_3_TIME_UNIT (23)

#define OFFSET_PAGE_WRITE_TIME (0x28)
#define OFFSET_BIT_PAGE_WRITE_TYPICAL_TIME (8)
#define OFFSET_BIT_PAGE_WRITE_TIME_UNIT (13)
#define OFFSET_BIT_CHIP_ERASE_TYPICAL_TIME (24)
#define OFFSET_BIT_CHIP_ERASE_TIME_UNIT (29)

#define TIME_MULTIPLIER_WIDTH (4)
#define TYPICAL_TIME_WIDTH (5)
#define TIME_UNIT_WIDTH (2)

/** Generic macro to generate bitfield mask of a given length */
#define BIT_MASK(len) ((1 << len) - 1)

typedef struct {
    uint8_t id;
    uint8_t minor_rev;
    uint8_t major_rev;
    uint8_t length;
    uint32_t data_offset : 24;
} sfdp_param_header;

/**
 * Function to get timing information of erase, page write and supported erase commands.
 *
 * @param sfdp_data output data structure to be updated with requested data.
 * @pram buff buffer which points to SFDP parameter table 0, to be parsed for aforementioned
 * information.
 *
 * @return 0 in case of success else non-negative value in case of error
 */
static int get_erase_info(sfdp_data_t* sfdp_data, uint8_t* buff);

/**
 * Function to get Flash ID information.
 *
 * @param spi_bus_drv driver instance of SPI Bus driver
 * @param slave_id Slave Id of the flash
 * @param sfdp_data output data structure to be updated with parsed data.
 *
 * @return 0 in case of success else non-negative value in case of error
 */
static int get_id_info(struct spi_bus_driver* spi_bus_drv, uint8_t slave_id,
                       sfdp_data_t* sfdp_data);
/**
 * Function to decode timeunit
 *
 * @param resol Input parameter represents time unit.
 *
 * @return return decoded time unit in milliseconds.
 **/
static uint32_t get_msec(uint32_t resol);

/**
 * Function to read data from SFDP table
 *
 * @param spi_bus_drv driver instance of SPI Bus driver
 * @param slave_id Slave Id of the flash
 * @param buff Output buffer to be udpated with read SFDP data
 * @param offset Address offset with in SFDP table.
 * @param size size of the data requested
 *
 * @return 0 in case of success else non-negative value in case of error
 */
static int flash_sfdp_priv(struct spi_bus_driver* spi_bus_drv, uint8_t slave_id, void* buff,
                           uint32_t offset, uint32_t size);

/**
 * Function to set sfdp data structure to meaningfull default values
 *
 * @param sfdp_info output data structure to be updated by the function.
 */
static void sfdp_default_config(sfdp_data_t* sfdp_info);

int sfdp_read_data(struct spi_bus_driver* spi_bus_drv, uint8_t slave_id, sfdp_data_t* sfdp_info) {
    int rc = SUCCESS, nph = -1, offset = 0;
    uint8_t buff[0x100];
    sfdp_param_header param_0_hdr;

    sfdp_default_config(sfdp_info);

    /* Read flash ID */
    if (get_id_info(spi_bus_drv, slave_id, sfdp_info)) {
        SPI_FLASH_LOGE("SFDP read ID failed, err = %u", rc);
    }

    /* Read SFDP header, (spans over two Double words(8bytes) DW0, DW1*/
    if ((rc = flash_sfdp_priv(spi_bus_drv, slave_id, buff, 0x0, 2 * DW_LENGTH))) {
        SPI_FLASH_LOGE("SFDP read failed, err = %u", rc);
        return rc;
    }
    /* Check for SFPD signature*/
    if (strncmp((char*)buff, "SFDP", 4)) {
        SPI_FLASH_LOGE("Not an SFDP compliant flash, err = %u", rc);
        return EFAILURE;
    }
    /* Deduct number of parameter headers*/
    nph = buff[NPH_INDEX];  // Value = 1 => two parameter headers
    if (nph >= 0) {
        /* Read parameter header 0, it is available after SFDP header */
        memset(&param_0_hdr, 0, sizeof(sfdp_param_header));
        if ((rc = flash_sfdp_priv(spi_bus_drv, slave_id, &param_0_hdr, SFDP_PARAM_0_OFFSET,
                                  sizeof(sfdp_param_header)))) {
            SPI_FLASH_LOGE("SFDP parameter header 0 read failed, err = %u", rc);
            return rc;
        }
        SPI_FLASH_LOGD("param_0 offset %08x,2Y len = %u", param_0_hdr.data_offset,
                       param_0_hdr.length * DW_LENGTH);
        /* Read parameter table 0 data */
        memset(buff, 0, param_0_hdr.length * DW_LENGTH);
        if ((rc = flash_sfdp_priv(spi_bus_drv, slave_id, buff, param_0_hdr.data_offset,
                                  param_0_hdr.length * DW_LENGTH))) {
            SPI_FLASH_LOGE("SFDP parameter 0 data read failed, err = %u", rc);
            return rc;
        }
        offset = OFFSET_FLASH_SIZE;
        sfdp_info->flash_info.flash_size = buff[offset] | (buff[offset + 1] << 8) |
                                           (buff[offset + 2] << 16) | (buff[offset + 3] << 24);
        /* Flags suggest if Quad and Dual reads are supported */
        offset = OFFSET_READ_SUPPORT;
        sfdp_info->flash_info.is_quad_rd_supp = (buff[offset] & 0x40) ? true : false;
        sfdp_info->flash_info.is_dual_rd_supp = (buff[offset] & 0x1) ? true : false;

        if (get_erase_info(sfdp_info, buff)) {
            SPI_FLASH_LOGE("Error in Parsing timeout information");
            return EFAILURE;
        } else {
            SPI_FLASH_LOGD(
                "rd = 0x%02x\nwr= 0x%02x\ners= 0x%02x-0x%02x-0x%02x\ntime= "
                "%ums-%ums-%ums-%ums-%ums",
                sfdp_info->cmds.read, sfdp_info->cmds.write, sfdp_info->cmds.erase_4k,
                sfdp_info->cmds.erase_32k, sfdp_info->cmds.erase_64k, sfdp_info->erase_4k_timeout,
                sfdp_info->erase_32k_timeout, sfdp_info->erase_64k_timeout,
                sfdp_info->page_write_timeout, sfdp_info->chip_erase_timeout);
        }
    }
    return SUCCESS;
}

static int get_erase_info(sfdp_data_t* sfdp_info, uint8_t* buff) {
    int offset = 0, loop_var = 0;
    uint8_t time_bit_offset[N_SECTOR_TYPES] = {OFFSET_BIT_SECTOR_1_TYPICAL_TIME,
                                               OFFSET_BIT_SECTOR_2_TYPICAL_TIME,
                                               OFFSET_BIT_SECTOR_3_TYPICAL_TIME};
    uint8_t resol_bit_offset[N_SECTOR_TYPES] = {OFFSET_BIT_SECTOR_1_TIME_UNIT,
                                                OFFSET_BIT_SECTOR_2_TIME_UNIT,
                                                OFFSET_BIT_SECTOR_3_TIME_UNIT};
    uint32_t data = 0, multiplier = 0, typical_time = 0, size = 0, max_time = 0;
    offset = OFFSET_ERASE_TIME;
    data = *((uint32_t*)(buff + offset));
    multiplier = data & BIT_MASK(TIME_MULTIPLIER_WIDTH);

    for (loop_var = 0; loop_var < N_SECTOR_TYPES; loop_var++) {
        offset = OFFSET_SECTOR_1_SIZE + loop_var * 2;
        size = buff[offset];
        typical_time =
            (((data >> time_bit_offset[loop_var]) & (BIT_MASK(TYPICAL_TIME_WIDTH))) + 1) *
            get_msec((data >> resol_bit_offset[loop_var]) & BIT_MASK(TIME_UNIT_WIDTH));
        max_time = 2 * (multiplier + 1) * typical_time;
        switch (size) {
            case SPI_FLASH_SECTOR_TYPE_4KB:
                /* 4KB sector*/
                sfdp_info->cmds.erase_4k = buff[offset + 1];
                sfdp_info->erase_4k_timeout = max_time;
                break;
            case SPI_FLASH_SECTOR_TYPE_32KB:
                /* 32KB sector*/
                sfdp_info->cmds.erase_32k = buff[offset + 1];
                sfdp_info->erase_32k_timeout = max_time;
                break;
            case SPI_FLASH_SECTOR_TYPE_64KB:
                /* 64KB sector*/
                sfdp_info->cmds.erase_64k = buff[offset + 1];
                sfdp_info->erase_64k_timeout = max_time;
                break;
            default:
                SPI_FLASH_LOGE("Unsupported Sector Type");
                return EFAILURE;
        }
    }
    offset = OFFSET_PAGE_WRITE_TIME;
    data = *((uint32_t*)(buff + offset));
    multiplier = data & BIT_MASK(TIME_MULTIPLIER_WIDTH);
    /* Extracting page write time information, Time unit information of page write is of 1-bit
     * length only */
    typical_time =
        (((data >> OFFSET_BIT_PAGE_WRITE_TYPICAL_TIME) & (BIT_MASK(TYPICAL_TIME_WIDTH))) + 1) *
        get_msec((data >> OFFSET_BIT_PAGE_WRITE_TIME_UNIT) & (BIT_MASK(TIME_UNIT_WIDTH) - 1));
    sfdp_info->page_write_timeout = 2 * (multiplier + 1) * typical_time;
    /* Extracting Chip erase time information */
    typical_time =
        (((data >> OFFSET_BIT_CHIP_ERASE_TYPICAL_TIME) & (BIT_MASK(TYPICAL_TIME_WIDTH))) + 1) *
        get_msec((data >> OFFSET_BIT_CHIP_ERASE_TIME_UNIT) & (BIT_MASK(TIME_UNIT_WIDTH)));
    sfdp_info->chip_erase_timeout = 2 * (multiplier + 1) * typical_time;
    return SUCCESS;
}

static uint32_t get_msec(uint32_t resol) {
    switch (resol) {
        case 0:
            return 1;
        case 1:
            return 16;
        case 2:
            return 128;
        case 3:
            return 1000;
        default:
            return 0;
    }
}

static int flash_sfdp_priv(struct spi_bus_driver* spi_bus_drv, uint8_t slave_id, void* buff,
                           uint32_t offset, uint32_t size) {
    uint8_t cmd_addr[4] = {0, 0, 0, 0};
    int rc = 0;
    spi_transaction_t spi_trans_conf;
    spi_transfer_t t_list[3];

    cmd_addr[0] = FLASH_CMD_RD_SFDP;
    cmd_addr[1] = (offset >> 16) & 0xFF;
    cmd_addr[2] = (offset >> 8) & 0xFF;
    cmd_addr[3] = (offset)&0xFF;
    spi_bus_transaction_init_write(t_list, cmd_addr, 4);
    spi_bus_transaction_init_write(t_list + 1, NULL,
                                   1); /** SFDP command needs 1 DUMMY byte in 1-bit spi mode*/
    spi_bus_transaction_init_read(t_list + 2, buff, size);

    spi_trans_conf.slave_id = slave_id;
    spi_trans_conf.xfer_count = 3;
    spi_trans_conf.xfer_list = t_list;
    if ((rc = spi_bus_transact(spi_bus_drv, &spi_trans_conf, NULL, NULL))) {
        SPI_FLASH_LOGE("SPI transaction failed, err = %u", rc);
        return EFAILURE;
    }
    return SUCCESS;
}

static int get_id_info(struct spi_bus_driver* spi_bus_drv, uint8_t slave_id,
                       sfdp_data_t* sfdp_info) {
    int rc = 0;
    uint8_t buff[4], cmd = 0;
    spi_transaction_t spi_trans_conf;
    spi_transfer_t t_list[2];

    memset(buff, 0, 4);
    cmd = sfdp_info->cmds.read_id;
    spi_bus_transaction_init_write(t_list, &cmd, 1);
    spi_bus_transaction_init_read(t_list + 1, buff, 4);

    spi_trans_conf.slave_id = slave_id;
    spi_trans_conf.xfer_count = 2;
    spi_trans_conf.xfer_list = t_list;
    if ((rc = spi_bus_transact(spi_bus_drv, &spi_trans_conf, NULL, NULL))) {
        SPI_FLASH_LOGE("SFDP read ID failed, err = %u", rc);
        return EFAILURE;
    }
    sfdp_info->flash_info.vendor_id = buff[0];
    sfdp_info->flash_info.device_id = ((buff[1] << 8) | buff[2]);

    return SUCCESS;
}

static void sfdp_default_config(sfdp_data_t* sfdp_info) {
    sfdp_info->cmds.read_sfdp = FLASH_CMD_RD_SFDP;
    sfdp_info->cmds.read_id = FLASH_CMD_RD_ID;
    sfdp_info->cmds.read = FLASH_CMD_RD;
    sfdp_info->cmds.write = FLASH_CMD_WR;
    sfdp_info->cmds.quad_read = FLASH_CMD_QUAD_RD;
    sfdp_info->cmds.quad_write = FLASH_CMD_QUAD_WR;
    sfdp_info->cmds.wr_enable = FLASH_CMD_WR_ENABLE;
    sfdp_info->cmds.wr_disable = FLASH_CMD_WR_DISABLE;
    sfdp_info->cmds.read_status_reg = FLASH_CMD_RD_STAT_REG;
    sfdp_info->cmds.read_flag_status_reg = FLASH_CMD_RD_FLAG_STAT_REG;
    sfdp_info->cmds.clear_flag_status_reg = FLASH_CMD_CLR_FLAG_STAT_REG;
#if CONFIG_ENABLE_BULK_ERASE
    sfdp_info->cmds.chip_erase = FLASH_CMD_WHOLE_FLASH_ERASE;
#endif
    sfdp_info->cmds.erase_4k = 0;
    sfdp_info->cmds.erase_32k = 0;
    sfdp_info->cmds.erase_64k = 0;
    sfdp_info->read_dummy_cycles = DUMMY_CYCLES_TO_BYTES(SPI_XFER_EXTENDED_SPI_MODE, 8);
}
