#include <stdarg.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>

#ifndef __CLI_CMD_H__
#define __CLI_CMD_H__

/**
 * @addtogroup infra
 * @{
 *
 * @addtogroup cli_cmd command line interface
 * contains the interface to implement command line interface commands
 *
 * @{
 */

struct cli_cmd_ctx;
struct cli_ctx;

struct cli_cmd_handler {
    const char* name;
    int (*handler)(int argc, char** argv, struct cli_cmd_ctx* ctx);
};

#define TTY_STATE_NORMAL 0
#define TTY_STATE_ESCAPE 1
#define TTY_STATE_ESCAPE_2 2

struct cli_cmd_ctx;
/**
 * cli input/output functions structure
 */
struct cli_ctx {
    int (*vprintf)(struct cli_ctx*, const char* fmt, va_list args);
    int (*get_char)(struct cli_ctx*);
    void (*puts)(struct cli_ctx*, char* s);
    void (*put)(struct cli_ctx*, int c);
    int (*write)(struct cli_ctx*, uint8_t* buf, int size);
    int (*read)(struct cli_ctx*, uint8_t* buf, int size);
    char buffer[256];
    int cmd_idx;
    int state;
    int bin_mode;

    /** pointer to cli execution context */
    struct cli_cmd_ctx* ctx;
};

/**
 * Cli execution context structure.
 */
struct cli_cmd_ctx {
    /** CLI input / output functions */
    struct cli_ctx* cli;
    /** Current command handler struture */
    struct cli_cmd_handler* h;
    /** Last command exection result code */
    int ans;
};

struct cli_channel_handler {
    char* name;
    struct cli_ctx* ctx;
    int (*init)(struct cli_channel_handler*);
    void* priv;
};

#ifndef CONFIG_CLI
#define CLI_HANDLER_DECL static int __attribute__((unused))
#define DECLARE_CMD(_name_, _handler_)
#define DECLARE_TEST_CMD(_name_, _handler_)
#define DECLARE_DEBUG_CMD(_name_, _handler_)
#define DECLARE_CONFIG_CMD(_name_, _handler_)
#define DECLARE_CLI_CHANNEL(_name_, _ctx_, _init_, _priv_)
#else
#define CLI_HANDLER_DECL static int
#if defined(__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050)
#define DECLARE_CMD(_name_, _handler_)                                            \
    struct cli_cmd_handler __attribute__((section("CMDS"))) __attribute__((used)) \
        __cmd_handler_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_TEST_CMD(_name_, _handler_)                                            \
    struct cli_cmd_handler __attribute__((section("TEST_CMDS"))) __attribute__((used)) \
        __cmd_test_handler_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_DEBUG_CMD(_name_, _handler_)                                            \
    struct cli_cmd_handler __attribute__((section("DEBUG_CMDS"))) __attribute__((used)) \
        __cmd_debug_handler_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_CONFIG_CMD(_name_, _handler_)                                            \
    struct cli_cmd_handler __attribute__((section("CONFIG_CMDS"))) __attribute__((used)) \
        __cmd_config_handler_config_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_CLI_CHANNEL(_name_, _ctx_, _init_, _priv_)                                    \
    struct cli_channel_handler __attribute__((section("CLI_CHANNELS"))) __attribute__((used)) \
        __cmd_channel_handler_##_name_ = {                                                    \
            .name = #_name_, .ctx = _ctx_, .init = _init_, .priv = _priv_};

#elif defined(__GNUC__)
#define DECLARE_CMD(_name_, _handler_)                               \
    struct cli_cmd_handler __attribute__((section(".cmd." #_name_))) \
        __cmd_handler_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_TEST_CMD(_name_, _handler_)                               \
    struct cli_cmd_handler __attribute__((section(".test.cmd." #_name_))) \
        __cmd_test_handler_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_DEBUG_CMD(_name_, _handler_)                               \
    struct cli_cmd_handler __attribute__((section(".debug.cmd." #_name_))) \
        __cmd_debug_handler_debug_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_CONFIG_CMD(_name_, _handler_)                               \
    struct cli_cmd_handler __attribute__((section(".config.cmd." #_name_))) \
        __cmd_config_handler_config_##_name_ = {.name = #_name_, .handler = _handler_};

#define DECLARE_CLI_CHANNEL(_name_, _ctx_, _init_, _priv_)                       \
    struct cli_channel_handler __attribute__((section(".cli.channel." #_name_))) \
        __cmd_channel_handler_##_name_ = {                                       \
            .name = #_name_, .ctx = _ctx_, .init = _init_, .priv = _priv_};
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif
int cli_printf(struct cli_cmd_ctx* ctx, const char* fmt, ...);
int cli_exec(char* cmd, struct cli_ctx* ctx);
void cli_dump_mem(struct cli_cmd_ctx* ctx, int address, int buffer, int len, int size);
void cli_handle_char(struct cli_ctx* ctx, char c);
void cli_set_bin_mode(struct cli_ctx* ctx, int bin_mode);
void cli_init();
void cli_poll();

/**
 * Retrieve the currently configured cli channel.
 *
 * @return the currently configured cli channel or NULL if none configured.
 */
struct cli_ctx* cli_get_ctx(void);

/**
 * Utilty function to convert a hex string into a binary buffer
 *
 * @param hex the null-terminated hex string
 * @param bin the buffer where to put the binary values
 * @param bin_len the length of the bin buffer
 *
 * @return int the number of bytes read from hex
 */
int cli_utils_hex2bin(char* hex, uint8_t* bin, int bin_len);

/**
 * Utility function to output a dump of binary data.
 *
 * @param ctx the cli context on which we want to write
 * @param address the address of the buffer to display
 * @param size the size of the buffer
 * @param data the data to display
 *
 * @return 0 if no error occureed
 */
int cli_utils_dump_data(struct cli_cmd_ctx* ctx, uint32_t address, uint32_t size, uint8_t* data);

/**
 * Output the welcome text when a console is opened.
 *
 * @param ctx the cli context on which we want to print the welcome text
 */
void cli_util_print_welcome(struct cli_cmd_ctx* ctx);

/**
 * Execute a cli script file
 *
 * @param ctx the cli context on which to execute the script
 * @param script the script text to execute (list of commands, one per line)
 * @param script_len length of the script text to execute
 *
 * @return int the last executed command status
 */
int cli_run_script(struct cli_cmd_ctx* ctx, char* script, int script_len);

#ifdef __cplusplus
}
#endif

/**
 * @} cli_cmd
 * @} infra
 */
#endif /* __CLI_CMD_H__ */
