#ifndef __COMMON_H__
#define __COMMON_H__

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "bsp_error.h"
#include "device.h"
#if defined(__CC_ARM)
#include "stdarg.h"
#endif
#include "elog.h"
#include "soc.h"
#include "systick.h"

/** @addtogroup infra
 * @{
 *
 * @addtogroup common common bsp components
 * @{
 */

/**
 * Write value to a register with a bit mask
 *
 * Attention, this macro will set all the bits in SETVAL even if not in CLEARMASK
 *
 * @param REG the register to write
 * @param SETVAL the value to set
 * @param CLEARMASK the bits to clear prior writing
 */
#define REG_WRITE_MASK(REG, SETVAL, CLEARMASK) (REG |= ((REG &= ~(CLEARMASK)) | (SETVAL)))

/**
 * Write value to a register
 *
 * @param REG the register to write
 * @param SETVAL the value to set
 */
#define REG_WRITE(REG, SETVAL) (REG = (SETVAL))

/**
 * Clear bits in a register
 *
 * @param REG the register to write
 * @param CLEARBITMASK the bits to clear
 */
#define REG_CLEAR_BIT_FIELD(REG, CLEARBITMASK) (REG &= (~CLEARBITMASK))

/**
 * Read and mask a register value
 *
 * @param REG the register to read
 * @param MASK the mask to apply
 */
#define REG_READ_MASK(REG, MASK) (REG & (MASK))

/**
 * Read a register value
 *
 * @param REG the register to read
 */
#define REG_READ(REG) (REG)

/*
 * Bit Operations
 */
#define BC_BIT(x) (1 << (x))
#define BC_GET_BIT(x, n) (((x) >> (n)) && 0x1)
#define DEFAULTU32BITMASK 0xFFFFFFFF

#define BCFW_VALIDATE_NULL(POINTER) \
    if (NULL == POINTER) {          \
        return ENULL_POINTER;       \
    }

#define BCFW_VALIDATE_RETURN(retval) \
    {                                \
        if (SUCCESS != retval)       \
            return retval;           \
    }

/**
 * Disable interrupts
 * @return current interrupt state.
 */
uint32_t interrupt_lock();
/**
 * Restore interrupt state
 *
 * @param flag interrupt state as returned by @ref interrupt_lock()
 */
void interrupt_unlock(uint32_t flag);

/**
 * This method performs the bsp initialization. should be called first
 * in application main()
 */
void bsp_init(void);

/** @} common */
/** @} infra */
#endif /*__COMMON_H__*/
