#ifndef _CMSIS_ELOG_INCLUDE_ELOG_DEF_H
#define _CMSIS_ELOG_INCLUDE_ELOG_DEF_H

#include <stdarg.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif
/**
 * @file elog.h
 * This header file elog.h defines the event logging module corresponding data structures.
.*  Elog facilitates the logging from various modules to render over log backends. logbackend
provides the
.* interfaces with Host using logical channel via RTT / UART..
.*
.* Log Format:       <Micro-seconds lasted> <LogLevel>< VarArgs>
.*
.* LogClass : CRITICAL,ERROR,WARNING,DEBUG,INFO
 */

/** @addtogroup infra infra
 * @{
 */

/** @addtogroup elog elog
 *  Documentation details of elog.
 *  @{
 */

/**
 * An enum type to identify supported log class type
 */
typedef enum {
    ELOG_LEVEL_VOID = 0,
    ELOG_LEVEL_CRITICAL, /**< defines the critical level log class */
    ELOG_LEVEL_ERROR,    /**< defines the error level log class */
    ELOG_LEVEL_WARNING,  /**< defines the warning level log class */
    ELOG_LEVEL_INFO,     /**< defines the info level log class */
    ELOG_LEVEL_DEBUG,    /**< defines the debug level log class */
    ELOG_LEVEL_INVALID   /**<define invalid log class, */
} elog_level_t;

struct log_module {
    /** Name of the log module */
    char* name;
    /** Default log level */
    elog_level_t def_log_level;
};

#define ELOG_MAX_MESSAGE_LENGTH 164

#define GET_LOG_INST(name) (&__log_##name)

#define BCLOG_RET_IF_MODULE_DISABLED(log_module, val) \
    if (!(log_module & val))                          \
        return;

#define ELOG_RET_IF_LEVEL_ABOVE(def_log_lvl, req_log_lvl) \
    if (req_log_lvl > def_log_lvl)                        \
        return;

/**
 *  LOGI: Used to prints the Information kind of logs
 *  Ex: BCLOGI(BC_MRBC,BC_LEVEL1," Received inputs validation done %s\n", __FUNCTION__);
 */
#define ELOGI(log_module, ...) elog_msg(log_module, ELOG_LEVEL_INFO, __VA_ARGS__)

/**
 *  LOGD: Used to prints the Debugging logs
 *  Ex: LOGD(BC_WDOG,BC_LEVEL2,"Variable value = %d  %s\n", ival, __FUNCTION__);
 */
#define ELOGD(log_module, ...) elog_msg(log_module, ELOG_LEVEL_DEBUG, __VA_ARGS__)

/**
 *  LOGW: Used to prints the Warning level logs
 *  Ex: LOGW(BC_UART,BC_LEVEL3,"Reached lower water mark level %d  %s\n", ival, __FUNCTION__);
 */
#define ELOGW(log_module, ...) elog_msg(log_module, ELOG_LEVEL_WARNING, __VA_ARGS__)

/**
 *  LOGE: Used to prints the Error detected logs
 *  Ex: LOGE(BC_SRTT,BC_LEVEL0,"Input arguments are out of range val = %d  %s\n", ival,
 * __FUNCTION__);
 */
#define ELOGE(log_module, ...) elog_msg(log_module, ELOG_LEVEL_ERROR, __VA_ARGS__)

/**
 *  LOGC: Used to prints the Critical logs, like reaching unexpected code regions
 *  Ex: LOGC(BC_GEN,BC_LEVEL4,"Reached unexptecd error condition ival = %d  %s\n", ival,
 * __FUNCTION__);
 */
#define ELOGC(log_module, ...) elog_msg(log_module, ELOG_LEVEL_CRITICAL, __VA_ARGS__)

/**
 *API to  intialises the elog module
 *@return     SUCCESS  elog module initialised successfully else EFAILURE
 */
int elog_init();

/**
 *API to set the new log_level configurations.Logs above the current level will be discarded.
 *@param[in] module Driver module name.
 *@param[in] log_level current interested logclass.
 *@see        elog_level_t
 *@return     SUCCESS  update the logclass sucessfully.
 *@return     EFAILURE Invalid argument passed
 */
int elog_set_log_level(const char* module, const elog_level_t log_level);

/**
 *API to get the current log_level configurations
 *@param[in] module Driver module name.
 *@param[out] log_level current interested logclass.
 *@see        elog_level_t
 *@return     SUCCESS  update the logclass sucessfully.
 *@return     EFAILURE Invalid argument passed
 */
int elog_get_log_level(const char* module, elog_level_t* log_level);

/**
 *API to print the logs over the HCI (Host Controller Interface)
 */
#ifndef CONFIG_LOG
#define elog_msg(...)
#else
/**
 *API to print the logs over the HCI (Host Controller Interface)
 *@param[in] module structure contains driver module information and log class
 *@param[in] req_log_level contains current request log level
 *@param[in] elogFmt Variable arguments logformat
 *
 *@see         elog_moduleId_t elog_level_t
 */
void elog_msg(struct log_module* module, elog_level_t req_log_level, const char* elogFmt, ...);
void elog_msgv(struct log_module* module, elog_level_t req_log_level, const char* elogFmt,
               va_list params);
#endif

/**
 *API to look for log module instance
 *@param[in] name Name of the module that uses ELOG
 *
 *@return      log module instance
 */
struct log_module* elog_module_lookup(const char* name);

/**
 *API to  Deintialises the elog module
 *@return     SUCCESS  elog module Deintialises successfully
 */
int elog_Deinit();

/**
 * Declare a per driver log module.
 */
#if defined(__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050)
#define DECLARE_LOG_MODULE(_name_, _lvl_)                             \
    struct log_module __attribute__((section("LOG_MODULES" #_name_))) \
        __log_##_name_ = {.name = #_name_, .def_log_level = _lvl_};
#elif defined(__GNUC__)
#define DECLARE_LOG_MODULE(_name_, _lvl_)                              \
    struct log_module __attribute__((section(".log_modules" #_name_))) \
        __log_##_name_ = {.name = #_name_, .def_log_level = _lvl_};
#endif

/** @} elog
 * @} infra
 */

#ifdef __cplusplus
}
#endif
#endif /* _CMSIS_ELOG_INCLUDE_ELOG_DEF_H */
