#ifndef __EVENT_LOG_H__
#define __EVENT_LOG_H__

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @addtogroup infra
 * @{
 */

/**
 * @addtogroup event_log
 *
 * Event log is a circular buffer containing binary log items for fast logging of events in a local
 * buffer.
 *
 * A separate component is responsible to flush the circular buffer asynchronously.
 *
 * @{
 */

#define EVENT_MASK_APP CONFIG_APP_EVENT_ID_OFFSET

/**
 * Event categories
 *
 * This enum specifies the possible event categories.
 *
 * Inside these categories, the events ids should be defined in a unique way.
 *
 * Event ids lower than 0x100 are text logs comming from elog subsystem.
 */
enum event_categories {
    EVENT_CAT_SOC = 0x100,
    EVENT_CAT_DRIVER = 0x10000,
    EVENT_CAT_INFRA = 0x20000,
    EVENT_CAT_POWER = 0x30000,
    EVENT_CAT_NETWORK = 0x40000
};

/* EVENT_INA_MEASURE is the first of the power category */
#define EVENT_INA_MEASURE (EVENT_CAT_POWER)

/* EVENT_WDT_KICK is the first of the soc category */
#define EVENT_WDT_KICK (EVENT_CAT_SOC)

#define CONFIG_LOG_BUFFER_SIZE 4096

struct event_header {
    uint32_t event_id;
    uint64_t timestamp;
    int data_size;
    uint8_t data[0];
} __attribute__((packed));

/**
 * Push an event to the event log buffer
 *
 * @param event_id a global identifier of the event
 * @param data a pointer to specific event data
 * @param len the length of the event data field
 *
 * @return 0 if event pushed sucessfully
 */
int event_log_push(uint32_t event_id, uint8_t* data, uint32_t len);

/**
 * Retrieve one event from the log buffer
 *
 * @param hdr the buffer where to place the event
 * @param len the size of the buffer passed ( in bytes ) it should be large enough
 *            to hold event header and event data.
 *
 * @return the size of the data retrieved
 */
int event_log_pull(struct event_header* hdr, int len);

/**
 * Initialize the event log buffer
 */
void event_log_init();

#ifdef __cplusplus
}
#endif

#endif /* __EVENT_LOG_H__ */