#include <cli_cmd.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "bsp_error.h"
#include "custom_sections.h"
#include "driver.h"
#include "version.h"

int cli_printf(struct cli_cmd_ctx* cmd_ctx, const char* fmt, ...) {
    va_list args;
    if (cmd_ctx == NULL)
        return 0;
    va_start(args, fmt);
    int ret = cmd_ctx->cli->vprintf(cmd_ctx->cli, fmt, args);
    va_end(args);
    return ret;
}

DECLARE_CUSTOM_SECTION(cmd, struct cli_cmd_handler, CMDS);
DECLARE_CUSTOM_SECTION(cmd_test, struct cli_cmd_handler, TEST_CMDS);
DECLARE_CUSTOM_SECTION(cmd_debug, struct cli_cmd_handler, DEBUG_CMDS);
DECLARE_CUSTOM_SECTION(cli_channel, struct cli_channel_handler, CLI_CHANNELS);

struct cli_ctx* cli_conf = NULL;

struct cli_ctx* cli_get_ctx(void) {
    return cli_conf;
}

struct cli_channel_handler* cli_channel_lookup(char* name) {
    struct cli_channel_handler* ptr = cli_channel_start;
    while (ptr < cli_channel_end) {
        if (!strcmp(ptr->name, name)) {
            return ptr;
        }
        ptr++;
    }
    return NULL;
}

void cli_init() {
    struct cli_channel_handler* chan = cli_channel_lookup(CONFIG_DEFAULT_CLI_BACKEND);
    if (chan) {
        if (chan->init) {
            chan->init(chan);
        }
        cli_conf = chan->ctx;
    }
}

void cli_poll() {
    static int init = 0;
    int c;
    if (!init) {
        cli_util_print_welcome(cli_conf->ctx);
        init = 1;
    }
    c = cli_conf->get_char(cli_conf);
    if (c >= 0) {
        cli_handle_char(cli_conf, (char)c);
        if ('\n' == c) {
            cli_handle_char(cli_conf, '\r');
        }
    }
}

#if defined(__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050)
char* strdup(const char* s) {
    size_t slen = strlen(s);
    char* result = malloc(slen + 1);
    if (result == NULL) {
        return NULL;
    }

    memcpy(result, s, slen + 1);
    return result;
}
#endif

struct cli_cmd_handler* find_cmd(char* name, struct cli_cmd_handler* start,
                                 struct cli_cmd_handler* end) {
    struct cli_cmd_handler* cmd = start;
    while (cmd < end) {
        if (!strcmp(name, cmd->name)) {
            return cmd;
        }
        cmd++;
    }
    return NULL;
}

void list_cmd(struct cli_cmd_handler* start, struct cli_cmd_handler* end, struct cli_cmd_ctx* ctx) {
    struct cli_cmd_handler* cmd = start;
    while (cmd < end) {
        cli_printf(ctx, "%s\r\n", cmd->name);
        cmd++;
    }
}

int cli_cmd_help(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    (void)argc;
    (void)argv;
    list_cmd(cmd_start, cmd_end, ctx);
    return 0;
}
DECLARE_CMD(help, cli_cmd_help);

int cli_handler(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    if (argc < 2 || !strcmp(argv[1], "list")) {
        struct cli_channel_handler* ptr = cli_channel_start;
        while (ptr < cli_channel_end) {
            cli_printf(ctx, "%10s %6s %c\r\n", ptr->name, ptr->ctx->bin_mode ? "bin" : "ascii",
                       (ptr->ctx == ctx->cli ? '*' : ' '));
            ptr++;
        }
        return 0;
    }
    return -1;
}
DECLARE_CMD(cli, cli_handler);

/**
 * Print arguments passed to the command
 */
int cli_cmd_echo(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    for (int i = 1; i < argc; i++) {
        cli_printf(ctx, "%s ", argv[i]);
    }
    cli_printf(ctx, "\r\n");
    return 0;
}
DECLARE_CMD(echo, cli_cmd_echo);

/**
 * Generate a true or false result.
 */
int cli_cmd_yesno(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    (void)argc;
    (void)ctx;
    return strcmp(argv[0], "yes");
}
DECLARE_CMD(yes, cli_cmd_yesno);
DECLARE_CMD(no, cli_cmd_yesno);

/**
 * Print result of previously executed command.
 */
int cli_cmd_ans(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    (void)argc;
    (void)argv;
    cli_printf(ctx, "%d\r\n", ctx->ans);
    return 0;
}
DECLARE_CMD(ans, cli_cmd_ans);

/**
 * Conditionaly execute command based on previous result code.
 *
 * this command has one implementation for 2 different commands
 * behavior is controlled with invocation name ( argv[0] )
 */
int cli_cmd_condexec(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    struct cli_cmd_handler* h = NULL;

    if (argc > 1)
        h = find_cmd(argv[1], cmd_start, cmd_end);

    if (h) {
        if ((!strcmp("ifok", argv[0]) && !ctx->ans) || (strcmp("ifok", argv[0]) && ctx->ans)) {
            h->handler(argc - 1, argv + 1, ctx);
        }
    }
    return ctx->ans; /* keep last error code */
}
DECLARE_CMD(ifok, cli_cmd_condexec);
DECLARE_CMD(iferr, cli_cmd_condexec);

void cli_print_arg(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    int indx = 0;

    cli_printf(ctx, "Args:");

    for (indx = 0; indx < argc; indx++) {
        cli_printf(ctx, " %s", argv[indx]);
    }
    cli_printf(ctx, "\n");
}

int cli_cmd_test(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    int ret = 0;
    struct cli_cmd_handler* h = NULL;

    if (argc > 1)
        h = find_cmd(argv[1], cmd_test_start, cmd_test_end);

    if (h) {
        ret = h->handler(argc - 1, argv + 1, ctx);
        cli_print_arg(argc, argv, ctx);
    } else {
        list_cmd(cmd_test_start, cmd_test_end, ctx);
    }
    return ret;
}
DECLARE_CMD(test, cli_cmd_test);

int cli_cmd_debug(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    int ret = 0;
    struct cli_cmd_handler* h = NULL;

    if (argc > 1)
        h = find_cmd(argv[1], cmd_debug_start, cmd_debug_end);

    if (h) {
        ret = h->handler(argc - 1, argv + 1, ctx);
        cli_print_arg(argc, argv, ctx);
    } else {
        list_cmd(cmd_debug_start, cmd_debug_end, ctx);
    }
    return ret;
}
DECLARE_CMD(debug, cli_cmd_debug);

#define CMD_HISTORY_COUNT 16

char* history_data[CMD_HISTORY_COUNT];
int history_idx = -1;
int history_count = 0;

void push_history(char* cmd) {
    if (history_data[0] && !strcmp(cmd, history_data[0])) {
        /* Ignore same command in history */
        return;
    }
    if (history_count == CMD_HISTORY_COUNT) {
        free(history_data[history_count - 1]);
    }
    for (int i = history_count; i > 0; i--) {
        if (i < CMD_HISTORY_COUNT) {
            history_data[i] = history_data[i - 1];
        }
    }
    history_data[0] = strdup(cmd);
    history_count++;
    if (history_count > CMD_HISTORY_COUNT) {
        history_count = CMD_HISTORY_COUNT;
    }
}

int split_arguments(char* buffer, char*** argv) {
    /* First pass: calculate argc */
    char* pc = buffer;
    int in_arg = 0;
    int argc = 0;
    int in_quotes = 0;

    while ((*pc) != '\0') {
        if (in_quotes && (*pc) == '\"') {
            in_quotes = 0;
        } else if ((*pc) == ' ' && in_arg) {
            in_arg = 0;
        } else if ((*pc) == '\"' && !in_quotes) {
            in_quotes = 1;
            argc++;
        } else if ((*pc) != ' ' && !in_arg && !in_quotes) {
            in_arg = 1;
            argc++;
        }
        pc++;
    }

    /* Second path, allocate argv and split in place */
    char** args = (char**)malloc(argc * sizeof(char*));
    char** tmp = args;

    in_arg = in_quotes = 0;
    pc = buffer;
    while ((*pc) != '\0') {
        if (in_quotes && (*pc) == '\"') {
            in_quotes = 0;
            *pc = '\0';
        } else if ((*pc) == ' ' && in_arg) {
            *pc = '\0';
            in_arg = 0;
        } else if ((*pc) == '\"' && !in_quotes) {
            *tmp++ = pc + 1;
            in_quotes = 1;
        } else if ((*pc) != ' ' && !in_arg && !in_quotes) {
            in_arg = 1;
            *tmp++ = pc;
        }
        pc++;
    }

    *argv = args;
    return argc;
}

int cli_exec(char* cmd, struct cli_ctx* ctx) {
    char** argv;
    char* line = "";

    line = strdup(cmd);
    int argc = split_arguments(line, &argv);
    struct cli_cmd_handler* h = find_cmd(line, cmd_start, cmd_end);
    if (h) {
        ctx->ctx->h = h;
        ctx->ctx->ans = h->handler(argc, argv, ctx->ctx);
    } else {
        cli_printf(ctx->ctx, "Command '%s' not found\r\n", cmd);
        ctx->ctx->ans = -1;
    }
    free(line);
    free(argv);

    return ctx->ctx->ans;
}

/**
 * Handle a character from a cli interface
 *
 * @param ctx the context of the cli interface
 * @param c the character to handle
 */
void cli_handle_char(struct cli_ctx* ctx, char c) {
    if (ctx->state == TTY_STATE_ESCAPE_2) {
        /*
         * Take into account escape sequences ( arrow up/down for history for now )
         * with RTT use "stty raw -echo && nc localhost 19021 && stty sane" to
         * connect with RTT server
         */
        switch (c) {
            case 'A': /* UP */
                if ((history_idx < (CMD_HISTORY_COUNT - 1)) && history_data[history_idx + 1] &&
                    strlen(&history_data[history_idx + 1][0])) {
                    history_idx += 1;
                    for (int i = 0; i < ctx->cmd_idx; i++) {
                        ctx->put(ctx, '\b');
                        ctx->put(ctx, ' ');
                        ctx->put(ctx, '\b');
                    }
                    strcpy(ctx->buffer, &history_data[history_idx][0]);
                    ctx->puts(ctx, ctx->buffer);
                    ctx->cmd_idx = strlen(ctx->buffer);
                }
                break;
            case 'B':
                if ((history_idx > 0) && strlen(&history_data[history_idx - 1][0])) {
                    history_idx -= 1;
                    for (int i = 0; i < ctx->cmd_idx; i++) {
                        ctx->put(ctx, '\b');
                        ctx->put(ctx, ' ');
                        ctx->put(ctx, '\b');
                    }
                    strcpy(ctx->buffer, &history_data[history_idx][0]);
                    ctx->puts(ctx, ctx->buffer);
                    ctx->cmd_idx = strlen(ctx->buffer);
                }
                break;
        }
        ctx->state = TTY_STATE_NORMAL;
        return;
    } else if (ctx->state == TTY_STATE_ESCAPE) {
        if (c == '[')
            ctx->state = TTY_STATE_ESCAPE_2;
        else
            ctx->state = TTY_STATE_NORMAL;
        return;
    }
    switch (c) {
        case 0x1b:
            ctx->state = TTY_STATE_ESCAPE;
            break;
        case 10:
            break;
        case 13:
            ctx->put(ctx, 10);
            ctx->put(ctx, c);
            ctx->buffer[ctx->cmd_idx] = 0;
            if (ctx->cmd_idx > 0) {
                push_history(ctx->buffer);
                cli_exec(ctx->buffer, ctx);
                history_idx = -1;
            }
            ctx->puts(ctx, "akida> ");
            ctx->cmd_idx = 0;
            ctx->buffer[ctx->cmd_idx] = 0;
            break;
        case 127: /* linux sends 127 */
        case 8:   /* terraterm sends 8 */
            if (ctx->cmd_idx > 0) {
                ctx->cmd_idx--;
                ctx->buffer[ctx->cmd_idx] = 0;
                ctx->put(ctx, '\b');
                ctx->put(ctx, ' ');
                ctx->put(ctx, '\b');
            }
            break;
        default:
            if (ctx->cmd_idx >= (int)sizeof(ctx->buffer))
                break; /* do not overrun buffer */
            ctx->buffer[ctx->cmd_idx++] = (char)c;
            ctx->put(ctx, c);
            break;
    }
}

void cli_set_bin_mode(struct cli_ctx* ctx, int bin_mode) {
    ctx->bin_mode = bin_mode;
}

int cli_cmd_history(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    (void)argc;
    (void)argv;
    for (int i = history_count - 1; i >= 0; i--) {
        cli_printf(ctx, "%3i %s\r\n", i, history_data[i]);
    }
    return 0;
}
DECLARE_CMD(history, cli_cmd_history);

int cli_version(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    (void)argc;
    (void)argv;
    cli_printf(ctx, "version: %s\r\n", __version_string);
    return 0;
}
DECLARE_CMD(version, cli_version);

static int nibble(char c) {
    if (c >= '0' && c <= '9') {
        return c - '0';
    }
    if (c >= 'a' && c <= 'f') {
        return c - 'a' + 10;
    }
    if (c >= 'A' && c <= 'F') {
        return c - 'A' + 10;
    }
    return 0;
}

static int hex2byte(char* hex) {
    return (nibble(hex[0]) << 4) | nibble(hex[1]);
}

int cli_utils_hex2bin(char* hex, uint8_t* bin, int bin_len) {
    int hex_len = strlen(hex) / 2;

    for (int i = 0; i < hex_len; i++) {
        if (i < bin_len) {
            bin[i] = hex2byte(hex + 2 * i);
        } else {
            return i;
        }
    }

    return hex_len;
}

#define DUMP_WRAPAROUND_COUNT 16

int cli_utils_dump_data(struct cli_cmd_ctx* ctx, uint32_t address, uint32_t size, uint8_t* data) {
    for (uint32_t i = 0; i < size; i++) {
        if ((i % DUMP_WRAPAROUND_COUNT) == 0) {
            if (i)
                cli_printf(ctx, "\r\n");
            cli_printf(ctx, "%08x:", address + i);
        }
        cli_printf(ctx, " %02x", data[i]);
    }
    cli_printf(ctx, "\r\n");
    return 0;
}

void cli_util_print_welcome(struct cli_cmd_ctx* ctx) {
    ctx->cli->puts(ctx->cli, "\r\nWelcome to Akida monitor (");
    ctx->cli->puts(ctx->cli, __version_string);
    ctx->cli->puts(ctx->cli, ")");
    ctx->cli->puts(ctx->cli, "\r\nakida> ");
}

int cli_run_script(struct cli_cmd_ctx* ctx, char* script, int script_len) {
    (void)script_len;
    char* tok_state;
    char* line = strtok_r(script, "\n", &tok_state);
    while (line != NULL) {
        int ret = cli_exec(line, ctx->cli);
        if (ret != 0) {
            cli_printf(ctx, "error executing '%s' (%d)\r\n", line, ret);
        }
        line = strtok_r(NULL, "\n", &tok_state);
    }
    return 0;
}
