#include <elog.h>
#include <stdlib.h>
#include <string.h>
#include "bsp_error.h"
#include "cli_cmd.h"
#include "common.h"
#include "custom_sections.h"
#include "event_log.h"
#include "log_backend.h"

struct log_common* gpHci = NULL;

DECLARE_CUSTOM_SECTION(log_module, struct log_module, LOG_MODULES);

uint8_t elogmessage[ELOG_MAX_MESSAGE_LENGTH];

static struct log_module default_module = {
    .name = "default",
    .def_log_level = ELOG_LEVEL_INFO,
};

static const char* elog_get_level_name(const elog_level_t log_lvl) {
    switch (log_lvl) {
        case ELOG_LEVEL_INFO:
            return "I";
        case ELOG_LEVEL_DEBUG:
            return "D";
        case ELOG_LEVEL_WARNING:
            return "W";
        case ELOG_LEVEL_ERROR:
            return "E";
        case ELOG_LEVEL_CRITICAL:
            return "C";
        default:
            return "U";
    }
}

int elog_set_log_level(const char* module, const elog_level_t log_lvl) {
    struct log_module* log_m = NULL;
    if (NULL != (log_m = elog_module_lookup(module))) {
        log_m->def_log_level = log_lvl;
    } else {
        return EFAILURE;
    }
    return SUCCESS;
}

int elog_get_log_level(const char* module, elog_level_t* log_lvl) {
    struct log_module* log_m = NULL;
    if (NULL != (log_m = elog_module_lookup(module))) {
        *log_lvl = log_m->def_log_level;
    } else {
        return EFAILURE;
    }
    return SUCCESS;
}

int elog_init() {
    if (!strcmp(CONFIG_DEFAULT_LOG_BACKEND, "uart")) {
        /* Lower the Uart modules log level */
    }

    if (NULL == (gpHci = log_backend_lookup(CONFIG_DEFAULT_LOG_BACKEND))) {
        return EFAILURE;
    }

    return SUCCESS;
}

int elog_Deinit() {
    gpHci = NULL;

    return SUCCESS;
}

void elog_msgv(struct log_module* module, elog_level_t req_log_level, const char* elogFmt,
               va_list params) {
    uint8_t bufIdx = 0;

    if (module == NULL) {
        module = (struct log_module*)&default_module;
    }

    /* Validate the log_lvl */
    ELOG_RET_IF_LEVEL_ABOVE(module->def_log_level, req_log_level);

    /*This section can be enabled when rtc_time.h file is added.*/
    /* Read the current time */
    uint64_t us = get_time_us();
    bufIdx = sprintf((char*)elogmessage, "%6lu.%06lu ", (uint32_t)(us / 1000000),
                     (uint32_t)(us % 1000000));

    bufIdx += sprintf((char*)elogmessage + bufIdx, "%s ", elog_get_level_name(req_log_level));

#ifdef HAS_OS
    bufIdx += sprintf((char*)elogmessage + bufIdx, "TID:%d ", *(int*)osThreadGetId());
#endif
    /* Add the time stamps, logclass and log level once the data is transmitted over RTT */
    if ((NULL != gpHci) && (NULL != gpHci->write))
        gpHci->write((char*)elogmessage, bufIdx);

    bufIdx = 0;
    bufIdx +=
        vsnprintf((char*)elogmessage + bufIdx, ELOG_MAX_MESSAGE_LENGTH - bufIdx, elogFmt, params);

#ifdef CONFIG_EVENT_LOG
    event_log_push(req_log_level, elogmessage, bufIdx);
#endif

    bufIdx += snprintf((char*)elogmessage + bufIdx, ELOG_MAX_MESSAGE_LENGTH - bufIdx, "\r\n");

    if ((NULL != gpHci) && (NULL != gpHci->write))
        gpHci->write((char*)elogmessage, bufIdx);
}

void elog_msg(struct log_module* module, elog_level_t req_log_level, const char* elogFmt, ...) {
    va_list elogParamList;
    va_start(elogParamList, elogFmt);

    elog_msgv(module, req_log_level, elogFmt, elogParamList);

    va_end(elogParamList);
    return;
}

struct log_module* elog_module_lookup(const char* name) {
    struct log_module* log_m = log_module_start;
    while (log_m < log_module_end) {
        if (!strcmp(log_m->name, name)) {
            return log_m;
        }
        log_m++;
    }
    return NULL;
}

static void elog_cli_print_all(struct cli_cmd_ctx* ctx) {
    elog_level_t level;
    struct log_module* log_m = log_module_start;
    while (log_m < log_module_end) {
        elog_get_log_level(log_m->name, &level);
        cli_printf(ctx, "%12s: %s\r\n", log_m->name, elog_get_level_name(log_m->def_log_level));
        log_m++;
    }
    cli_printf(ctx, "%12s: %s\r\n", default_module.name,
               elog_get_level_name(default_module.def_log_level));
}
static void elog_set_all(elog_level_t level) {
    struct log_module* log_m = log_module_start;
    while (log_m < log_module_end) {
        log_m->def_log_level = level;
        log_m++;
    }
}
/**
 * @page Commands CLI
 * @section elog ELOG
 * @ref ELOG-CLI-1 \n
 */
/**
 * @page Commands
 * @subsection ELOG-CLI-1 log level [{module_name|all}[new_log_level]]
 * CLI to set/get the log level per module or for all\n
 * __syntax: log level [{module_name|all}[new_log_level]__\n
 * __log_level__: 1-5:CRITICAL|WARNING|ERROR|INFO|DEBUG \n
 **/

CLI_HANDLER_DECL loglevel_handler(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    elog_level_t level;
    int ret = SUCCESS;
    const char* mod;
    bool check = true;

    if (((2 == argc) || (argc == 3)) && (!strncmp(argv[1], "level", 5))) {
        if (2 == argc) {
            elog_cli_print_all(ctx);
        } else {
            mod = argv[2];
            ret = elog_get_log_level(mod, &level);
            cli_printf(ctx, "%12s: %s\r\n", mod, elog_get_level_name(level));
        }

    } else if (4 == argc) {
        mod = argv[2];
        level = atoi(argv[3]);
        if ((level < 1) || (level > 5)) {
            cli_printf(ctx, "Invalid log level, choose 1-5 \n");
            return EFAILURE;
        }
        if (!strcmp(mod, "all")) {
            elog_set_all(level);
            default_module.def_log_level = level;
            cli_printf(ctx, "log_level is set to: %s \n", elog_get_level_name(level));
        } else {
            ret = elog_set_log_level(mod, level);
        }
    } else {
        check = false;
    }

    if (false == check) {
        cli_printf(ctx, "%s command not supported\n", argv[0]);
        cli_printf(ctx, "Syntax: log level [{module|all} {log_level}]\n");
        return EFAILURE;
    }
    if (ret) {
        cli_printf(ctx, "Error: Incorrect module name\n");
    }
    return ret;
}
DECLARE_CMD(log, loglevel_handler);
