#include <stdlib.h>
#include <string.h>

#include "cli_cmd.h"
#include "common.h"
#include "event_log.h"

#include "uart.h"

static struct logbuf {
    uint8_t* wptr;
    uint8_t* rptr;
    int size;
    uint8_t data[CONFIG_LOG_BUFFER_SIZE];
} log_buffer, *buf;

static uint8_t* _log_buffer_read(uint8_t* ptr, uint8_t* dst, int len);

/**
 * buffer free
 *
 * @return the number of bytes available in the log buffer
 */
static uint32_t _buffer_free() {
    int diff = buf->wptr - buf->rptr;
    if (diff < 0) {
        return -diff;
    } else {
        return buf->size - diff;
    }
}

/**
 * Advance read pointer one log item
 *
 * This method 'discards' the older item in the log buffer. It is used when
 * room is needed in the buffer to add a new item.
 */
static void _log_buffer_read_advance() {
    struct event_header hdr;
    _log_buffer_read(buf->rptr, (uint8_t*)&hdr, sizeof(hdr));
    buf->rptr += sizeof(struct event_header) + hdr.data_size;
    if ((buf->rptr - buf->data) >= buf->size) {
        buf->rptr -= buf->size;
    }
}

/**
 * Allocate data in the buffer
 *
 * This method 'allocates' data in the buffer by incrementing the write pointer
 * and discarding old items if required.
 *
 * @param size the number of bytes to allocate for the new item
 *
 * @return the write pointer to write the item to
 */
static uint8_t* _buffer_alloc(uint32_t size) {
    uint8_t* ret = buf->wptr;
    uint32_t flags = interrupt_lock();
    while (_buffer_free() <= size) {
        /* drop oldest events if no room availble */
        _log_buffer_read_advance();
    }
    buf->wptr += size;
    if ((buf->wptr - buf->data) >= buf->size) {
        buf->wptr -= buf->size;
    }
    interrupt_unlock(flags);
    return ret;
}

/**
 * Write to the log buffer taking care of wraparound
 *
 * This method copies data to the buffer taking care of wrapping around if end
 * of circular buffer is reached.
 *
 * @param ptr the write pointer in the buffer
 * @param src the pointer to the data to write to the buffer
 * @param len the size of data to write to the buffer
 *
 * @return the new write pointer
 */
static uint8_t* _log_buffer_write(uint8_t* ptr, uint8_t* src, int len) {
    int left = buf->size - (ptr - buf->data);
    uint8_t* ret;
    if (len >= left) {
        memcpy(ptr, src, left);
        memcpy(buf->data, src + left, len - left);
        ret = buf->data + (len - left);
    } else {
        memcpy(ptr, src, len);
        ret = ptr + len;
    }
    return ret;
}

/**
 * Read from the log buffer taking care of wraparound
 *
 * This method copies data from the buffer taking care of wrapping around if end
 * of circular buffer is reached.
 *
 * @param ptr the read pointer in the buffer
 * @param src the pointer to the location where to copy data from the buffer
 * @param len the size of data to read
 *
 * @return the new read pointer
 */
static uint8_t* _log_buffer_read(uint8_t* ptr, uint8_t* dst, int len) {
    int left = buf->size - (ptr - buf->data);
    uint8_t* ret;
    if (len > left) {
        memcpy(dst, ptr, left);
        memcpy(dst + left, buf->data, len - left);
        ret = buf->data + (len - left);
    } else {
        memcpy(dst, ptr, len);
        ret = ptr + len;
    }
    if (ret >= (buf->data + buf->size)) {
        ret -= buf->size;
    }
    return ret;
}

/**
 * Push a log event to the log buffer
 *
 * This method allocates space and copies data from the log event and data passed
 *
 * @param hdr the log event header structure to write
 * @param src the log event data to write can be NULL
 * @param len the size of the event data parameter if not NULL
 *
 * @return 0 if no error occured
 */
static int _log_buffer_push(struct event_header* hdr, uint8_t* event_data, uint32_t len) {
    uint8_t* ptr = _buffer_alloc(sizeof(*hdr) + len);
    ptr = _log_buffer_write(ptr, (uint8_t*)hdr, sizeof(*hdr));
    if (len) {
        ptr = _log_buffer_write(ptr, event_data, len);
    }
    return 0;
}

int event_log_push(uint32_t event_id, uint8_t* data, uint32_t len) {
    struct event_header hdr;
    hdr.event_id = event_id;
    hdr.timestamp = get_time_ms();
    hdr.data_size = len;
    _log_buffer_push(&hdr, data, len);
    return 0;
}

int event_log_pull(struct event_header* hdr, int len) {
    uint8_t* ptr = buf->rptr;
    int count = 0;

    uint32_t flags = interrupt_lock();
    if (buf->wptr == buf->rptr) {
        interrupt_unlock(flags);
        return -1;
    }
    _log_buffer_read_advance();

    ptr = _log_buffer_read(ptr, (uint8_t*)hdr, sizeof(*hdr));
    if (hdr->data_size) {
        count = hdr->data_size;
        if (count > (len - (int)sizeof(*hdr))) {
            count = len - sizeof(*hdr);
        }
        ptr = _log_buffer_read(ptr, &hdr->data[0], count);
    }
    interrupt_unlock(flags);
    return count + sizeof(*hdr);
}

void event_log_init() {
    buf = &log_buffer;
    buf->size = sizeof(buf->data);
    buf->rptr = buf->wptr = &buf->data[0];
}

int event_log_handler(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    struct event_header* log_hdr = (struct event_header*)malloc(256);

    if (argc == 1) {
        while (event_log_pull(log_hdr, 256) >= 0) {
            cli_printf(ctx, "%8d.%03d ", (uint32_t)(log_hdr->timestamp / 1000),
                       (uint32_t)(log_hdr->timestamp % 1000));
            switch (log_hdr->event_id) {
                case EVENT_INA_MEASURE:
                    cli_printf(ctx, "ina %dmA %dmV\r\n", *(uint32_t*)(log_hdr->data + 4),
                               *(uint32_t*)(log_hdr->data) / 1000);
                    break;
                case EVENT_WDT_KICK:
                    cli_printf(ctx, "wdt kick\r\n");
                    break;
                default:
                    if (log_hdr->event_id < 0x100) {
                        log_hdr->data[log_hdr->data_size] = 0;
                        cli_printf(ctx, "%d %s\r\n", log_hdr->event_id, log_hdr->data);
                    } else {
                        cli_printf(ctx, "%d %d\r\n", log_hdr->event_id, log_hdr->data_size);
                    }
                    break;
            }
        }
    } else if (argc > 2 && !strcmp(argv[1], "push")) {
        uint32_t event_id = strtoul(argv[2], NULL, 0);
        event_log_push(event_id, NULL, 0);
    }
    return 0;
}
DECLARE_CMD(event_log, event_log_handler);
