#include "systimer.h"
#include <stddef.h>

#include "common.h"
#include "interrupt.h"
#include "systick.h"

/* Ordered list of timed works */
struct systimer* sys_timer = NULL;

/**
 * Remove a timer item from the timer list
 * This function required to be called with interrupt disabled
 *
 * @param timer the timer to dequeue
 *
 * @return error code incase of failure or 0 if success
 */
static int systimer_dequeue(struct systimer* timer) {
    int found = 0;

    /* remove from pending work if present */
    struct systimer* ptimer = NULL;

    /* remove from timed work list if found */
    ptimer = sys_timer;
    if (timer == ptimer) {
        sys_timer = sys_timer->next;
        found = 1;
    } else if (ptimer != NULL) {
        while (ptimer->next && ptimer->next != timer) {
            ptimer = ptimer->next;
        }
        if (ptimer->next != NULL) {
            ptimer->next = ptimer->next->next;
            found = 1;
        }
    }

    if (!found) {
        return EFAILURE;
    }

    return SUCCESS;
}

/**
 * Insert a timer item in the timer list.
 * This function required to be called with interrupt disabled
 *
 *
 * @param timer the timer to queue
 * @param delay the delay in milliseconds after which the timer has to be executed.
 *
 * @return void
 */
static void systimer_queue_delayed(struct systimer* timer, uint32_t delay) {
    timer->sheduled_time = get_time_ms() + delay;

    /* Remove the work from queue if present */
    systimer_dequeue(timer);

    /* insert to list ordered by scheduled time */
    timer->next = NULL;

    if (sys_timer == NULL) {
        sys_timer = timer;
    } else {
        if (timer->sheduled_time < sys_timer->sheduled_time) {
            timer->next = sys_timer;
            sys_timer = timer;
        } else {
            struct systimer* ptr = sys_timer->next;
            struct systimer* prev = sys_timer;
            while (ptr && (ptr->sheduled_time < timer->sheduled_time)) {
                prev = ptr;
                ptr = ptr->next;
            }
            prev->next = timer;
            timer->next = ptr;
        }
    }
}

int systimer_register(struct systimer* timer, uint32_t delay) {
    BCFW_VALIDATE_NULL(timer);

    uint32_t flags = interrupt_lock();

    systimer_queue_delayed(timer, delay);

    interrupt_unlock(flags);

    return SUCCESS;
}

int systimer_deregister(struct systimer* timer) {
    BCFW_VALIDATE_NULL(timer);

    int ret = SUCCESS;

    uint32_t flags = interrupt_lock();

    ret = systimer_dequeue(timer);

    interrupt_unlock(flags);

    return ret;
}

void systimer_run(uint64_t time_ms) {
    struct systimer* temp_timer = NULL;

    while (sys_timer && (time_ms >= sys_timer->sheduled_time)) {
        /* timer will be set to null or next pending work */
        temp_timer = sys_timer;

        sys_timer = sys_timer->next;

        if (temp_timer->cb) {
            temp_timer->cb(temp_timer);
        }
    }
}

int systimer_pending(void) {
    if (sys_timer && (get_time_ms() < sys_timer->sheduled_time)) {
        return 1;
    }

    return 0;
}

void systimer_init(void) {
    systick_register_cb(systimer_run);
}
